// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$DESCRIPTION 'ALISP3 v0.4.3b by Ketmar // Vampire Avalon. dedicated to Enota'}

{$INCLUDE alisp3def.inc}

{$IFNDEF ALISP3_DEBUG}
{$MESSAGE ''}
{$MESSAGE 'compiling ALISP3 v0.4.3b by Ketmar // Vampire Avalon'}
{$MESSAGE 'ALISP is dedicated to Enota, the girl i''ll never forget'}
{$IFDEF ALISP3_MULTIENGINE}
  {$MESSAGE 'multiengine version'}
{$ENDIF}
{$MESSAGE ''}

{$ELSE}

{$IFDEF ALISP3_MULTIENGINE}
  {$MESSAGE ''}
  {$MESSAGE 'multiengine ALISP3'}
  {$MESSAGE ''}
{$ENDIF}

{$ENDIF}
unit alisp3U;

interface

uses
  Windows, SysUtils, StrUtils, Classes;


const
  (* version info *)
  alispVersionHi = 0;
  alispVersionLo = 4;
  alispVersionMod = 3;
  alispVersionFlags = -1; // -2:alpha; -1: beta; 0: normal; 1..255: RC
  alispVersionString = '0.4.3b';
  alispName = 'ALISP3 internal Avalonian edition';
  //alispCompileTime = __DATE__+'  '+__TIME__;


  AL_MAX_POOL_SIZE: Integer = $200000;
    // max number of cells: 2097152; max memory used: 16 mb
  AL_MIN_POOL_LEFT: Integer = 8192;
    // alloc more cells when there's less than AL_MIN_POOL_LEFT cells left
  AL_POOL_GROW_DELTA: Integer = 64*1024;
  AL_CALL_TRACE_DEPTH: Integer = 16; // # of names to show in ShowCallStack
  AL_INPUT_BUFFER_SIZE = 8192;

  DT_INTEGER = $01;
  DT_FLOAT = $02;
  DT_STRING = $03;
  DT_PRIMITIVE = $04;
  DT_VECTOR = $05;  // TList of integers
  DT_BYTEVECTOR = $06;
  DT_TOBJECT = $07;  // TLispObject


type
  EALispError = class (Exception);
  EALispOutOfMemoryError = class (EALispError);

  {$IFDEF ALISP3_MULTIENGINE}
  PALISP3 = Pointer;
  {$ENDIF}

  TALPrimitiveFn = function ({$IFDEF ALISP3_MULTIENGINE}_eng: PALISP3;{$ENDIF}n: Integer): Integer;
    // n: arglist (no primitive name passed!)

  TALDebuggerProc = function ({$IFDEF ALISP3_MULTIENGINE}_eng: PALISP3;{$ENDIF}n: Integer; beforeEval: Boolean): Integer;
    // n: eval arg/eval result
    // result <> -1: new EVAL value

  PALPrimitive = ^TALPrimitive;
  TALPrimitive = record
    name: string;
    func: TALPrimitiveFn;
    macro: Boolean;
    minArg, maxArg: Integer; // maxArg < 0: not limited
  end;

  PALInStreamInfo = ^TALInStreamInfo;
  TALInStreamInfo = record
    inStream: TStream;
    fileName: string;
    rejectedChar: Char;    // unread character
    line: Integer;         // input line number
    loadDepth: Integer;    // nesting level of ALLoadXXX cmds

    buf: packed array [0..AL_INPUT_BUFFER_SIZE-1] of Char;
    bufPos: Integer;
    bufRead: Integer;
  end;

  TLispObject = class (TObject)
  public
    //constructor Create (); virtual; abstract;
    procedure Externalize (st: TStream); virtual; abstract;
    procedure Internalize (st: TStream); virtual; abstract;
  end;

  {$IFDEF ALISP3_DOUBLE_FLOAT}
  TALFloat = Double;
  {$ELSE}
  TALFloat = Single;
  {$ENDIF}

  TALListSymbolsCallback = function (const name: string; value: Integer; user: Pointer): Boolean;
    // return false to stop enumeration


{$IFDEF ALISP3_MULTIENGINE}
function ALSpawnEngine (): PALISP3;
procedure ALKillEngine (var _eng: PALISP3);
{$ENDIF}

procedure ALInitInternals ();
procedure ALClearInternals ();

procedure ALDefineStdPrimitives ();

procedure ALGetLineInfo (n: Integer; out fn: string; out line, pos: Integer);

procedure ALAddPrimitive (const pName: string; pFunc: TALPrimitiveFn;
  pMacro: Boolean; pMinArg, pMaxArg: Integer);
procedure ALRebindPrimitives ();

procedure ALReinit ();

procedure ALClearState ();
  // reset all stacks, frames, etc.

procedure ALRegisterGCRoot (n: Integer);
  // this cell will become GC root
procedure ALUnregisterGCRoot (n: Integer);
function ALGC (): Integer;
  // return the number of the free cells
function ALCountFree (): Integer;
  // count the cells in the freelist

procedure ALPushCell (n: Integer);
  // save the cell N on the stack
function ALPopCells (k: Integer): Integer;
  // pop K cells off the stack and return
  // the most recently popped one

function ALCons (pcar, pcdr: Integer): Integer;
  // allocate a fresh cell and initialize with PCAR,PCDR
  // PCAR & PCDR will not be recycled during ALCons()'ing

procedure ALEnter (n: Integer);
  // enter lambda. N is a function name (atom), [M]LAMBDA or NIL
  // set up new frame for local vars
procedure ALLeave ();
  // leave lambda, restore previous frame
procedure ALDumpFrames (dumpLocalFn: Boolean = false);

procedure ALPrint (n: Integer);
procedure ALPrintNoQ (n: Integer);
function ALPrintToStr (n: Integer): string;
function ALEscapeStr (const s: string): string;
function ALEscapeStrEx (const s: string; quoteAnyway: Boolean; quoteCh: Char): string;

// the following functions adds a new atom.
// the resulting atom is unprotected.
function ALNewStrAtom (const s: string): Integer;
function ALNewIntAtom (value: Integer): Integer;
function ALNewFloatAtom (value: TALFloat): Integer;
function ALNewObjectAtom (value: TObject): Integer;
function ALNewVector (size: Integer): Integer;
function ALNewByteVector (size: Integer): Integer;
function ALNewVectorFromList (nlst: Integer): Integer;
function ALNewContextAtom (): Integer;

// ALEnterContext()/ALLeaveContexts(): no calls to ALCons()!
function ALEnterContext (ctxatom: Integer): Boolean;
procedure ALLeaveContexts (cnt: Integer);
function ALContextStackDepth (): Integer;

function ALGetAtomType (n: Integer): Integer;
  // returns DT_XXX or -1 if not an atom
function ALGetAtomStr (n: Integer): string;
  // returns string value or '' if not an atom
function ALGetAtomInt (n: Integer): Integer;
  // returns int value or 0 if not a number
function ALGetAtomFloat (n: Integer): TALFloat;
  // returns float value or 0 if not a number
function ALGetAtomObject (n: Integer): TObject;
  // returns int value or nil if not an object
function ALGetVectorItem (n, index: Integer): Integer;
  // for normal and byte vectors
procedure ALSetVectorItem (n, index, value: Integer);
  // for normal and byte vectors
function ALGetVectorLength (n: Integer): Integer;
  // for normal and byte vectors
procedure ALSetVectorLength (n, newLen: Integer);
  // for normal and byte vectors
function ALGetByteVectorDataPtr (n: Integer): Pointer;

// ALFindXXX returns symbol pair or NIL
function ALFindSymbolInList (const s: string; symList: Integer): Integer;
function ALFindSymbol (const s: string): Integer;
  // find a symbol named S in the symbol table
  // (starting from local frames)
function ALFindSymbolEx (const s: string; si: Integer = -1): Integer;
  // find a symbol named S in the symbol table
  // (starting from local frames)
  // processes context dereferences
function ALFindGlobalSymbol (const s: string): Integer;
// ALSetXXX returns symbol pair
function ALSetSymbol (const s: string; v: Integer = -1): Integer;
function ALSetSymbolEx (const s: string; v: Integer = -1; si: Integer = -1;
  onlyTopFrame: Boolean = false; onlyTopCtx: Boolean = false;
  viewGlobals: Boolean = false): Integer;
// if onlyTopFrame is true and viewGlobals is false: do not search in globals
function ALSetGlobalSymbol (const s: string; v: Integer = -1;
  onlyTopFrame: Boolean = false): Integer;
procedure ALDeleteSymbolFromCurrentContext (const s: string);
  // what a wonderful name! %-)
procedure ALSetSymbolProtectionFlag (n: Integer; value: Boolean);
function ALGetSymbolProtectionFlag (n: Integer): Boolean;
procedure ALEnumSymbolsInCurrentContext (CBack: TALListSymbolsCallback; user: Pointer);
  // another cool name...
procedure ALEnumContextSymbols (ctx: Integer; CBack: TALListSymbolsCallback; user: Pointer);
  // ctx: context atom
function ALGetMainContext (): Integer;
  // returns context atom

function ALGetStringFromPool (si: Integer): string;

function ALNil (): Integer;
function ALTrue (): Integer;
function ALFalse (): Integer;
function ALLambda (): Integer;
function ALMLambda (): Integer;
function ALQuote (): Integer;
function ALXQuote (): Integer;
function ALSymbols (): Integer;
function ALFrames (): Integer;

function ALIsContext (n: Integer): Boolean;
function ALIsNil (n: Integer): Boolean;
function ALIsTrue (n: Integer): Boolean;
function ALIsFalse (n: Integer): Boolean;
function ALIsAtom (n: Integer): Boolean; // is N a 'real' (non-NIL) atom?

function ALError (const m: string; n: Integer = -1): Integer;
function ALThrowError (const m: string; n: Integer = -1): Integer;

function ALCar (n: Integer): Integer;
function ALCdr (n: Integer): Integer;

procedure ALSetCar (n, v: Integer);
procedure ALSetCdr (n, v: Integer);


procedure ALNewInStream (st: TStream; const fName: string; out old: TALInStreamInfo);
procedure ALRestoreInStream (const si: TALInStreamInfo);

function ALReadExpr (): Integer;
  // -1: EOF
function ALInterpretNext (): Integer;
  // read expression and interpret it
  // -1: EOF
function ALEvalStr (const s: string): Integer;

function ALEval (n: Integer): Integer;

function ALLoadStream (st: TStream; const fName: string = ''): Integer;
  // load S-expressions from the stream.
  // return result of evaluation.
function ALLoadFile (const fn: string): Integer;
  // load S-expressions from the external file.
  // return result of evaluation.

function ALNextStrArg (const funcName: string; var n: Integer; out s: string; out res: Integer): Boolean;
function ALNextIntArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
function ALNextFloatArg (const funcName: string; var n: Integer; out v: TALFloat; out res: Integer): Boolean;
function ALNextVectorArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
function ALNextByteVectorArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
function ALNextObjectArg (const funcName: string; var n: Integer; out v: TObject; out res: Integer): Boolean;

function ALNextAnyVectorArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
  // byte vector or ordinary vector

function ALGetBreakFlag (): Integer;
procedure ALSetBreakFlag (new: Integer);
function ALGetAList (): Integer;
function ALGetSymbols (): Integer;
procedure ALSetDebugger (dbgProc: TALDebuggerProc);

procedure ALDumpPrimitives ();

function ALGetDefPathList (): string;
procedure ALSetDefPathList (const s: string);

function FileSearchEx (const name, pathList: string): string;


const
  ALBREAK_NONE = 0;
  ALBREAK_ERROR = -1;
  ALBREAK_RETURN = 1;
  ALBREAK_BREAK = 2;
  ALBREAK_THROW = 3;
  ALBREAK_CONT = 4;
  ALBREAK_QUIT = 5;


implementation


{$INCLUDE alisp3IntListDef.pas}
{$INCLUDE alisp3IntListImp.pas}

const
  RBT_BLACK = 0;
  RBT_RED = 1;
  RBT_UNUSED = 255;

  GROW_DELTA = 128;

{$INCLUDE alisp3strPoolU.pas}
{$INCLUDE alisp3symPoolU.pas}


const
  MAX_LOAD_LEVEL = 64; // max. load level

  AFLAG = Integer($80000000);
  MFLAG = $40000000;
  PROTFLAG = $20000000; // "protected" flag. can be set on symbols
  TYPE_MASK = $FF;

  // for ReadChar()
  EOF = #0; // EOF indicator
  EOL = #1; // EOL indicator
  ERR = #2; // error indicator

  // for ReadExpr()
  EOFa = -1;   // EOF indicator
  DOTa = -2;   // dot character
  RPARENa = -3;// right parenthesis ")"
  RSQPARENa = -4;// right square parenthesis "]"


type
  TLispByteVector = class
    vector: packed array of Byte;
  end;

  TLineInfo = record
    fileName: string;
    line: Integer;
    pos: Integer;
  end;

{$IFDEF ALISP3_MULTIENGINE}
  PXALISP3 = ^TXALISP3;
{$ENDIF}

  TLispContextObject = class
  protected
    {$IFDEF ALISP3_MULTIENGINE}
    _eng: PXALISP3;
    {$ENDIF}
    fSymbols: Integer;     // symbol table (GC root)
    fSymTree: TSymbolTree; // keep in sync with fSymbols!

  public
    constructor Create ({$IFDEF ALISP3_MULTIENGINE}e: PXALISP3{$ENDIF});
    destructor Destroy (); override;
  end;

{$IFDEF ALISP3_MULTIENGINE}
  TXALISP3 = record
{$ELSE}
  var
{$ENDIF}
    al_DefPathList: string;
    al_BreakFlag: Integer;
      // -1: error
      // 0: no error/break/cont/throw/etc...
      // 1: return
      // 2: break
      // 3: runerror (ex-bottom)
      // 4: continue
    al_ShowCSFlag: Boolean;
      // show call stack flag. default: TRUE
    al_VerbLoadFlag: Boolean;
      // verbose LOAD? default: TRUE
    al_Optimize: Boolean;
      // do lambda optimizations? default: TRUE
    al_ThrowSysError: Boolean;
      // do THROW instead of ALError? default: FALSE
    al_Debugger: TALDebuggerProc;
      // return evaluated value or -1

    al_Primitives: TList;  // PALPrimitive

    al_CAR: array of Integer;
    al_CDR: array of record
        case Byte of
          0: (vi: Integer);
          1: (vf: TALFloat);
      end;
    al_LineInfo: array of TLineInfo;
    al_StringPool: TStringTree;

    al_Input: TALInStreamInfo;
    al_LoadNestingLevel: Integer; // nesting level during input

    al_TObjects: TSortedIntList;  // list of TLispObjects, TLispByteVectors
    al_TVectors: TSortedIntList;  // list of vectors (TList)
    al_CustomGCRoots: TList;
    al_CurrentCtx: TLispContextObject; // (GC root)
    al_ContextStack: TList; // GC root (TLispContextObjects)
    al_MainContextCell: Integer;
    al_MainContextObj: TLispContextObject;
    al_Frames: Integer;       // frame list (GC root), nilp -- no frames
      // frame is:
      // list of: (<func-name> . symbols)
    al_FluidList: TStringList;

    al_Nil: Integer;       // Not in List (or pool)
    al_CellStack: Integer; // global stack (GC root)
    al_FreeList: Integer;  // freelist

    al_CallDepth: Integer;

    // builtin symbol pointers (for various lookups/codegens/checks)
    // DO NOT CONCATENATE 'EM, OR PREPROCESSOR WILL DIE! %-)
    al_snil: Integer;
    al_sfalse: Integer;
    al_strue: Integer;
    al_ssymbols: Integer;
    al_sframes: Integer;
    al_slambda: Integer;
    al_smlambda: Integer;
    al_squote: Integer;
    al_sxquote: Integer;
    al_sfluid: Integer;
    al_sunfluid: Integer;
    al_sdefun: Integer;
    al_smacro: Integer;
    al_ssetq: Integer;

    al_opt_args: Integer;
      // first cell will never be changed by processor,
      // so we can make it GC root
    al_opt_psi: Integer;
    al_opt_pipe: Integer;  // pipe atom & pipe si
    al_opt_fluids: TStringList;

    al_GC_sl: packed array of Byte;// = nil;
    al_GC_ol: packed array of Byte;// = nil;
    al_GC_vl: packed array of Byte;// = nil;
    al_GC_touched_vectors0: TSortedIntList;// = nil;
    al_GC_touched_vectors1: TSortedIntList;// = nil;
{$IFDEF ALISP3_MULTIENGINE}
  end;
{$ENDIF}


{ TLispContextObject }
constructor TLispContextObject.Create ({$IFDEF ALISP3_MULTIENGINE}e: PXALISP3{$ENDIF});
begin
  {$IFDEF ALISP3_MULTIENGINE}
    _eng := e;
    //fSymbols := _eng.al_Nil;
  {$ELSE}
  {$ENDIF}
  fSymbols := al_Nil;
  fSymTree := TSymbolTree.Create();
end;

destructor TLispContextObject.Destroy ();
begin
  fSymbols := al_Nil;
  fSymTree.Free();
  {$IFDEF ALISP3_MULTIENGINE}
    _eng := nil;
  {$ENDIF}
end;


{$INCLUDE alisp3misc.pas}

function LispPrint (n: Integer; doQuote: Boolean): string; forward;
function ReadList (): Integer; forward;
function ReadVector (): Integer; forward;
//function ExtStr2Int (const s: string; out res: Integer): Boolean; forward;


function ALGetDefPathList (): string;
begin
  result := al_DefPathList;
end;

procedure ALSetDefPathList (const s: string);
begin
  al_DefPathList := s;
end;


function ALCar (n: Integer): Integer;
begin
  if (n = al_Nil) or (n < 0) or (n > High(al_CAR)) then result := al_Nil else result := al_CAR[n];
end;

function ALCdr (n: Integer): Integer;
begin
  if (n = al_Nil) or (n < 0) or (n > High(al_CDR)) then result := al_Nil else result := al_CDR[n].vi;
end;

procedure ALSetCar (n, v: Integer);
begin
  //if (v < 0) or (v > High(car)) or ((car[v] and MFLAG) <> 0) then v := nilp;
  if (n >= 0) and (n <= High(al_CAR)) and (n <> al_Nil) and ((al_CAR[n] and MFLAG) = 0) then al_CAR[n] := v;
end;

procedure ALSetCdr (n, v: Integer);
begin
  //if (v < 0) or (v > High(car)) or ((car[v] and MFLAG) <> 0) then v := nilp;
  if (n >= 0) and (n <= High(al_CDR)) and (n <> al_Nil) and ((al_CAR[n] and MFLAG) = 0) then al_CDR[n].vi := v;
end;

procedure ALGetLineInfo (n: Integer; out fn: string; out line, pos: Integer);
begin
  fn := ''; line := 0; pos := 0;
  if (n >= 0) and (n <= High(al_CDR)) and (n <> al_Nil) and ((al_CAR[n] and MFLAG) = 0) then
  begin
    fn := al_LineInfo[n].fileName;
    line := al_LineInfo[n].line;
    pos := al_LineInfo[n].pos;
  end;
end;

function ALGetAList (): Integer;
begin
  result := al_Frames;
end;

function ALGetSymbols (): Integer;
begin
  result := al_CurrentCtx.fSymbols;
end;

function ALIsNil (n: Integer): Boolean;
begin
  result := (n = al_Nil);
end;

function ALIsTrue (n: Integer): Boolean;
begin
  result := (n = al_strue);
end;

function ALIsFalse (n: Integer): Boolean;
begin
  result := (n = al_sfalse);
end;

function ALNil (): Integer;
begin
  result := al_Nil;
end;

function ALTrue (): Integer;
begin
  result := al_strue;
end;

function ALFalse (): Integer;
begin
  result := al_sfalse;
end;

function ALLambda (): Integer;
begin
  result := al_slambda;
end;

function ALMLambda (): Integer;
begin
  result := al_smlambda;
end;

function ALQuote (): Integer;
begin
  result := al_squote;
end;

function ALXQuote (): Integer;
begin
  result := al_sxquote;
end;

function ALSymbols (): Integer;
begin
//!!!
  result := al_CurrentCtx.fSymbols;
end;

function ALFrames (): Integer;
begin
  result := al_Frames;
end;

function ALGetBreakFlag (): Integer;
begin
  result := al_BreakFlag;
end;

procedure ALSetBreakFlag (new: Integer);
begin
  al_BreakFlag := new;
end;

procedure ALSetDebugger (dbgProc: TALDebuggerProc);
begin
  al_Debugger := dbgProc;
end;


function GetFuncName (n: Integer; const nameless: string = '(nameless)'): string;
var
  f: Integer;
begin
  if ALGetAtomType(n) <> DT_STRING then
  begin
    result := nameless;
    f := al_CurrentCtx.fSymbols;//!!!al_Symbols;
    while f <> al_Nil do
    begin
      if al_CDR[al_CAR[f]].vi = n then
      begin
        result := ALGetAtomStr(al_CAR[al_CAR[f]]);
        exit;
      end;
      f := al_CDR[f].vi;
    end;
  end
  else result := ALGetAtomStr(n);
end;

procedure ShowCallStack ();
// print function names on call stack
var
  f, t, n: Integer;
begin
  if al_Frames = al_Nil then exit;
  f := AL_CALL_TRACE_DEPTH; t := al_Frames;
  Write('>>>stack:');
  while (f > 0) and (t <> al_Nil) do
  begin
    n := al_CAR[t];
    Write(' ', GetFuncName(al_CAR[n]));
    t := al_CDR[t].vi;
    Dec(f);
  end;
  WriteLn;
end;

procedure ALDumpFrames (dumpLocalFn: Boolean = false);
var
  oa, nn: Integer;
  s: string;
begin
  oa := al_Frames;
  while al_Frames <> al_Nil do
  begin
    s := GetFuncName(al_CAR[al_CAR[al_Frames]], #0);
    if s = #0 then
    begin
      Write(' '); ALPrint(al_CAR[al_CAR[al_Frames]]); WriteLn;
    end
    else WriteLn(' ', s);
    nn := al_CDR[al_CAR[al_Frames]].vi;
    while nn <> al_Nil do
    begin
      Write('   ');
      if not dumpLocalFn and ((al_CAR[al_CDR[al_CAR[nn]].vi] = al_slambda) or
         (al_CAR[al_CDR[al_CAR[nn]].vi] = al_smlambda)) then
      begin
        ALPrint(al_CAR[al_CAR[nn]]); Write(' -- local func');
      end
      else ALPrint(al_CAR[nn]);
      WriteLn;
      nn := al_CDR[nn].vi;
    end;
    al_Frames := al_CDR[al_Frames].vi;
  end;
  al_Frames := oa;
end;

procedure ALClearState ();
begin
  al_Frames := al_Nil; al_CallDepth := 0; al_CellStack := al_Nil;
  al_LoadNestingLevel := 0;
  // restore main context
  ALLeaveContexts(al_ContextStack.Count);
  al_CurrentCtx := al_MainContextObj;
end;

function ALError (const m: string; n: Integer = -1): Integer;
// print error message M and set alBreakFlag.
// if N <> -1 print an additional line containing '>>> N'.
var
  fn: Integer;
begin
  result := al_Nil; if al_BreakFlag = ALBREAK_ERROR then begin ALClearState(); exit; end;

  Write('>>> error: [');
  if al_Input.fileName <> '' then Write(al_Input.fileName, ':');
  Write(al_Input.line, ']: ');
  if al_Frames <> al_Nil then
  begin
    fn := al_CAR[al_CAR[al_Frames]];
    Write(GetFuncName(fn));
  end
  else Write('(top level)');
  WriteLn(': ', m);
  if n <> -1 then begin Write('>>> '); ALPrint(n); WriteLn; end;
  if al_ShowCSFlag then ShowCallStack();
  {$IFDEF ALISP3_SHOW_FRAMES_ON_ERROR}ALDumpFrames();{$ENDIF}
  al_BreakFlag := ALBREAK_ERROR; ALClearState();
end;

function ALThrowError (const m: string; n: Integer = -1): Integer;
var
  s: string;
begin
  if al_BreakFlag = ALBREAK_ERROR then
  begin
    ALClearState(); result := al_Nil; exit;
  end;

  if al_ThrowSysError then
  begin
    if al_BreakFlag = ALBREAK_THROW then
    begin
      result := ALError('double THROW');
      exit;
    end;
    if n = -1 then n := al_Nil;
    // build list:
    // (n func filename linenum)
    // func can be "" -- nameless func
    result := ALCons(n, al_Nil); ALPushCell(result);
    if al_Frames <> al_Nil then
    begin
      s := GetFuncName(al_CAR[al_CAR[al_Frames]], '');
    end
    else s := '';
    al_CDR[result].vi := ALCons(ALNewStrAtom(s), al_Nil);
    result := al_CDR[result].vi;
    al_CDR[result].vi := ALCons(ALNewStrAtom(al_Input.fileName), al_Nil);
    result := al_CDR[result].vi;
    al_CDR[result].vi := ALCons(ALNewIntAtom(al_Input.line), al_Nil);
    n := ALNewStrAtom(m);
    result := ALPopCells(1);
    // now build pair: list . msg
    result := ALCons(result, n);
    al_BreakFlag := ALBREAK_THROW;
  end
  else result := ALError(m, n);
end;


function ALCountFree (): Integer;
// count the cells in the freeList
var
  n, k: Integer;
begin
  k := 0;
  n := al_FreeList;
  while n <> al_Nil do
  begin
    Inc(k);
    n := al_CDR[n].vi;
  end;
  result := k;
end;

procedure GCMarkCells (n: Integer); forward;

procedure GCMarkCells (n: Integer);
// mark cells which can be accessed through N.
// this routine uses the Deutsch/Schorr/Waite algorithm
// (aka pointer reversal algorithm) which marks the
// cells of a pool in constant space.
// in addition to the mark flag (MFLAG), it uses
// another bit to save the state of the current
// cell (S). the (otherwise unused) MFLAG of cdr is
// used to store the state.
// each visited cell goes through these states:
// 1) M=0 S=0 unvisited, process car
// 2) M=1 S=1 car visited, process cdr
// 3) M=1 S=0 completely visited, return to parent
var
  parent: Integer; // parent pointer
  f, t: Integer;
begin
  if (n < 0) or (n > High(al_CDR)) then exit;
  parent := al_Nil; // initially, there is no parent cell
  repeat
    // reached a leaf?
    if (n = al_Nil) or ((al_CAR[n] and MFLAG) <> 0) then
    begin
      // when the current cell is a leaf and there is
      // no parent, the entire tree is marked.
      if parent = al_Nil then exit;
      if (al_CDR[parent].vi and MFLAG) <> 0 then
      begin
        // state 2: the cdr of the parent has not yet been marked (M of cdr set).
        // swap car and cdr pointers and proceed with cdr. set state=3.
        t := al_CDR[parent].vi and not MFLAG;
        al_CDR[parent].vi := al_CAR[parent] and not MFLAG; // S=0
        al_CAR[parent] := n or MFLAG; // M=1
        n := t;
      end
      else
      begin
        // state 3: car and cdr of parent done.
        // return to the parent and restore parent of parent.
        t := parent; parent := al_CDR[t].vi; al_CDR[t].vi := n; n := t;
      end;
    end
    else
    begin
      // state 1: the current cell has not yet been visited.
      if (al_CAR[n] and AFLAG) <> 0 then
      begin
        // if the cell is an atom, mark the object as used & go to state 3.
        al_CAR[n] := al_CAR[n] or MFLAG; // M=1
        f := al_CDR[n].vi; {.$R+}
        case (al_CAR[n] and TYPE_MASK) of
          DT_STRING:
            if not al_StringPool.ValidHandle(f) then raise EALispError.Create('GC: orphan string')
            else if f > 0 then al_GC_sl[f] := 1;
          DT_VECTOR:
            if al_TVectors.Find(f, f) then
            begin
              if al_GC_vl[f] = 0 then
              begin
                // found untouched vector. remember it.
                al_GC_touched_vectors0.Add(al_TVectors[f]);
                al_GC_vl[f] := 1;
              end;
            end
            else raise EALispError.Create('GC: orphan vector');
          DT_BYTEVECTOR, DT_TOBJECT:
            begin
              if al_TObjects.Find(f, f) then
              begin
                if al_GC_ol[f] <> 1 then
                begin
                  al_GC_ol[f] := 1;
                  if TObject(al_TObjects[f]) is TLispContextObject then
                  begin
                    GCMarkCells(TLispContextObject(al_TObjects[f]).fSymbols);
                  end;
                end;
              end
              else raise EALispError.Create('GC: orphan object');
            end;
        end; {.$R-}
      end
      else
      begin
        // go to state 2: like above, but save the parent in car.
        t := al_CAR[n];
        al_CAR[n] := parent or MFLAG; // M=1
        parent := n; n := t;
        al_CDR[parent].vi := al_CDR[parent].vi or MFLAG; // S=1
      end;
    end;
  until false;
end;

{.$R+}
function GC (tc0: Integer = -1; tc1: Integer = -1): Integer;
// Mark and Sweep Garbage Collection.
// first, mark all "root cells" and then reclaim untagged cells
var
  f, r: Integer;
  tmpv: TSortedIntList;
  tl: TList;
  {$IFDEF ALISP3_VERBOSE_GC}
  t: LongWord;
  {$ENDIF}
begin
  {$IFDEF ALISP3_VERBOSE_GC}Write('*** GC started...'); t := GetTickCount();{$ENDIF}
  result := 0;

  // unmark freeList
  while al_FreeList <> al_Nil do
  begin
    al_CAR[al_FreeList] := AFLAG or DT_INTEGER;
    al_FreeList := al_CDR[al_FreeList].vi;
  end;

  // create bytemaps
  // gc_sl := nil; gc_ol := nil; gc_vl := nil;
  al_GC_touched_vectors0 := nil; al_GC_touched_vectors1 := nil;
  try
    if Length(al_GC_sl) < al_StringPool.maxHandle+1 then
    begin
      SetLength(al_GC_sl, al_StringPool.maxHandle+1);
    end;
    if Length(al_GC_sl) > 0 then FillChar(al_GC_sl[0], Length(al_GC_sl), 0);
    if Length(al_GC_ol) < al_TObjects.Count then
    begin
      SetLength(al_GC_ol, al_TObjects.Count);
    end;
    if Length(al_GC_ol) > 0 then FillChar(al_GC_ol[0], Length(al_GC_ol), 0);
    if Length(al_GC_vl) < al_TVectors.Count then
    begin
      SetLength(al_GC_vl, al_TVectors.Count);
    end;
    if Length(al_GC_vl) > 0 then FillChar(al_GC_vl[0], Length(al_GC_vl), 0);

    try
      al_GC_touched_vectors0 := TSortedIntList.Create();
      al_GC_touched_vectors1 := TSortedIntList.Create();
      al_GC_touched_vectors0.Capacity := al_TVectors.Count;

      al_CAR[al_Nil] := al_CAR[al_Nil] or MFLAG;
      GCMarkCells(al_CellStack);
      GCMarkCells(al_Frames);
      for f := 0 to al_ContextStack.Count-1 do
      begin
        GCMarkCells(TLispContextObject(al_ContextStack[f]).fSymbols);
        if al_TObjects.Find(Integer(al_ContextStack[f]), r) then
        begin
          al_GC_ol[r] := 1; // do not wipe this object
        end;
      end;
      GCMarkCells(al_CurrentCtx.fSymbols);
      GCMarkCells(al_MainContextObj.fSymbols);
      GCMarkCells(al_MainContextCell);
      if al_TObjects.Find(Integer(al_MainContextObj), r) then
      begin
        al_GC_ol[r] := 1; // do not wipe this object
      end;

      GCMarkCells(al_snil);
      GCMarkCells(al_sfalse);
      GCMarkCells(al_strue);
      GCMarkCells(al_ssymbols);
      GCMarkCells(al_sframes);
      GCMarkCells(al_slambda);
      GCMarkCells(al_smlambda);
      GCMarkCells(al_squote);
      GCMarkCells(al_sxquote);
      GCMarkCells(al_sfluid);
      GCMarkCells(al_sunfluid);
      GCMarkCells(al_sdefun);
      GCMarkCells(al_smacro);
      GCMarkCells(al_ssetq);

      GCMarkCells(tc0);
      GCMarkCells(tc1);

      f := al_CustomGCRoots.Count-1;
      while f >= 0 do
      begin
        GCMarkCells(Integer(al_CustomGCRoots[f]));
        Dec(f);
      end;

      // process fresh vectors
      while al_GC_touched_vectors0.Count > 0 do
      begin
        //{$IFDEF ALISP3_AGRESSIVE_GC}WriteLn('vector-loop.');{$ENDIF}
        // swap "touched" buffers
        // gc_touched_vectors0: just filled one
        tmpv := al_GC_touched_vectors0; al_GC_touched_vectors0 := al_GC_touched_vectors1;
        al_GC_touched_vectors1 := tmpv;
        al_GC_touched_vectors0.Clear();
        al_GC_touched_vectors0.Capacity := al_TVectors.Count;
        // mark all touched vector's items
        f := al_GC_touched_vectors1.Count;
        while f > 0 do
        begin
          Dec(f); tl := TList(al_GC_touched_vectors1[f]);
          r := tl.Count;
          while r > 0 do
          begin
            Dec(r);
            if (al_CAR[Integer(tl[r])] and MFLAG) = 0 then
              GCMarkCells(Integer(tl[r]));
          end;
        end;
        // do everything again
      end;
    finally
      al_GC_touched_vectors0.Free(); al_GC_touched_vectors1.Free();
      al_GC_touched_vectors0 := nil; al_GC_touched_vectors1 := nil;
    end;

    // now we can wipe out all unised thingy
    //
    // wipe out unused vectors
    f := al_TVectors.Count-1; while f >= 0 do
    begin
      if al_GC_vl[f] = 0 then
      begin
        //{$IFDEF ALISP3_AGRESSIVE_GC}WriteLn('dead vector.');{$ENDIF}
        TObject(al_TVectors[f]).Free();
        al_TVectors.Delete(f);
      end;
      Dec(f);
    end;

    // wipe out unused objects
    f := al_TObjects.Count-1; while f >= 0 do
    begin
      if (al_GC_ol[f] = 0) {and (al_TObjects[f] <> Integer(al_MainContextObj))} then
      begin
        //{$IFDEF ALISP3_AGRESSIVE_GC}WriteLn('dead object.');{$ENDIF}
        TObject(al_TObjects[f]).Free();
        al_TObjects.Delete(f);
      end;
      Dec(f);
    end;

    // wipe out unused strings
    f := al_StringPool.MaxHandle; while f > 0  do
    begin
      if al_GC_sl[f] = 0 then
      begin
        if al_StringPool.ValidHandle(f) then al_StringPool.Delete(f);
      end;
      Dec(f);
    end;

    // all used cells marked, build freeList
    al_FreeList := al_Nil;
    for f := High(al_CAR) downto 0 do
    begin
      if (al_CAR[f] and MFLAG) = 0 then
      begin
        // free cell
        al_CAR[f] := MFLAG; al_CDR[f].vi := al_FreeList;
        al_FreeList := f;
        Inc(result);
        {$IFDEF ALISP3_DEBUG}{$R+}{$ENDIF}
        al_LineInfo[f].fileName := '';
        al_LineInfo[f].line := 0;
        al_LineInfo[f].pos := 0;
        {$IFDEF ALISP3_DEBUG}{$R-}{$ENDIF}
      end
      else al_CAR[f] := al_CAR[f] and not MFLAG;
    end;

    {$IFDEF ALISP3_VERBOSE_GC}
    t := GetTickCount()-t;
    WriteLn(result, ' of ', Length(al_CAR), ' cells free (', t, ' ms)');
    {$ENDIF}
  finally
    //SetLength(gc_sl, 0); SetLength(gc_ol, 0); SetLength(gc_vl, 0);
  end;
end;
{.$R-}

function ALGC (): Integer;
begin
  result := GC(-1, -1);
end;

procedure ALRegisterGCRoot (n: Integer);
begin
  al_CustomGCRoots.Add(Pointer(n));
end;

procedure ALUnregisterGCRoot (n: Integer);
begin
  n := al_CustomGCRoots.IndexOf(Pointer(n));
  if n <> -1 then al_CustomGCRoots.Delete(n);
end;


function ALCons (pcar, pcdr: Integer): Integer;
// allocate a fresh cell and initialize with PCAR,PCDR
// PCAR & PCDR will not be recycled during ALCons()'ing
var
  n, c: Integer;
begin
  {$IFDEF ALISP3_AGRESSIVE_GC}GC(pcar, pcdr);{$ENDIF}
  if al_FreeList = al_Nil then
  begin
    {$IFDEF ALISP3_GC_FIRST}
      c := GC(pcar, pcdr);
      if (al_FreeList = al_Nil) or (c < AL_MIN_POOL_LEFT) then
      begin
        if Length(al_CAR) >= AL_MAX_POOL_SIZE then
        begin
          if al_FreeList = al_Nil then raise EALispOutOfMemoryError.Create('ALISP3: out of cells');
        end;
        n := Length(al_CAR); if n = 1 then n := 0;
        SetLength(al_CAR, n+AL_POOL_GROW_DELTA);
        SetLength(al_CDR, n+AL_POOL_GROW_DELTA);
        SetLength(al_LineInfo, n+AL_POOL_GROW_DELTA);
        c := High(al_CAR); if n = 0 then Inc(n);
        while c >= n do
        begin
          al_CAR[c] := MFLAG; al_CDR[c].vi := al_FreeList;
          al_FreeList := c;
          {$IFDEF ALISP3_DEBUG}{$R+}{$ENDIF}
          al_LineInfo[c].fileName := '';
          al_LineInfo[c].line := 0;
          al_LineInfo[c].pos := 0;
          {$IFDEF ALISP3_DEBUG}{$R-}{$ENDIF}
          Dec(c);
        end;
      end;
    {$ELSE}
      if Length(al_CAR) >= AL_MAX_POOL_SIZE then GC(pcar, pcdr)
      else
      begin
        n := Length(al_CAR); if n = 1 then n := 0;
        SetLength(al_CAR, n+AL_POOL_GROW_DELTA);
        SetLength(al_CDR, n+AL_POOL_GROW_DELTA);
        SetLength(al_LineInfo, n+AL_POOL_GROW_DELTA);
        c := High(al_CAR); if n = 0 then Inc(n);
        while c >= n do
        begin
          al_CAR[c] := MFLAG; al_CDR[c].vi := al_FreeList;
          al_FreeList := c;
          {$IFDEF ALISP3_DEBUG}{$R+}{$ENDIF}
          al_LineInfo[c].fileName := '';
          al_LineInfo[c].line := 0;
          al_LineInfo[c].pos := 0;
          {$IFDEF ALISP3_DEBUG}{$R-}{$ENDIF}
          Dec(c);
        end;
      end;
      if al_FreeList = al_Nil then raise EALispOutOfMemoryError.Create('ALISP3: out of cells');
    {$ENDIF}
  end;
  n := al_FreeList;
  al_FreeList := al_CDR[al_FreeList].vi;
  al_CAR[n] := pcar; al_CDR[n].vi := pcdr;
  result := n;
  {if pcdr = nilp then
  begin
    if pcar <> nilp then lineInfo[n] := lineInfo[pcar];
  end
  else lineInfo[n] := lineInfo[pcdr];}
end;


function ALIsAtom (n: Integer): Boolean;
// is N a 'real' (non-NIL) atom?
begin
  result := (n >= 0) and (n <= High(al_CAR)) and
    (n <> al_Nil) and ((al_CAR[n] and AFLAG) <> 0);
end;

function ALIsContext (n: Integer): Boolean;
begin
  result := (n >= 0) and (n <= High(al_CAR)) and (n <> al_Nil) and
    ((al_CAR[n] and AFLAG) <> 0) and
    ((al_CAR[n] and TYPE_MASK) = DT_TOBJECT) and
    (TObject(al_CDR[n].vi) is TLispContextObject);
end;

function ALNewIntAtom (value: Integer): Integer;
begin
  result := ALCons(al_Nil, al_Nil);
  al_CAR[result] := DT_INTEGER or AFLAG;
  al_CDR[result].vi := value;
end;

function ALNewFloatAtom (value: TALFloat): Integer;
begin
  result := ALCons(al_Nil, al_Nil);
  al_CAR[result] := DT_FLOAT or AFLAG;
  al_CDR[result].vf := value;
end;

function ALNewObjectAtom (value: TObject): Integer;
{var
  f: Integer;}
begin
  result := ALCons(al_Nil, al_Nil);
  {f := tObjects.IndexOf(Integer(value)); if f = -1 then }
    al_TObjects.Add(Integer(value));
  al_CAR[result] := DT_TOBJECT or AFLAG;
  al_CDR[result].vi := Integer(value);
end;

function ALNewContextAtom (): Integer;
var
  ctx: TLispContextObject;
begin
  result := ALCons(al_Nil, al_Nil);
  ALPushCell(result);
  ctx := TLispContextObject.Create({$IFDEF ALISP3_MULTIENGINE}_eng{$ENDIF});
  ALPopCells(1);
  al_TObjects.Add(Integer(ctx));
  al_CAR[result] := DT_TOBJECT or AFLAG;
  al_CDR[result].vi := Integer(ctx);
end;

function ALNewStrAtom (const s: string): Integer;
var
  si: Integer;
begin
  result := ALCons(al_Nil, al_Nil);
  si := al_StringPool.Append(s);
  al_CAR[result] := AFLAG or DT_STRING;
  al_CDR[result].vi := si;
end;

function ALNewVector (size: Integer): Integer;
var
  lst: TList;
begin
  result := ALCons(al_Nil, al_Nil);
  lst := TList.Create();
  try
    if size > 0 then lst.Capacity := size;
    while size > 0 do begin lst.Add(Pointer(al_Nil)); Dec(size); end;
    al_TVectors.Add(Integer(lst));
  except
    lst.Free();
    raise;
  end;
  al_CAR[result] := DT_VECTOR or AFLAG;
  al_CDR[result].vi := Integer(lst);
end;

function ALNewByteVector (size: Integer): Integer;
var
  lst: TLispByteVector;
begin
  result := ALCons(al_Nil, al_Nil);
  if size < 0 then size := 0;
  lst := TLispByteVector.Create();
  try
    SetLength(lst.vector, size);
    while size > 0 do begin Dec(size); lst.vector[size] := 0; end;
    al_TObjects.Add(Integer(lst));
  except
    lst.Free();
    raise;
  end;
  al_CAR[result] := DT_BYTEVECTOR or AFLAG;
  al_CDR[result].vi := Integer(lst);
end;

function ALNewVectorFromList (nlst: Integer): Integer;
var
  lst: TList;
begin
  ALPushCell(nlst);
  result := ALCons(al_Nil, al_Nil);
  ALPopCells(1);
  lst := TList.Create();
  try
    while (nlst >= 0) and (nlst <= High(al_CAR)) and (nlst <> al_Nil) do
    begin
      lst.Add(Pointer(al_CAR[nlst])); nlst := al_CDR[nlst].vi;
    end;
    al_TVectors.Add(Integer(lst));
  except
    lst.Free();
    raise;
  end;
  al_CAR[result] := DT_VECTOR or AFLAG;
  al_CDR[result].vi := Integer(lst);
end;

function ALGetAtomType (n: Integer): Integer;
// returns DT_XXX or -1 if not an atom
begin
  if not ALIsAtom(n) then result := -1 else result := al_CAR[n] and TYPE_MASK;
end;

function ALGetAtomStr (n: Integer): string;
// returns string value or '' if not an atom
var
  fn: TALFloat;
  f: Integer;
begin
  if not ALIsAtom(n) then result := ''
  else
  begin
    case al_CAR[n] and TYPE_MASK of
      DT_INTEGER: result := IntToStr(al_CDR[n].vi);
      DT_FLOAT:
        begin
          fn := al_CDR[n].vf;
          Str(fn:17:8, result);
          //Str(fn:50:24, result);
          result := Trim(result);
          f := Pos('.', result);
          if f > 0 then
          begin
            f := Length(result);
            while result[f] = '0' do Dec(f);
            if result[f] = '.' then Inc(f);
            Delete(result, f+1, Length(result));
          end;
        end;
      DT_STRING: result := al_StringPool[al_CDR[n].vi];
      DT_PRIMITIVE: result := '@'+PALPrimitive(al_CDR[n].vi).name;
      DT_VECTOR: result := '<v:'+IntToHex(LongWord(al_CDR[n].vi), 8)+'>';
      DT_BYTEVECTOR: result := '<bv:'+IntToHex(LongWord(al_CDR[n].vi), 8)+'>';
      DT_TOBJECT: result := '<o:'+IntToHex(LongWord(al_CDR[n].vi), 8)+'>';
      else result := '<???>';
    end;
  end;
end;

function ALGetAtomInt (n: Integer): Integer;
// returns int value or 0 if not a number
begin
  if not ALIsAtom(n) then result := 0
  else
  begin
    case al_CAR[n] and TYPE_MASK of
      DT_INTEGER: result := al_CDR[n].vi;
      DT_FLOAT:  result := round(al_CDR[n].vf);
      DT_STRING: result := StrToIntDef(al_StringPool[al_CDR[n].vi], 0);
      else result := 0;
    end;
  end;
end;

function ALGetAtomObject (n: Integer): TObject;
begin
  if not ALIsAtom(n) then result := nil
  else
  begin
    case al_CAR[n] and TYPE_MASK of
      DT_VECTOR, DT_BYTEVECTOR, DT_TOBJECT: result := TObject(al_CDR[n].vi);
      else result := nil;
    end;
  end;
end;

function ALGetAtomFloat (n: Integer): TALFloat;
// returns float value or 0 if not a number
var
  e: Integer;
begin
  if not ALIsAtom(n) then result := 0.0
  else
  begin
    case al_CAR[n] and TYPE_MASK of
      DT_INTEGER: result := al_CDR[n].vi;
      DT_FLOAT:  result := al_CDR[n].vf;
      DT_STRING:
        begin
          Val(al_StringPool[al_CDR[n].vi], result, e);
          if e <> 0 then Val(al_StringPool[al_CDR[n].vi], result, e);
          if e <> 0 then result := 0.0;
        end;
      else result := 0.0;
    end;
  end;
end;

function ALGetVectorItem (n, index: Integer): Integer;
// for normal and byte vectors
var
  tl: TList;
  tv: TLispByteVector;
begin
  result := al_Nil;
  if (index < 0) or (n < 0) or (n > High(al_CAR)) or ((al_CAR[n] and AFLAG) = 0) then exit;
  case (al_CAR[n] and TYPE_MASK) of
    DT_VECTOR:
      begin
        tl := TList(al_CDR[n].vi);
        if index < tl.Count then result := Integer(tl[index]);
      end;
    DT_BYTEVECTOR:
      begin
        tv := TLispByteVector(al_CDR[n].vi);
        if index < Length(tv.vector) then result := tv.vector[index];
      end;
  end;
end;

function ALGetByteVectorDataPtr (n: Integer): Pointer;
var
  tv: TLispByteVector;
begin
  result := nil;
  if (n < 0) or (n > High(al_CAR)) or ((al_CAR[n] and AFLAG) = 0) then exit;
  if (al_CAR[n] and TYPE_MASK) = DT_BYTEVECTOR then
  begin
    tv := TLispByteVector(al_CDR[n].vi);
    if Length(tv.vector) > 0 then result := @(tv.vector[0]);
  end;
end;

function ALGetVectorLength (n: Integer): Integer;
// for normal and byte vectors
var
  tl: TList;
  tv: TLispByteVector;
begin
  result := 0;
  if (n < 0) or (n > High(al_CAR)) or ((al_CAR[n] and AFLAG) = 0) then exit;
  case (al_CAR[n] and TYPE_MASK) of
    DT_VECTOR:
      begin
        tl := TList(al_CDR[n].vi);
        result := tl.Count;
      end;
    DT_BYTEVECTOR:
      begin
        tv := TLispByteVector(al_CDR[n].vi);
        result := Length(tv.vector);
      end;
  end;
end;

procedure ALSetVectorItem (n, index, value: Integer);
// for normal and byte vectors
var
  tl: TList;
  tv: TLispByteVector;
begin
  if (index < 0) or (n < 0) or (n > High(al_CAR)) or ((al_CAR[n] and AFLAG) = 0) then exit;
  case (al_CAR[n] and TYPE_MASK) of
    DT_VECTOR:
      begin
        if (value < 0) or (value > High(al_CAR)) or ((al_CAR[value] and MFLAG) <> 0) then value := al_Nil;
        tl := TList(al_CDR[n].vi);
        if index < tl.Count then tl[index] := Pointer(value);
      end;
    DT_BYTEVECTOR:
      begin
        tv := TLispByteVector(al_CDR[n].vi);
        if index < Length(tv.vector) then tv.vector[index] := (value and $FF);
      end;
  end;
end;

procedure ALSetVectorLength (n, newLen: Integer);
// for normal and byte vectors
var
  tl: TList;
  tv: TLispByteVector;
begin
  if newLen < 0 then newLen := 0;
  if (n < 0) or (n > High(al_CAR)) or ((al_CAR[n] and AFLAG) = 0) then exit;
  case (al_CAR[n] and TYPE_MASK) of
    DT_VECTOR:
      begin
        tl := TList(al_CDR[n].vi);
        if newLen = 0 then tl.Clear()
        else
        begin
          if tl.Count < newLen then
          begin
            tl.Capacity := newLen;
            while tl.Count < newLen do tl.Add(Pointer(al_Nil));
          end
          else while tl.Count > newLen do tl.Delete(tl.Count-1);
        end;
      end;
    DT_BYTEVECTOR:
      begin
        tv := TLispByteVector(al_CDR[n].vi);
        n := Length(tv.vector); SetLength(tv.vector, newLen);
        while n < newLen do begin tv.vector[n] := 0; Inc(n); end;
      end;
  end;
end;


function ALEnterContext (ctxatom: Integer): Boolean;
begin
  result := false;
  if not ALIsContext(ctxatom) then exit;
  al_ContextStack.Add(Pointer(al_CurrentCtx));
  al_CurrentCtx := TLispContextObject(al_CDR[ctxatom].vi);
  result := true;
end;

procedure ALLeaveContexts (cnt: Integer);
var
  res: Integer;
begin
  if cnt < 1 then exit;
  res := al_ContextStack.Count;
  if cnt >= res then
  begin
    al_CurrentCtx := al_MainContextObj;
    al_ContextStack.Clear();
    //cnt := res;
    exit;
  end
  else al_CurrentCtx := TLispContextObject(al_ContextStack[res-cnt]);
  while cnt > 0 do
  begin
    Dec(res); al_ContextStack.Delete(res);
    Dec(cnt);
  end;
end;

function ALContextStackDepth (): Integer;
begin
  result := al_ContextStack.Count;
end;

function ALGetMainContext (): Integer;
// returns context atom
begin
  result := al_MainContextCell;
end;

function ALGetStringFromPool (si: Integer): string;
begin
  if al_StringPool.ValidHandle(si) then result := al_StringPool[si]
  else result := '';
end;

procedure ALEnumContextSymbols (ctx: Integer; CBack: TALListSymbolsCallback; user: Pointer);
// ctx: context atom
var
  n, nn: Integer;
begin
  if not ALIsContext(ctx) then exit;
  n := TLispContextObject(al_CDR[ctx].vi).fSymbols;
  while n <> al_Nil do
  begin
    nn := al_CAR[n]; // symbol pair
    if not CBack(al_StringPool[al_CDR[al_CAR[nn]].vi], al_CDR[nn].vi, user) then exit;
    n := al_CDR[n].vi;
  end;
end;

procedure ALEnumSymbolsInCurrentContext (CBack: TALListSymbolsCallback; user: Pointer);
var
  n, nn: Integer;
begin
  n := al_CurrentCtx.fSymbols;
  while n <> al_Nil do
  begin
    nn := al_CAR[n]; // symbol pair
    if not CBack(al_StringPool[al_CDR[al_CAR[nn]].vi], al_CDR[nn].vi, user) then exit;
    n := al_CDR[n].vi;
  end;
end;

function FindSymbolInList (const s: string; symList: Integer;
  si: Integer = -1): Integer;
// if si was not specified -- determine it
begin
  if si = -1 then si := al_StringPool.Find(s);
  if al_StringPool.ValidHandle(si) then
  begin
    while (symList >= 0) and (symList <= High(al_CAR)) and (symList <> al_Nil) and
          ((al_CAR[symList] and AFLAG) = 0) do
    begin
      result := al_CAR[al_CAR[symList]];
      if ((al_CAR[result] and AFLAG) = 0) or ((al_CAR[result] and TYPE_MASK) <> DT_STRING) then
      begin
        raise EALispError.Create('shit! '+ IntToStr(al_CAR[result] and TYPE_MASK));
      end;
      if al_CDR[result].vi = si then begin result := al_CAR[symList]; exit; end;
      symList := al_CDR[symList].vi;
    end;
  end;
  result := al_Nil;
end;

function FindGlobalSymbol (const s: string; si: Integer = -1; onlyTop: Boolean = false): Integer;
// find a symbol named S in the context stack
var
  stk: Integer;
  ctx: TLispContextObject;
begin
  if si = -1 then si := al_StringPool.Find(s);
  if al_StringPool.ValidHandle(si) then
  begin
    ctx := al_CurrentCtx;
    result := ctx.fSymTree.Find(si);
    if ctx.fSymTree.ValidHandle(result) then
    begin
      result := ctx.fSymTree.Cells[result];
      exit;
    end;
    if onlyTop then begin result := al_Nil; exit; end;
    stk := al_ContextStack.Count;
    while stk > 0 do
    begin
      Dec(stk);
      ctx := TLispContextObject(al_ContextStack[stk]);
      result := ctx.fSymTree.Find(si);
      if ctx.fSymTree.ValidHandle(result) then
      begin
        result := ctx.fSymTree.Cells[result];
        exit;
      end;
    end;
    result := al_Nil;
  end
  else result := al_Nil;
end;

function FindSymbol (const s: string; si: Integer = -1;
  onlyTopFrame: Boolean = false; viewGlobals: Boolean = false): Integer;
// find a symbol named S in the symbol table
// (starting from local frames)
// if si was not specified -- determine it
// does not processes "::"
var
  lst: Integer;
begin
  if si = -1 then si := al_StringPool.Find(s);
  if al_StringPool.ValidHandle(si) then
  begin
    lst := al_Frames;
    while lst <> al_Nil do
    begin
      result := FindSymbolInList(s, al_CDR[al_CAR[lst]].vi, si);
      if onlyTopFrame or (result <> al_Nil) then
      begin
        if (result = al_Nil) and onlyTopFrame and viewGlobals then
          result := FindGlobalSymbol(s, si);
        exit;
      end;
      lst := al_CDR[lst].vi;
    end;
    if onlyTopFrame then
    begin
      if viewGlobals then result := FindGlobalSymbol(s, si)
      else result := al_Nil;
    end
    else result := FindGlobalSymbol(s, si);
    //if result = al_Nil then result := FindGlobalSymbol(s, si, onlyTopFrame);
  end
  else result := al_Nil;
end;

function DoSetSymbol (const s: string; v: Integer; si: Integer = -1;
  onlyTopFrame: Boolean = false; viewGlobals: Boolean = false): Integer;
var
  valid: Boolean;
begin
  if si = -1 then si := al_StringPool.Find(s);
  valid := al_StringPool.ValidHandle(si);
  if valid then result := FindSymbol(s, si, onlyTopFrame, viewGlobals) else result := al_Nil;
  if result = al_Nil then
  begin
    // construct new symbol:
    // construct string atom
    // I mimic ALNewStrAtom() to speedup the things (is there any speedup at all?)
    if v = -1 then result := ALCons(al_Nil, al_Nil) else result := ALCons(v, al_Nil);
    al_CAR[result] := AFLAG or DT_STRING;
    if not valid then si := al_StringPool.Append(s);
    al_CDR[result].vi := si;
    // construct symbol pair
    if v = -1 then result := ALCons(result, result)
    else result := ALCons(result, v);

    if al_Frames <> al_Nil then
    begin
      // add to local frame
      si := al_CDR[al_CAR[al_Frames]].vi; // local symbols
      // add new local symbol
      al_CDR[al_CAR[al_Frames]].vi := ALCons(result, si);
    end
    else
    begin
      // add to current context
      al_CurrentCtx.fSymTree.Append(si, result);
      al_CurrentCtx.fSymbols := ALCons(result, al_CurrentCtx.fSymbols);
    end;
  end
  else
  begin
    // the symbol aready exists. check protection
    if (al_CAR[al_CAR[result]] and PROTFLAG) = 0 then
    begin
      // the symbol is not protected, change it's value
      if v = -1 then al_CDR[result].vi := al_CAR[result] else al_CDR[result].vi := v;
    end;
  end;
end;

function ALSetSymbol (const s: string; v: Integer = -1): Integer;
begin
  result := DoSetSymbol(s, v);
end;

function SymContextImmerse (vn: string; dpos: Integer; out pcnt: Integer;
  onlyTop: Boolean = false): Integer;
// returns symbol pair or nil
var
  cn: string;
  f, si: Integer;
begin
  {$IFDEF ALISP3_VERBOSE_IMMERSION}WriteLn('*>immersion for "', vn, '"');{$ENDIF}
  pcnt := 0; result := al_Nil;
  // process context dereferences
  if dpos = 0 then f := Pos('::', vn) else f := dpos;
  if f = 0 then
  begin
    //result := FindGlobalSymbol(vn);
    result := FindSymbol(vn);
    exit;
  end;
  repeat
    // split string
    cn := Copy(vn, 1, f-1); vn := Copy(vn, f+2, Length(vn));
    // empty "::" means current context. just skip "::"
    if f <> 1 then
    begin
      // find symbol <cn>
      if onlyTop then
      begin
        si := al_StringPool.Find(cn);
        if al_StringPool.ValidHandle(si) then
        begin
          result := al_CurrentCtx.fSymTree.Find(si);
          if al_CurrentCtx.fSymTree.ValidHandle(result) then
          begin
            result := al_CurrentCtx.fSymTree.Cells[result];
          end
          else result := al_Nil;
        end
        else result := al_Nil;
      end
      else
      begin
        result := FindSymbol(cn);
        onlyTop := true;
      end;
      //if result = al_Nil then break; // no such symbol? quit
      if not ALEnterContext(al_CDR[result].vi) then
      begin
        // not a context (ALEnterContext() failed)? quit
        result := al_Nil;
        break;
      end;
      Inc(pcnt);
      {$IFDEF ALISP3_VERBOSE_IMMERSION}WriteLn('>context "', cn, '"');{$ENDIF}
    end
    else onlyTop := true;
    if vn = '' then begin result := al_Nil; break; end; // quit when no id
    f := Pos('::', vn);
    if f = 0 then
    begin
      // no more immersions
      {$IFDEF ALISP3_VERBOSE_IMMERSION}WriteLn('>var "', vn, '"');{$ENDIF}
      //result := FindGlobalSymbol(vn);
      //if vn = '*symbols*' then result := al_ssymbols
      //else if vn = '*nil*' then result := al_snil
      //else if vn = '*t*' then result := al_strue
      //else if vn = '*f*' then result := al_sfalse
      //else
      begin
        si := al_StringPool.Find(vn);
        if al_StringPool.ValidHandle(si) then
        begin
          result := al_CurrentCtx.fSymTree.Find(si);
          if al_CurrentCtx.fSymTree.ValidHandle(result) then
          begin
            result := al_CurrentCtx.fSymTree.Cells[result];
          end
          else result := al_Nil;
        end
        else result := al_Nil;
      end;
      {$IFDEF ALISP3_VERBOSE_IMMERSION}WriteLn('>var-pair ', result);{$ENDIF}
      break;
    end;
  until false;
  {$IFDEF ALISP3_VERBOSE_IMMERSION}WriteLn('>', al_ContextStack.Count, '; ', pcnt);{$ENDIF}
end;

function ALSetSymbolEx (const s: string; v: Integer = -1; si: Integer = -1;
  onlyTopFrame: Boolean = false; onlyTopCtx: Boolean = false;
  viewGlobals: Boolean = false): Integer;
// processes context dereferences
var
  f, pcnt: Integer;
  valid: Boolean;
  t: string;
begin
  if si = -1 then si := al_StringPool.Find(s);
  valid := false;
  if not al_StringPool.ValidHandle(si) then si := al_StringPool.Append(s)
  else valid := true;

  f := al_StringPool.DColonPos[si];
  if f <> 0 then
  begin
    result := SymContextImmerse(al_StringPool[si], f, pcnt);
    if result = al_Nil then
    begin
      t := Copy(al_StringPool[si], f+2, maxInt-100);
      f := 1; while f < Length(t) do
      begin
        if (t[f] = ':') and (t[f+1] = ':') then
        begin
          Delete(t, 1, f+1);
          f := 1;
        end
        else Inc(f);
      end;
      result := ALSetGlobalSymbol(t, v, onlyTopCtx);
    end
    else
    begin
      if v = -1 then al_CDR[result].vi := al_CAR[result] else al_CDR[result].vi := v;
    end;
    ALLeaveContexts(pcnt);
  end
  else
  begin
    if valid then result := DoSetSymbol(s, v, si, onlyTopFrame, viewGlobals)
    else
    begin
      //al_StringPool.Delete(si);
      result := DoSetSymbol(s, v, -1, onlyTopFrame);
    end;
  end;
end;

function ALFindSymbolInList (const s: string; symList: Integer): Integer;
begin
  result := FindSymbolInList(s, symList);
end;

function ALFindSymbol (const s: string): Integer;
// find a symbol named S in the symbol table
// (starting from local frames)
begin
  result := FindSymbol(s);
end;

function ALFindSymbolEx (const s: string; si: Integer = -1): Integer;
// find a symbol named S in the symbol table
// (starting from local frames)
// processes context dereferences
var
  f, pcnt: Integer;
begin
  if si = -1 then si := al_StringPool.Find(s);
  if not al_StringPool.ValidHandle(si) then si := al_StringPool.Append(s);

  f := al_StringPool.DColonPos[si];
  if f <> 0 then
  begin
    result := SymContextImmerse(al_StringPool[si], f, pcnt);
    ALLeaveContexts(pcnt);
  end
  else
  begin
    result := FindSymbol(s, si);
  end;
end;

function ALFindGlobalSymbol (const s: string): Integer;
// find a symbol named S in the context stack
begin
  result := FindGlobalSymbol(s);
end;

function ALSetGlobalSymbol (const s: string; v: Integer = -1;
  onlyTopFrame: Boolean = false): Integer;
var
  si: Integer;
  valid: Boolean;
begin
  //WriteLn('>>global set: "', s, '"');
  si := al_StringPool.Find(s);
  valid := al_StringPool.ValidHandle(si);
  if valid then
  begin
    if onlyTopFrame then
    begin
      result := al_CurrentCtx.fSymTree.Find(si);
      if al_CurrentCtx.fSymTree.ValidHandle(result) then
      begin
        result := al_CurrentCtx.fSymTree.Cells[result];
      end
      else result := al_Nil;
      //WriteLn('>>global set: "', s, '": ', result);
      //if al_CurrentCtx = al_MainContextObj then WriteLn('>>main context');
    end
    else result := FindGlobalSymbol(s, si);
  end
  else result := al_Nil;

  if result = al_Nil then
  begin
    // construct new symbol:
    // construct string atom
    // I mimic ALNewStrAtom() to speedup the things (is there any speedup at all?)
    if v = -1 then result := ALCons(al_Nil, al_Nil) else result := ALCons(v, al_Nil);
    al_CAR[result] := AFLAG or DT_STRING;
    if not valid then si := al_StringPool.Append(s);
    al_CDR[result].vi := si;
    // construct symbol pair
    if v = -1 then result := ALCons(result, result)
    else result := ALCons(result, v);

    // add to current context
    al_CurrentCtx.fSymTree.Append(si, result);
    al_CurrentCtx.fSymbols := ALCons(result, al_CurrentCtx.fSymbols);
  end
  else
  begin
    // the symbol aready exists. check protection
    if (al_CAR[al_CAR[result]] and PROTFLAG) = 0 then
    begin
      // the symbol is not protected, change it's value
      if v = -1 then al_CDR[result].vi := al_CAR[result] else al_CDR[result].vi := v;
    end;
  end;
end;

procedure ALDeleteSymbolFromCurrentContext (const s: string);
var
  n, prev, nn, si: Integer;
begin
  si := al_StringPool.Find(s);
  if not al_StringPool.ValidHandle(si) then exit; // no such symbol

  n := al_CurrentCtx.fSymbols; prev := al_Nil;
  while n <> al_Nil do
  begin
    nn := al_CAR[n]; // symbol pair
    if al_CDR[al_CAR[nn]].vi = si then
    begin
      // symbol found. kill it
      if prev = al_Nil then
      begin
        al_CurrentCtx.fSymbols := al_CDR[al_CurrentCtx.fSymbols].vi; // first symbol
      end
      else
      begin
        al_CDR[prev].vi := al_CDR[n].vi;
      end;
      al_CurrentCtx.fSymTree.Delete(al_CurrentCtx.fSymTree.Find(si));
      exit;
    end;
    prev := n;
    n := al_CDR[n].vi;
  end;
end;

procedure ALSetSymbolProtectionFlag (n: Integer; value: Boolean);
begin
  if (n >= 0) and (n <= High(al_CAR)) and (n <> al_Nil) and
     ((al_CAR[n] and MFLAG) = 0) and ((al_CAR[n] and AFLAG) <> 0) and
     ((al_CAR[n] and TYPE_MASK) = DT_STRING) then
  begin
    if value then al_CAR[n] := al_CAR[n] or PROTFLAG
    else al_CAR[n] := al_CAR[n] and (not PROTFLAG);
  end;
end;

function ALGetSymbolProtectionFlag (n: Integer): Boolean;
begin
  if (n >= 0) and (n <= High(al_CAR)) and (n <> al_Nil) and
     ((al_CAR[n] and MFLAG) = 0) and ((al_CAR[n] and AFLAG) <> 0) and
     ((al_CAR[n] and TYPE_MASK) = DT_STRING) then result := false
  else result := (al_CAR[n] and PROTFLAG) <> 0;
end;

procedure ALPushCell (n: Integer);
// save the cell N on the stack
begin
  al_CellStack := ALCons(n, al_CellStack);
end;

function ALPopCells (k: Integer): Integer;
// pop K cells off the stack and return
// the most recently popped one.
begin
  result := -1;
  while k > 0 do
  begin
    if al_CellStack = al_Nil then begin result := ALError('ALISP3: stack underflow'); exit; end;
    result := al_CAR[al_CellStack];
    al_CellStack := al_CDR[al_CellStack].vi;
    Dec(k);
  end;
end;

procedure ALEnter (n: Integer);
// enter lambda. N is a function name (atom), [M]LAMBDA or NIL
// set up new frame for local vars
begin
  if al_BreakFlag = ALBREAK_ERROR then exit;
  al_Frames := ALCons(ALCons(n, al_Nil), al_Frames);  // function . symlist
  Inc(al_CallDepth);
  {$IFDEF ALISP3_VERBOSE_ENTER_LEAVE}
  WriteLn('!>>ENTER (', al_CallDepth, '): '); ALDumpFrames(true);//ALPrint(eng alist); WriteLn;
  {$ENDIF}
end;

procedure ALLeave ();
// leave lambda, restore previous frame
begin
  if al_BreakFlag = ALBREAK_ERROR then exit;
  if al_CallDepth < 1 then
  begin
    ALError('ALISP3: unbalanced calls (internal)');
    exit;
  end;

  {$IFDEF ALISP3_VERBOSE_ENTER_LEAVE}
  WriteLn('!>>LEAVE (', al_CallDepth, '): '); ALDumpFrames(true);//ALPrint(eng alist); WriteLn;
  {$ENDIF}
  if al_Frames <> al_Nil then al_Frames := al_CDR[al_Frames].vi;
  Dec(al_CallDepth);
  {$IFDEF ALISP3_VERBOSE_ENTER_LEAVE}
  {$IFDEF ALISP3_EXTRA_VERBOSE_LEAVE}
  WriteLn(' <<LEAVE (', al_CallDepth, '): '); ALDumpFrames(true);//ALPrint(eng alist); WriteLn;
  {$ENDIF}
  {$ENDIF}
end;


function ReadChar (): Char;
// read a single character from the input stream
// and return it.
// ReadChar()=EOF indicates that the input is exhausted.
// all blanks will be converted to spaces.
// does line counting.

  function NextBuf (): Boolean;
  // returns TRUE on EOF
  begin
    if al_Input.bufPos >= al_Input.bufRead then
    begin
      al_Input.bufPos := 0;
      if (al_Input.inStream = nil) or (al_Input.inStream.Position >= al_Input.inStream.Size) then
        al_Input.bufRead := 0
      else
      begin
        al_Input.bufRead := al_Input.inStream.Size-al_Input.inStream.Position;
        if al_Input.bufRead > SizeOf(al_Input.buf) then al_Input.bufRead := SizeOf(al_Input.buf);
        al_Input.inStream.ReadBuffer(al_Input.buf, al_Input.bufRead);
      end;
      result := (al_Input.bufRead = 0);
    end
    else result := false;
  end;

var
  c: Char;
begin
  if al_Input.rejectedChar <> EOF then
  begin
    c := al_Input.rejectedChar;
    al_Input.rejectedChar := EOF;
    result := c;
    exit;
  end;

  if NextBuf() then
  begin
    // EOF
    result := EOF;
    exit;
  end;

  c := al_Input.buf[al_Input.bufPos]; Inc(al_Input.bufPos);

  if (c = #13) or (c = #10) then
  begin
    //!WriteLn('{', input.line, '}: ', lsss); lsss := '';
    Inc(al_Input.line);
    result := EOL;
    if (c = #13) and not NextBuf() and (al_Input.buf[al_Input.bufPos] = #10) then Inc(al_Input.bufPos);
    exit;
  end;

  //if c < ' ' then result := ' ' else result := c;
  //!lsss := lsss+c;
  if c <= #2 then c := ' ';
  result := c;
end;

procedure UnreadChar (ch: Char);
begin
  if (ch <> EOF) and (al_Input.rejectedChar <> EOF) then
    raise EALispError.Create('invalid UnreadChar() usage');
  al_Input.rejectedChar := ch;
end;

function SkipSpaces (): Char;
// read characters from the input device until a non-space
// character is found. return that character.
begin
  result := ReadChar();
  while (result > #0) and (result <= ' ') do result := ReadChar();
end;

function SkipToEOL (): Char;
// read characters from the input device until EOL.
begin
  result := ReadChar();
  while (result <> EOL) and (result <> EOF) do result := ReadChar();
end;

function SkipComment (): Char;
// skip /* ... */ comment
var
  pch: Char;
begin
  result := ReadChar();
  repeat
    pch := result; result := ReadChar();
    if (pch = '*') and (result = '/') then
    begin
      result := ReadChar();
      exit;
    end;
  until result = EOF;
  result := ERR;
end;

function SkipBlanks (): Char;
// skip spaces and comments, return first token char
var
  ch: Char;
begin
  repeat
    result := SkipSpaces();
    case result of
      ';': result := SkipToEOL();
      '/':
        begin
          ch := ReadChar();
          case ch of
            '*': result := SkipComment();
            '/': result := SkipToEOL();
            else UnreadChar(ch);
          end;
        end;
    end;
  until (result = EOF) or (result = ERR) or (result > ' ');
end;

function ReadExpr (): Integer;
// read an expression from the current input stream
// and return (a pointer to) it.
// each valid expression is either an Atom or a List.
var
  ch, nch: Char;
  l, e: Integer;
  s: string;
  fn: TALFloat;
  wasW: Boolean;
  li_l, li_p: Integer;
begin
  ch := SkipBlanks();
  if ch = ERR then begin result := ALThrowError('unfinished comment'); exit; end;
  if ch = EOF then begin result := EOFa; exit; end;
  li_l := al_Input.line; li_p := al_Input.inStream.Position-al_Input.bufRead+al_Input.bufPos;

  if ch = '(' then result := ReadList()
  else if ch = '[' then result := ReadVector()
  else if ch = #39 then  //'
  begin
    l := ReadExpr();
    result := ALCons(al_squote, ALCons(l, al_Nil));
    {if ALIsAtom(eng l) then result := ALCons(eng s_quote, ALCons(eng l, nilp))
    else result := ALCons(eng s_quote, l);}
  end
  else if ch = ')' then
  begin
    if al_LoadNestingLevel = 0 then result := ALThrowError('unexpected ")"') else result := RPARENa;
  end
  else if ch = ']' then
  begin
    if al_LoadNestingLevel = 0 then result := ALThrowError('unexpected "]"') else result := RSQPARENa;
  end
  else if ch = '.' then
  begin
    if al_LoadNestingLevel = 0 then result := ALThrowError('unexpected "."') else result := DOTa;
  end
  else
  begin
    // accept symbol
    s := '';
    if (ch = '"') or (ch = '`') then
    begin
      // autoquoted string
      nch := ch;
      ch := ReadChar();
      if (ch = EOL) or (ch = EOF) then
      begin
        result := ALThrowError('unfinished string');
        UnreadChar(ch);
        exit;
      end;
      while (ch <> nch) and (ch <> EOL) and (ch <> EOF) do
      begin
        if ch = '\' then
        begin
          ch := ReadChar();
          case ch of
            //'\', '"', #39: ;
            EOL, EOF: begin result := ALThrowError('invalid "\" usage'); exit; end;
            'r': ch := #13;
            'v': ch := #10;
            'n': begin s := s+#13; ch := #10; end;
            't': ch := #9;
            'b': ch := #8;
            'e': ch := #27;
            '$','x','X':
              begin
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'9','A'..'F']) then begin result := ALThrowError('invalid "\x" usage'); exit; end;
                l := ord(ch)-ord('0'); if ch > '9' then Dec(l, 7);
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'9','A'..'F']) then begin result := ALThrowError('invalid "\x" usage'); exit; end;
                l := l*16+ord(ch)-ord('0'); if ch > '9' then Dec(l, 7);
                ch := chr(l);
              end;
            '0','o','O':
              begin
                if ch <> '0' then ch := UpCase(ReadChar());
                if not (ch in ['0'..'7']) then begin result := ALThrowError('invalid "\o" usage'); exit; end;
                l := ord(ch)-ord('0');
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'7']) then begin result := ALThrowError('invalid "\o" usage'); exit; end;
                l := l*8+ord(ch)-ord('0');
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'7']) then begin result := ALThrowError('invalid "\o" usage'); exit; end;
                l := l*8+ord(ch)-ord('0');
                if l > 255 then begin result := ALThrowError('invalid "\o" charcode'); exit; end;
                ch := chr(l);
              end;
            'd','D':
              begin
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'9']) then begin result := ALThrowError('invalid "\d" usage'); exit; end;
                l := ord(ch)-ord('0');
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'9']) then begin result := ALThrowError('invalid "\d" usage'); exit; end;
                l := l*10+ord(ch)-ord('0');
                ch := UpCase(ReadChar());
                if not (ch in ['0'..'9']) then begin result := ALThrowError('invalid "\d" usage'); exit; end;
                l := l*10+ord(ch)-ord('0');
                if l > 255 then begin result := ALThrowError('invalid "\d" charcode'); exit; end;
                ch := chr(l);
              end;
            //else
          end;
        end;
        s := s+ch;
        ch := ReadChar();
      end;
      if ch <> nch then result := ALThrowError('unfinished string')
      else if nch = '`' then result := ALNewStrAtom(s)
      else result := ALCons(al_squote, ALCons(ALNewStrAtom(s), al_Nil));
    end
    else
    begin
      wasW := false;
      while (ch <> ' ') and (ch <> EOL) and (ch <> EOF) and
            (ch <> '(') and (ch <> ')') and (ch <> '[') and (ch <> ']') and
            (ch <> ';') do
      begin
        if ch = '!' then
        begin
          wasW := true;
          ch := ReadChar();
          if (ch = EOL) or (ch = EOF) then
          begin
            result := ALThrowError('invalid "!" usage');
            UnreadChar(ch);
            exit;
          end;
        end
        else if ch = '/' then
        begin
          nch := ReadChar();
          case nch of
            '/': begin ch := SkipToEOL(); break; end;
            '*': begin ch := SkipComment(); break; end;
            else
              begin
                s := s+ch; ch := nch;
                continue;
              end;
          end;
        end;
        s := s+ch;
        ch := ReadChar();
      end;
      if ch <> ' ' then UnreadChar(ch);

      if s = al_StringPool[al_CDR[al_strue].vi] then result := al_strue
      else if s = al_StringPool[al_CDR[al_sfalse].vi] then result := al_sfalse
      else if s = al_StringPool[al_CDR[al_snil].vi] then result := al_snil
      else if s = al_StringPool[al_CDR[al_ssymbols].vi] then result := al_ssymbols
      else if s = al_StringPool[al_CDR[al_sframes].vi] then result := al_sframes
      else if not wasW then
      begin
        l := 0; // shut up the compiler
        e := 1;
        if (s <> '') and (s[1] in ['0'..'9','$','+','-']) then
        begin
          if (Length(s) < 2) or (s[1] <> '0') or (UpCase(s[2]) <> 'X') then Val(s, l, e);
        end;
        if e = 0 then result := ALNewIntAtom(l)
        else
        begin
          Val(s, fn, e);
          if e = 0 then result := ALNewFloatAtom(fn)
          else result := ALNewStrAtom(s);
        end;
      end
      else result := ALNewStrAtom(s);
    end;
  end;
  {$IFDEF ALISP3_DEBUG}{$R+}{$ENDIF}
  if result > 0 then
  begin
    al_LineInfo[result].fileName := al_Input.fileName;
    al_LineInfo[result].line := li_l;
    al_LineInfo[result].pos := li_p;
  end;
  {$IFDEF ALISP3_DEBUG}{$R-}{$ENDIF}
end;

function ReadList (): Integer;
// read a list (S0 ... SN) and return (a pointer to) it
// this routine also recognizes pairs of the form (S0.S1)
// for empty lists, it returns nilp
const
  badPairMsg = 'bad pair';
var
  n: Integer; // cell read
  m: Integer; // ptr to list
  a: Integer; // used to append cells to m
  c: Integer; // member counter
  li_l, li_p: Integer;
begin
  result := al_Nil; Inc(al_LoadNestingLevel);
  m := ALCons(al_Nil, al_Nil); // root cell
  ALPushCell(m); // protect root cell
  a := al_Nil;
  c := 0;
  li_l := al_Input.line; li_p := al_Input.inStream.Position-al_Input.bufRead+al_Input.bufPos;
  repeat
    n := ReadExpr();
    if al_BreakFlag <> ALBREAK_NONE then begin result := n; break; end;
    if n = EOFa then begin result := ALThrowError('EOF in list'); break; end;
    if n = DOTa then
    begin
      if c = 0 then begin result := ALThrowError(badPairMsg); break; end;
      n := ReadExpr();
      if n < 0 then begin result := ALThrowError(badPairMsg); break; end;
      if a = al_Nil then a := m;
      al_CDR[a].vi := n;
      if ReadExpr() <> RPARENa then begin result := ALThrowError(badPairMsg); break; end;
      ALPopCells(1);
      Dec(al_LoadNestingLevel);
      result := m;
      exit;
    end;
    if n = RSQPARENa then begin result := ALThrowError('unexpected "]"'); break; end;
    if n = RPARENa then break;
    if a = al_Nil then a := m   // first member: insert at root
    else a := al_CDR[a].vi;  // following members: append
    al_CAR[a] := n;
    al_CDR[a].vi := ALCons(al_Nil, al_Nil);  // allocate space for next member
    Inc(c);
  until false;
  Dec(al_LoadNestingLevel);
  ALPopCells(1);
  if al_BreakFlag <> ALBREAK_NONE then exit;
  if a <> al_Nil then al_CDR[a].vi := al_Nil;// remove trailing empty cell
  if c <> 0 then
  begin
    result := m;
    {$IFDEF ALISP3_DEBUG}{$R+}{$ENDIF}
    if result > 0 then
    begin
      al_LineInfo[result].fileName := al_Input.fileName;
      al_LineInfo[result].line := li_l;
      al_LineInfo[result].pos := li_p;
    end;
    {$IFDEF ALISP3_DEBUG}{$R-}{$ENDIF}
  end
  else result := al_Nil;
end;

function ReadVector (): Integer;
var
  n: Integer;
  li_l, li_p: Integer;
begin
  Inc(al_LoadNestingLevel); result := al_Nil; // root cell
  li_l := al_Input.line; li_p := al_Input.inStream.Position-al_Input.bufRead+al_Input.bufPos;
  repeat
    n := ReadExpr();
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag <> ALBREAK_ERROR then
      begin
        if result <> al_Nil then ALPopCells(1);
        Dec(al_LoadNestingLevel);
      end;
      result := n;
      exit;
    end;
    if n = EOFa then
    begin
      Dec(al_LoadNestingLevel);
      if result <> al_Nil then ALPopCells(1);
      result := ALThrowError('EOF in vector');
      exit;
    end;
    if n = DOTa then
    begin
      Dec(al_LoadNestingLevel);
      if result <> al_Nil then ALPopCells(1);
      result := ALThrowError('"." in vector');
      exit;
    end;
    if n = RPARENa then
    begin
      Dec(al_LoadNestingLevel);
      if result <> al_Nil then ALPopCells(1);
      result := ALThrowError('unexpected ")"');
      exit;
    end;
    if n = RSQPARENa then break;
    // add item
    if result = al_Nil then
    begin
      result := ALCons(n, al_Nil);
      ALPushCell(result);
    end
    else
    begin
      al_CDR[result].vi := ALCons(n, al_Nil);
      result := al_CDR[result].vi;
    end;
  until false;
  Dec(al_LoadNestingLevel);
  if result <> al_Nil then result := ALPopCells(1);
  result := ALNewVectorFromList(result);
  {$IFDEF ALISP3_DEBUG}{$R+}{$ENDIF}
  if result > 0 then
  begin
    al_LineInfo[result].fileName := al_Input.fileName;
    al_LineInfo[result].line := li_l;
    al_LineInfo[result].pos := li_p;
  end;
  {$IFDEF ALISP3_DEBUG}{$R-}{$ENDIF}
end;


function ALEscapeStr (const s: string): string;
begin
  result := ALEscapeStrEx(s, false, '`');
end;

{$HINTS OFF} // shut up the compiler
function ALEscapeStrEx (const s: string; quoteAnyway: Boolean; quoteCh: Char): string;
var
  f, e: Integer;
  i64: Int64;
  fp: TALFloat;
  t: string;
  ch: Char;
  wasScr: Boolean;
begin
  if s = '' then
  begin
    result := quoteCh+quoteCh;
    exit;
  end;
  result := s; wasScr := false;
  f := 1; while f <= Length(result) do
  begin
    ch := result[f];
    if ch < ' ' then
    begin
      case ch of
        #8: t := '\b';
        #9: t := '\t';
        #10: t := '\v';
        #13:
          begin
            if (f < Length(result)) and (result[f+1] = #10) then
            begin
              Delete(result, f, 1); t := '\n';
            end
            else t := '\r';
          end;
        else t := '\$'+IntToHex(ord(ch), 2);
      end;
      Insert(t, result, f); Inc(f, Length(t));
      Delete(result, f, 1);
      wasScr := true;
    end
    else if ch in ['\','"'] then
    begin
      Insert('\', result, f); Inc(f, 2); wasScr := true;
    end
    else if ch in [' ','!','`','(',')','[',']',';'] then begin wasScr := true; Inc(f); end
    else if (ch = '/') and (f < Length(result)) and (result[f+1] in ['/','*']) then
    begin
      wasScr := true; Inc(f);
    end
    else Inc(f);
  end;
  if not quoteAnyway and not wasScr then
  begin
    if result <> '' then
    begin
      if result[1] = '.' then wasScr := true;
    end;
    if not wasScr then begin Val(result, i64, e); wasScr := (e = 0); end;
    if not wasScr then begin Val(result, fp, e); wasScr := (e = 0); end;
  end;
  if wasScr or quoteAnyway then result := quoteCh+result+quoteCh;
end;
{$HINTS ON} // shut up the compiler


function LispPrint (n: Integer; doQuote: Boolean): string;
// recursively print the term N
begin
  if n = al_Nil then result := '*nil*'
  else if n = al_snil then result := '*nil*'
  else if (n and MFLAG) <> 0 then result := '<free_cell>'
  else if (n and AFLAG) <> 0 then result := '<unprintable_value>'
  else if (n < 0) or (n > High(al_CAR)) then result := '<invalid_cell>'
  else if (al_CAR[n] and AFLAG) <> 0 then
  begin
    result := ALGetAtomStr(n);
    if (al_CAR[n] and TYPE_MASK) = DT_STRING then result := ALEscapeStr(result);
  end
  else if doQuote and (ALGetAtomStr(al_CAR[n]) = '@quote') and
          (al_CDR[n].vi <> al_Nil) and
          not ALIsAtom(al_CDR[n].vi) then
  begin
    n := al_CDR[n].vi;
    if al_CDR[n].vi = al_Nil then
    begin
      n := al_CAR[n];
      result := LispPrint(n, doQuote);
      if (ALGetAtomType(n) = DT_STRING) and (Length(result) > 1) and (result[1] = '`') then
        result := '"'+Copy(result, 2, Length(result)-2)+'"'
      else result := #39+result;
    end
    else result := #39+LispPrint(n, doQuote);
  end
  else
  begin
    // print list or pair
    result := '(';
    while n <> al_Nil do
    begin
      if (al_CAR[n] and MFLAG) <> 0 then
      begin
        result := result+'<free_cell!>';
        break;
      end;
      result := result+LispPrint(al_CAR[n], doQuote);
      n := al_CDR[n].vi;
      if (n and MFLAG) <> 0 then
      begin
        result := result+'<free_cell!!>';
        break;
      end;
      if ALIsAtom(n) then
      begin
        result := result+' . ';
        result := result+LispPrint(n, doQuote);
        n := al_Nil;
      end;
      if n <> al_Nil then result := result+' ';
    end;
    result := result+')';
  end;
end;

procedure ALPrint (n: Integer);
// print expression.
begin
  Write(LispPrint(n, true));
end;

procedure ALPrintNoQ (n: Integer);
// print expression.
begin
  Write(LispPrint(n, false));
end;

function ALPrintToStr (n: Integer): string;
// print expression.
begin
  result := LispPrint(n, true);
end;


procedure ALNewInStream (st: TStream; const fName: string; out old: TALInStreamInfo);
begin
  old := al_Input;
  al_Input.inStream := st; al_Input.fileName := fName; al_Input.rejectedChar := EOF;
  al_Input.line := 1;
  al_Input.bufPos := 0; al_Input.bufRead := 0;
  Inc(al_Input.loadDepth);
end;

procedure ALRestoreInStream (const si: TALInStreamInfo);
begin
  al_Input := si;
end;

function ALReadExpr (): Integer;
// read next expression.
begin
  al_BreakFlag := ALBREAK_NONE;
  al_LoadNestingLevel := 0; result := ReadExpr();
  if al_BreakFlag <> ALBREAK_NONE then exit;
  if al_LoadNestingLevel <> 0 then begin al_LoadNestingLevel := 0; result := ALThrowError('READEXPR: unfinished list'); exit; end;
  if result = EOFa then begin result := -1; exit; end;
end;

function ALInterpretNext (): Integer;
begin
  result := ALReadExpr();
  if (result = -1) or (al_BreakFlag <> ALBREAK_NONE) then exit;
  result := ALEval(result);
  case al_BreakFlag of
    //ALBREAK_NONE, ALBREAK_ERROR: ;
    ALBREAK_RETURN: al_BreakFlag := ALBREAK_NONE;
    ALBREAK_BREAK: result := ALThrowError('BREAK without loop', result);
    //ALBREAK_THROW: result := ALError(eng 'uncatched THROW', result);
    ALBREAK_CONT: result := ALThrowError('CONTINUE without loop', result);
  end;
end;

function InterpretX (): Integer;
var
  lastRes: Integer;
begin
  result := al_Nil;
  repeat
    lastRes := result; result := ALInterpretNext();
    //if al_BreakFlag <> ALBREAK_NONE then WriteLn(al_BreakFlag);
  until (result = -1) or (al_BreakFlag <> ALBREAK_NONE);
  if result = -1 then result := lastRes;
end;

function ALLoadStream (st: TStream; const fName: string = ''): Integer;
// load S-expressions from the stream.
// return result of evaluation
var
  si: TALInStreamInfo;
begin
  if al_Input.loadDepth+1 >= MAX_LOAD_LEVEL then
  begin
    result := ALThrowError('loading level too deep');
    exit;
  end;
  if al_VerbLoadFlag and (fName <> '') then WriteLn('':al_Input.loadDepth, 'loading "', fName, '"');
  ALNewInStream(st, fName, si);
  try
    result := InterpretX();
    if al_BreakFlag >= ALBREAK_NONE then
    begin
      if al_VerbLoadFlag and (fName <> '') then WriteLn('':al_Input.loadDepth-1, '"', fName, '" loaded');
    end;
  finally
    ALRestoreInStream(si);
  end;
end;

function ALLoadFile (const fn: string): Integer;
// load S-expressions from the external file
var
  st: TStream;
begin
  try
    st := TFileStream.Create(fn, fmOpenRead or fmShareDenyWrite);
  except
    result := ALNewStrAtom('cannot open file "'+fn+'"');
    al_BreakFlag := ALBREAK_THROW;
    exit;
  end;
  try
    result := ALLoadStream(st, fn);
  finally
    st.Free();
  end;
end;

function ALEvalStr (const s: string): Integer;
// evaluate string
var
  st: TStream;
begin
  if al_Input.loadDepth+1 >= MAX_LOAD_LEVEL then
  begin
    result := ALThrowError('ALEVALSTR: nesting level too deep');
    exit;
  end;
  st := TMemoryStream.Create();
  try
    if s <> '' then
    begin
      st.WriteBuffer(s[1], Length(s));
      st.Position := 0;
    end;
    result := ALLoadStream(st);
  finally
    st.Free();
  end;
end;


function EvalAtomEx (n: Integer; out pcnt: Integer): Integer;
// atom MUST be correct
// this function evaluates atom (with possible context dereferencing)
var
  f: Integer;
begin
  pcnt := 0;
  // only the string atoms needs to be evaluated
  // (any other atom represents itself and need no evaluation)
  if ((al_CAR[n] and AFLAG) <> 0) and ((al_CAR[n] and TYPE_MASK) = DT_STRING) then
  begin
    result := al_CDR[n].vi;
    if result = al_CDR[al_snil].vi then begin result := al_Nil; exit; end;

    // check if this is context dereferencing.
    // dereferences will be walked down recursively.
    f := al_StringPool.DColonPos[result];
    if f <> 0 then
    begin
      // process context dereferences
      //WriteLn('EvalAtomEx: ', al_StringPool[result]);
      result := SymContextImmerse(al_StringPool[result], f, pcnt);
      //WriteLn('EvalAtomEx: ', result, '; ', pcnt);
    end
    else
    begin
      // just find a symbol
      result := FindSymbol(al_StringPool[result], result);
    end;
    if result = al_Nil then
    begin
      // oops %-)
      ALLeaveContexts(pcnt); pcnt := 0;
      result := ALThrowError('undefined variable', n);
      exit;
    end;
    result := al_CDR[result].vi; // symbol value
    // the followings is just a strings...
    if result = al_ssymbols then result := al_CurrentCtx.fSymbols
    else if result = al_sframes then result := al_Frames
    else if result = al_snil then result := al_Nil;
  end
  else result := n;
end;

function EvalAtom (n: Integer): Integer;
// atom MUST be correct
var
  pcnt: Integer;
begin
  result := EvalAtomEx(n, pcnt);
  if pcnt <> 0 then ALLeaveContexts(pcnt);
end;

function DoFluidList (n: Integer; list: TStringList; addToList: Boolean): Integer;
var
  c: Integer;
begin
  while n <> al_Nil do
  begin
    if (ALGetAtomType(n) <> -1) or (ALGetAtomType(al_CAR[n]) <> DT_STRING) then
    begin
      if addToList then result := ALThrowError('FLUID: invalid idlist', n)
      else result := ALThrowError('UNFLUID: invalid idlist', n);
      exit;
    end;
    if addToList then list.Append(al_StringPool[al_CDR[al_CAR[n]].vi])
    else
    begin
      if list.Find(al_StringPool[al_CDR[al_CAR[n]].vi], c) then list.Delete(c);
    end;
    n := al_CDR[n].vi;
  end;
  result := al_snil;
end;

{$INCLUDE alisp3opt.pas}

function EvalArgs (n: Integer; var ctxUp: Integer; out cnt: Integer): Integer;
// N is the arguments. collect evaluation results in the
// new list and return this list
var
  svCUp, f: Integer;
  svCtx: array [0..15] of TLispContextObject;
begin
  cnt := 0;
  ALPushCell(n);

  // unwind contexts
  svCUp := ctxUp;
  if svCUp > 0 then
  begin
    for f := 0 to svCUp-1 do
    begin
      svCtx[f] := al_CurrentCtx; ALLeaveContexts(1);
    end;
    ctxUp := 0;
  end;

  result := al_Nil; // the list of evaluated args
  while n <> al_Nil do
  begin
    if ALIsAtom(n) then
    begin
      result := ALThrowError('invalid arglist', n);
      if al_BreakFlag <> ALBREAK_ERROR then
      begin
        if result <> al_Nil then ALPopCells(1);
        ALPopCells(1);
      end;
      exit;
    end;
    f := al_CAR[n];
    if (al_CAR[f] and AFLAG) <> 0 then
    begin
      // atom. only the string atoms needs to be evaluated
      // (any other atom represents itself and need no evaluation)
      if (al_CAR[f] and TYPE_MASK) = DT_STRING then f := EvalAtom(f);
    end
    else
    begin
      // list. no alternatives.
      f := ALEval(f);
    end;

    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag = ALBREAK_ERROR then begin result := f; exit; end;
      if result <> al_Nil then ALPopCells(1);
      ALPopCells(1);
      result := f;
      exit;
    end;
    if result = al_Nil then
    begin
      // first item
      result := ALCons(f, al_Nil);
      ALPushCell(result); // protect new list
    end
    else
    begin
      // append item to the list
      al_CDR[result].vi := ALCons(f, al_Nil);
      result := al_CDR[result].vi;
    end;
    n := al_CDR[n].vi;
    Inc(cnt);
  end;
  if result <> al_Nil then result := ALPopCells(1); // get list of evaluated args
  ALPopCells(1); // drop original list

  // wind contexts
  if svCUp > 0 then
  begin
    for f := svCUp-1 downto 0 do
    begin
      al_ContextStack.Add(Pointer(al_CurrentCtx));
      al_CurrentCtx := svCtx[f];
    end;
  end;
  ctxUp := svCUp;
end;

function BindArgs (varl, argl: Integer): Integer;
// !!call only after ALEnter()!!
// bind lambda arguments
// varl: list of names; argl: list of evaluated args
// if name = '|': stop binding, locals started
// argl is always correct (it's created by EvalArgs)
var
  n, ss, al, dsi, nsi: Integer;
begin
  if (varl and AFLAG) <> 0 then begin result := ALThrowError('invalid idlist', varl); exit; end;
  result := al_Nil;
  // protect lists
  ALPushCell(ALCons(varl, argl));
  varl := al_CAR[varl];
  dsi := al_StringPool.Find('|');
  al := al_CDR[al_CAR[al_Frames]].vi;
  while varl <> al_Nil do
  begin
    if (varl and AFLAG) <> 0 then
    begin
      ALPopCells(1);
      result := ALThrowError('invalid idlist', varl);
      exit;
    end;
    n := al_CAR[varl]; varl := al_CDR[varl].vi;
    if (al_CAR[n] and TYPE_MASK) <> DT_STRING then
    begin
      ALPopCells(1);
      result := ALThrowError('identifier expected', n);
      exit;
    end;
    nsi := al_CDR[n].vi;
    if nsi = dsi then begin argl := al_Nil; continue; end;
    {$IFDEF ALISP3_CHECK_LMB_ARGS}
    n := FindSymbolInList(al_StringPool[nsi], al, nsi);
    if n <> al_Nil then
    begin
      ALPopCells(1);
      result := ALThrowError('duplicate identifier', n);
      exit;
    end;
    {$ENDIF}
    // add symbol...
    // new string atom
    ss := ALCons(al_Nil, al_Nil);
    al_CAR[ss] := AFLAG or DT_STRING; al_CDR[ss].vi := nsi;
    al := ALCons(ALCons(ss, al_CAR[argl]), al);
    al_CDR[al_CAR[al_Frames]].vi := al;
    argl := al_CDR[argl].vi;
  end;
  ALPopCells(1); // drop lists
end;

function ALEval (n: Integer): Integer;
label eval_end;
var
  pr: PALPrimitive;
  c: Integer;
  pcnt: Integer;
begin
  result := al_Nil; pcnt := 0;
  if (n < 0) or (n > High(al_CAR)) or ((al_CAR[n] and MFLAG) <> 0) then exit; // paranoid? %-)
{$IFDEF ALISP3_ENABLE_DEBUGGER}
  if Assigned(al_Debugger) then
  begin
    result := al_Debugger({$IFDEF ALISP3_MULTIENGINE}_eng,{$ENDIF}n, true);
    if result <> -1 then exit;
    result := al_Nil;
  end;
{$ENDIF}
  if (n = al_Nil) or (n = al_snil) then goto eval_end;
  if al_BreakFlag <> ALBREAK_NONE then begin result := n; goto eval_end; end;

  // atom?
  if (al_CAR[n] and AFLAG) <> 0 then
  begin
    // evaluate atom
    result := EvalAtom(n);
    goto eval_end;
  end;

  // evaluate lists
  result := al_CAR[n];
  if ALIsAtom(result) then
  begin
    //Write('>>> result: '); ALPrintNoQ(result); WriteLn;
    result := EvalAtomEx(result, pcnt); // calling fn
    //Write('>>> ev-result: '); ALPrintNoQ(result); WriteLn;
    if al_BreakFlag <> ALBREAK_NONE then goto eval_end;
    //if pcnt <> 0 then WriteLn('[method call]');

    if (al_CAR[result] and AFLAG) <> 0 then
    begin
      // atom
      // this must be primitive
      if (al_CAR[result] and TYPE_MASK) <> DT_PRIMITIVE then
      begin
        //WriteLn('>>> type: ', al_CAR[result] and TYPE_MASK);
        //Write('>>> result: '); ALPrintNoQ(result); WriteLn;
        //Write('>>> comes: '); ALPrintNoQ(n); WriteLn;
        result := ALThrowError('primitive expected', n);
        goto eval_end;
      end;
      // primitive
      pr := PALPrimitive(al_CDR[result].vi);
      if pr.macro then
      begin
        result := al_CDR[n].vi;
        // count args number
        c := 0; n := result;
        while n <> al_Nil do
        begin
          if ALIsAtom(n) then
          begin
            result := ALThrowError('invalid arglist', result);
            goto eval_end;
          end;
          Inc(c); n := al_CDR[n].vi;
        end;
      end
      else
      begin
        result := EvalArgs(al_CDR[n].vi, pcnt, c);
        if al_BreakFlag <> ALBREAK_NONE then goto eval_end;
      end;
      if (c < pr.minArg) or ((pr.maxArg >= pr.minArg) and (c > pr.maxArg)) then
      begin
        result := ALThrowError('invalid number of arguments for primitive '+pr.name);
        goto eval_end;
      end;
      //!result := ALCons(eng ALNewStrAtom(eng pr.name), result);
      ALPushCell(result); // save arglist
      result := pr.func({$IFDEF ALISP3_MULTIENGINE}_eng,{$ENDIF}result);
      if al_BreakFlag <> ALBREAK_ERROR then ALPopCells(1); // drop arglist
      goto eval_end;
    end;
  end;

  // first item is list
  // first item now in <result>. it can be taken directly
  // or obtained thru EvalAtom()
  // ( (...) ...)
  // check for lambda-call
  ALPushCell(n); // push original list
  ALPushCell(result); // push lambda definition
  c := al_CAR[result];
  if (al_CAR[c] and not PROTFLAG) = (AFLAG or DT_STRING) then
  begin
    if al_StringPool[al_CDR[c].vi] = 'lambda' then c := al_slambda
    else if al_StringPool[al_CDR[c].vi] = 'mlambda' then c := al_smlambda;
  end;
  if c = al_smlambda then
  begin
    // macro
    ALEnter(al_CAR[n]);
    result := BindArgs(al_CDR[result].vi, ALCons(al_CDR[n].vi, al_Nil));
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag <> ALBREAK_ERROR then
      begin
        // clear stacks
        ALLeave();
        ALPopCells(2); // drop lambda definition & original list
      end;
      goto eval_end;
    end;
    result := ALPopCells(1); // pop lambda definition
  end
  else if c = al_slambda then
  begin
    // normal
    ALEnter(al_CAR[n]);
    //if pcnt <> 0 then WriteLn('{method call}');
    result := EvalArgs(al_CDR[n].vi, pcnt, c);
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag <> ALBREAK_ERROR then
      begin
        // some kind of BREAK/THROW
        ALLeave(); // this "return" is not for the current function
        ALPopCells(2); // drop lambda definition & original list
      end;
      goto eval_end;
    end;
    c := result;
    result := ALPopCells(1); // pop lambda definition
    ALPushCell(ALCons(c, result)); // push evaluated args & lambda definition
    c := BindArgs(al_CDR[result].vi, c);
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag <> ALBREAK_ERROR then
      begin
        ALPopCells(1); // pop evaluated args & lambda definition
        ALLeave();
        ALPopCells(1); // drop original list
      end;
      result := c;
      goto eval_end;
    end;
    ALPopCells(1); // pop evaluated args & lambda definition
  end
  else
  begin
    result := ALThrowError('function expected', result); //n
    if al_BreakFlag <> ALBREAK_ERROR then
    begin
      // clear stacks
      ALLeave();
      ALPopCells(2); // drop lambda definition and original list
    end;
    goto eval_end;
  end;
  // now: ALEnter() done; args binded
  // result: lambda definition
  //if pcnt <> 0 then WriteLn('<method call>') else WriteLn('<call>');
  //ALPrintNoQ(al_CAR[al_CDR[result].vi]); WriteLn; WriteLn('------------');
  ALPopCells(1); // drop original list
  n := al_CDR[al_CDR[result].vi].vi; // body
  ALPushCell(n); result := al_Nil;
  // execute body
  while n <> al_Nil do
  begin
    if (al_CAR[n] and AFLAG) <> 0 then
    begin
      // atom %-(
      ALPopCells(1); // drop body
      ALLeave();
      result := ALThrowError('invalid function body');
      goto eval_end;
    end;
    result := ALEval(al_CAR[n]);
    if al_BreakFlag <> ALBREAK_NONE then
    begin
      if al_BreakFlag = ALBREAK_ERROR then goto eval_end;
      break;
    end;
    n := al_CDR[n].vi;
  end;
  ALPopCells(1); // drop body
  ALLeave();
  if al_BreakFlag = ALBREAK_RETURN then al_BreakFlag := ALBREAK_NONE;

eval_end:
  if pcnt <> 0 then ALLeaveContexts(pcnt);
{$IFDEF ALISP3_ENABLE_DEBUGGER}
  if Assigned(al_Debugger) then
  begin
    n := al_Debugger({$IFDEF ALISP3_MULTIENGINE}_eng,{$ENDIF}n, false);
    if n <> -1 then result := n;
  end;
{$ENDIF}
end;


function ALNextStrArg (const funcName: string; var n: Integer; out s: string; out res: Integer): Boolean;
begin
  res := al_CAR[n];
  if ALGetAtomType(res) <> DT_STRING then
  begin
    s := '';
    res := ALThrowError(funcName+': string expected', al_CAR[n]);
    result := false;
  end
  else
  begin
    s := al_StringPool[al_CDR[res].vi];
    n := al_CDR[n].vi;
    result := true;
  end;
end;

function ALNextIntArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
begin
  res := al_CAR[n];
  if ALGetAtomType(res) <> DT_INTEGER then
  begin
    v := 0;
    res := ALThrowError(funcName+': integer expected', al_CAR[n]);
    result := false;
  end
  else
  begin
    v := al_CDR[res].vi;
    n := al_CDR[n].vi;
    result := true;
  end;
end;

function ALNextFloatArg (const funcName: string; var n: Integer; out v: TALFloat; out res: Integer): Boolean;
begin
  res := al_CAR[n];
  case ALGetAtomType(res) of
    DT_INTEGER:
      begin
        v := al_CDR[res].vi;
        n := al_CDR[n].vi;
        result := true;
        exit;
      end;
    DT_FLOAT:
      begin
        v := al_CDR[res].vf;
        n := al_CDR[n].vi;
        result := true;
        exit;
      end;
  end;
  v := 0;
  res := ALThrowError(funcName+': number expected', al_CAR[n]);
  result := false;
end;

function ALNextVectorArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
begin
  res := al_CAR[n];
  if ALGetAtomType(res) <> DT_VECTOR then
  begin
    v := 0;
    res := ALThrowError(funcName+': vector expected', al_CAR[n]);
    result := false;
  end
  else
  begin
    v := res;
    n := al_CDR[n].vi;
    result := true;
  end;
end;

function ALNextByteVectorArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
begin
  res := al_CAR[n];
  if ALGetAtomType(res) <> DT_BYTEVECTOR then
  begin
    v := 0;
    res := ALThrowError(funcName+': byte vector expected', al_CAR[n]);
    result := false;
  end
  else
  begin
    v := res;
    n := al_CDR[n].vi;
    result := true;
  end;
end;

function ALNextAnyVectorArg (const funcName: string; var n: Integer; out v: Integer; out res: Integer): Boolean;
var
  t: Integer;
begin
  res := al_CAR[n];
  t := ALGetAtomType(res);
  if (t <> DT_BYTEVECTOR) and (t <> DT_VECTOR) then
  begin
    v := 0;
    res := ALThrowError(funcName+': vector or byte-vector expected', al_CAR[n]);
    result := false;
  end
  else
  begin
    v := res;
    n := al_CDR[n].vi;
    result := true;
  end;
end;

function ALNextObjectArg (const funcName: string; var n: Integer; out v: TObject; out res: Integer): Boolean;
begin
  res := al_CAR[n];
  if ALGetAtomType(res) <> DT_TOBJECT then
  begin
    v := nil;
    res := ALThrowError(funcName+': object expected', al_CAR[n]);
    result := false;
  end
  else
  begin
    v := TObject(al_CDR[res].vi);
    n := al_CDR[n].vi;
    result := true;
  end;
end;

procedure ALRebindPrimitives ();
var
  f, n: Integer;
  p: PALPrimitive;
begin
  al_snil := al_CAR[ALSetGlobalSymbol('*nil*')];
  al_sfalse := al_CAR[ALSetGlobalSymbol('*f*')];
  al_strue := al_CAR[ALSetGlobalSymbol('*t*')];
  al_ssymbols := al_CAR[ALSetGlobalSymbol('*symbols*')];
  al_sframes := al_CAR[ALSetGlobalSymbol('*frames*')];
  ALSetGlobalSymbol('MAIN', al_MainContextCell);

  for f := 0 to al_Primitives.Count-1 do
  begin
    p := PALPrimitive(al_Primitives[f]);
    n := ALCons(al_Nil, al_Nil);
    al_CAR[n] := AFLAG or DT_PRIMITIVE;
    al_CDR[n].vi := Integer(p);
    ALSetGlobalSymbol(p.name, n);
  end;

  al_sdefun := al_CDR[ALFindGlobalSymbol('defun')].vi;
  al_smacro := al_CDR[ALFindGlobalSymbol('macro')].vi;
  al_ssetq := al_CDR[ALFindGlobalSymbol('setq')].vi;
  al_slambda := al_CDR[ALFindGlobalSymbol('lambda')].vi;
  al_smlambda := al_CDR[ALFindGlobalSymbol('mlambda')].vi;
  al_squote := al_CDR[ALFindGlobalSymbol('quote')].vi;
  al_sxquote := al_CDR[ALFindGlobalSymbol('xquote')].vi;
  al_sfluid := al_CDR[ALFindGlobalSymbol('fluid')].vi;
  al_sunfluid := al_CDR[ALFindGlobalSymbol('unfluid')].vi;
end;

procedure ALAddPrimitive (const pName: string; pFunc: TALPrimitiveFn;
  pMacro: Boolean; pMinArg, pMaxArg: Integer);
var
  n, f, ccell: Integer;
  p: PALPrimitive;
  s, cn: string;
  ccnt: Integer;
begin
  if pName = '' then raise EALispError.Create('can''t add nameless primitive');
  if not Assigned(pFunc) then raise EALispError.Create('invalid primitive "'+pName+'" handler');

  f := al_Primitives.Count;
  while f > 0 do
  begin
    Dec(f); if PALPrimitive(al_Primitives[f]).name = pName then
      raise EALispError.Create('primitive "'+pName+'" already defined');
  end;

  n := ALCons(al_Nil, al_Nil);
  New(p);
  p.name := pName; p.func := pFunc; p.macro := pMacro;
  p.minArg := pMinArg; p.maxArg := pMaxArg;
  try
    al_Primitives.Add(p);
  except
    Dispose(p);
    raise;
  end;
  al_CAR[n] := AFLAG or DT_PRIMITIVE;
  al_CDR[n].vi := Integer(p);

  //ALSetGlobalSymbol(pName, n);
  {ALPushCell(al_Frames); al_Frames := al_Nil;
  ALSetSymbolEx(pName, n, -1, true, true);
  al_Frames :=ALPopCells(1);}

  ALPushCell(n);
  // find (or create) all contexts
  s := pName; cn := ''; ccnt := 0;
  f := Pos('::', s);
  while (f <> 0) and (f < Length(s)-1) do
  begin
    if f > 1 then
    begin
      cn := Copy(s, 1, f-1);
      ccell := ALFindGlobalSymbol(cn);
      if ccell = al_Nil then
      begin
        // create new context atom & context symbol
        ccell := ALNewContextAtom();
        ccell := ALSetGlobalSymbol(cn, ccell, true);
        ALSetSymbolProtectionFlag(ccell, true);
      end;
      // try to enter into the context
      if not ALIsContext(al_CDR[ccell].vi) then
      begin
        ALLeaveContexts(ccnt);
        ALPopCells(1);
        raise EALispError.Create('symbol "'+cn+'" is not a context');
      end;
      if not ALEnterContext(al_CDR[ccell].vi) then
      begin
        ALLeaveContexts(ccnt);
        ALPopCells(1);
        raise EALispError.Create('can''t enter context "'+cn+'"');
      end;
      Inc(ccnt);
    end;
    Delete(s, 1, f+1);
    f := Pos('::', s);
  end;
  ALSetSymbolProtectionFlag(ALSetGlobalSymbol(s, n, true), true);
  ALLeaveContexts(ccnt);
  ALPopCells(1);
end;

procedure ALReinit ();
// initialize the interpreter variables,
// build the freeList list,
// create builtin symbols.
// register user primitives.
var
  f: Integer;
begin
  // misc. variables
  SetLength(al_CAR, 1); SetLength(al_CDR, 1);
  SetLength(al_LineInfo, 1);
  al_LineInfo[0].fileName := '';
  al_LineInfo[0].line := 0;
  al_LineInfo[0].pos := 0;
  al_Nil := 0; al_CAR[al_Nil] := al_Nil; al_CDR[al_Nil].vi := al_Nil;
  al_LoadNestingLevel := 0;
  al_ContextStack.Clear();
  al_Frames := al_Nil;
  al_CallDepth := 0;
  al_CellStack := al_Nil;
  al_ShowCSFlag := true;
  al_VerbLoadFlag := true;
  al_Optimize := true;
  al_ThrowSysError := false;

  al_StringPool.Clear();
  al_CustomGCRoots.Clear();
  al_ContextStack.Clear();

  f := al_TObjects.Count-1;
  while f >= 0 do begin TObject(al_TObjects[f]).Free(); Dec(f); end;
  al_TObjects.Clear();
  al_MainContextObj := nil;

  f := al_TVectors.Count-1;
  while f >= 0 do begin TList(al_TVectors[f]).Free(); Dec(f); end;
  al_TVectors.Clear();

  // initialize freeList
  al_FreeList := al_Nil;

  // clear input buffer
  al_Input.inStream := nil;
  al_Input.bufPos := 0;
  al_Input.bufRead := 0;
  al_Input.fileName := '';
  al_Input.rejectedChar := EOF;
  al_Input.line := 1;
  al_Input.loadDepth := 0;

  // second stage of initialization:
  // create built-in symbols, rebind primitives
  al_MainContextObj := TLispContextObject.Create({$IFDEF ALISP3_MULTIENGINE}_eng{$ENDIF});
  al_CurrentCtx := al_MainContextObj;
  al_MainContextCell := ALNewObjectAtom(al_MainContextObj);

  ALRebindPrimitives();
end;

procedure ALInitInternals ();
begin
  al_GC_sl := nil; al_GC_ol := nil; al_GC_vl := nil;
  al_GC_touched_vectors0 := nil; al_GC_touched_vectors1 := nil;

  al_DefPathList := '.;$(appdir)';
  al_CAR := nil; al_CDR := nil;
  al_FluidList := TStringList.Create();
  al_FluidList.Sorted := true;
  al_FluidList.Duplicates := dupIgnore;
  al_FluidList.CaseSensitive := true;
  al_CustomGCRoots := TList.Create();
  al_TObjects := TSortedIntList.Create();
  al_TVectors := TSortedIntList.Create();
  al_Primitives := TList.Create();
  al_StringPool := TStringTree.Create();
  al_MainContextObj := nil;
  al_CurrentCtx := nil;
  al_ContextStack := TList.Create();
  al_Debugger := nil;
end;

procedure ALClearInternals ();
var
  il: TSortedIntList;
  tl: TList;
  f: Integer;
begin
  al_DefPathList := '';
  SetLength(al_GC_sl, 0); SetLength(al_GC_ol, 0); SetLength(al_GC_vl, 0);
  FreeAndNil(al_GC_touched_vectors0); FreeAndNil(al_GC_touched_vectors1);
  FreeAndNil(al_FluidList);
  FreeAndNil(al_CustomGCRoots);
  FreeAndNil(al_StringPool);
  SetLength(al_CAR, 0); SetLength(al_CDR, 0);
  SetLength(al_LineInfo, 0);

  il := al_TObjects;
  if il <> nil then
  begin
    f := il.Count; while f > 0 do begin Dec(f); TObject(il[f]).Free(); end;
    FreeAndNil(al_TObjects);
  end;
  al_MainContextObj := nil;
  al_CurrentCtx := nil;

  il := al_TVectors;
  if il <> nil then
  begin
    f := il.Count; while f > 0 do begin Dec(f); TObject(il[f]).Free(); end;
    FreeAndNil(al_TVectors);
  end;

  tl := al_Primitives;
  if tl <> nil then
  begin
    f := tl.Count;
    while f > 0 do
    begin
      Dec(f);
      Dispose(PALPrimitive(tl[f]));
    end;
    FreeAndNil(al_Primitives);
  end;

  al_Debugger := nil;
end;

{$INCLUDE alisp3fn.pas}
{$INCLUDE alisp3builtprim.pas}

{$IFDEF ALISP3_MULTIENGINE}
function ALSpawnEngine (): PALISP3;
var
  _eng: PALISP3;
begin
  New(PXALISP3(result));
  // fool preprocessor %-)
  _eng := result;
  ALInitInternals();
  ALReinit();
  ALDefineBuiltInPrimitives();
  ALDefineStdPrimitives();
  ALReinit();
end;

procedure ALKillEngine (var _eng: PALISP3);
begin
  ALClearInternals();
  Dispose(PXALISP3(_eng));
  _eng := nil;
end;
{$ENDIF}


procedure ALDumpPrimitives ();
var
  f: Integer;
  pr: PALPrimitive;
begin
  for f := 0 to al_Primitives.Count-1 do
  begin
    pr := PALPrimitive(al_Primitives[f]);
    if pr.macro then Write('M') else Write(' ');;
    Write(' "', pr.name, '"'#9'(', pr.minArg, ';');
    if pr.maxArg >= pr.minArg then Write(pr.maxArg) else Write('*');
    WriteLn(')');
  end;
end;


initialization
  //alDefPathList := '.;$(appdir)';
  {$IFNDEF ALISP3_MULTIENGINE}
  ALInitInternals();
  ALReinit();
  ALDefineBuiltInPrimitives();
  ALDefineStdPrimitives();
  ALReinit();
  {$ENDIF}
finalization
  {$IFNDEF ALISP3_MULTIENGINE}
  ALClearInternals();
  {$ENDIF}
end.
