// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//

function LispProcessLambda (n: Integer): Integer;
// n was PUSHed by calee
// this function will process all lists (except quoted)
// and will change all ids that denotes primitives to
// direct primitive calls.
// convert local defun/macro to setq, add locals if necessary.
// opt_args: list of all arglists of all lambdas
//  (the 1st one is the opt_args of this lambda)

  function XEvalAtom (n: Integer): Integer;
  var
    f, pcnt: Integer;
  begin
    pcnt := 0;
    if ((al_CAR[n] and AFLAG) <> 0) and ((al_CAR[n] and TYPE_MASK) = DT_STRING) then
    begin
      result := al_CDR[n].vi;
      if result = al_CDR[al_snil].vi then begin result := al_Nil; exit; end;
      f := al_StringPool.DColonPos[result];
      if f <> 0 then
      begin
        //result := SymContextImmerse(al_StringPool[result], f, pcnt);
        result := n; exit;
      end
      else
      begin
        result := FindSymbol(al_StringPool[result], result);
      end;
      if result = al_snil then result := al_Nil;
      if result = al_Nil then result := -1 else result := al_CDR[result].vi;
    end
    else result := n;
    ALLeaveContexts(pcnt);
  end;

  function HasId (si: Integer; lst: Integer): Boolean;
  // lst: arglist
  begin
    result := false;
    while lst <> al_Nil do
    begin
      if ALIsAtom(lst) then exit;
      if (ALGetAtomType(al_CAR[lst]) = DT_STRING) and (al_CDR[al_CAR[lst]].vi = si) then
      begin
        result := true;
        exit;
      end;
      lst := al_CDR[lst].vi;
    end;
  end;

  function IsFluidId (si: Integer): Boolean;
  // search thru fluids
  var
    l: Integer;
    s: string;
  begin
    s := al_StringPool[si];
    result := al_FluidList.Find(s, l);
    if result then exit;
    result := al_opt_fluids.Find(s, l);
    if result then exit;

    l := Pos('::', s); if l = 0 then exit;
    Delete(s, 1, l+1);
    result := al_FluidList.Find(s, l);
    if result then exit;
    result := al_opt_fluids.Find(s, l);
  end;

  function IsLocalId (si: Integer): Boolean;
  // search thru all opt_args
  // search thru fluids
  var
    l: Integer;
  begin
    l := al_CDR[al_opt_args].vi;
    while l <> al_Nil do
    begin
      result := HasId(si, al_CAR[l]);
      if result then exit;
      l := al_CDR[l].vi;
    end;
    result := IsFluidId(si);
  end;

  function AddLocalVar (si: Integer): Boolean;
  var
    l, pl: Integer;
  begin
    {WriteLn('si: ', si);
    WriteLn('add local: "', stringPool[si], '"');}
    result := true;
    l := al_CAR[al_CDR[al_opt_args].vi];
    if HasId(si, l) then begin {WriteLn('has!');} exit; end;
    if l = al_Nil then
    begin
      // create id atom
      l := ALCons(al_Nil, al_Nil);
      al_CAR[l] := AFLAG or DT_STRING; al_CDR[l].vi := si;
      l := ALCons(l, al_Nil); // atom becomes list
      l := ALCons(al_opt_pipe, l); // arglist
      al_CAR[al_CDR[al_opt_args].vi] := l; // set new arglist for current lambda
    end
    else
    begin
      // there is at least one id in arglist
      // find opt_pipe
      result := false;
      pl := al_Nil;
      while l <> al_Nil do
      begin
        if (al_CAR[l] and AFLAG) <> 0 then exit; // invalid arglist
        if ALGetAtomType(al_CAR[l]) <> DT_STRING then exit; // invalid arglist
        //if cdr[car[l]] = opt_psi then begin WriteLn('HAS LOCALS!'); end;
        if al_CDR[al_CAR[l]].vi = al_opt_psi then break; // locals found
        pl := l; l := al_CDR[l].vi; // next id
      end;
      // l=nilp: no locals yet
      //   pl: last arg
      // l<>nilp: opt_pipe cell

      if l <> al_Nil then
      begin
        pl := l;
        //ALPrintNoQ(pl); WriteLn;
      end
      else
      begin
        ASSERT(pl <> al_Nil);
        // append opt_pipe
        al_CDR[pl].vi := ALCons(al_opt_pipe, al_Nil);
        pl := al_CDR[pl].vi; // pl: opt_pipe cell
      end;
      // create id atom
      l := ALCons(al_Nil, al_Nil);
      al_CAR[l] := AFLAG or DT_STRING; al_CDR[l].vi := si;
      // insert id
      al_CDR[pl].vi := ALCons(l, al_CDR[pl].vi);
      result := true;
    end;
  end;

  function ProcessList (n: Integer): Integer;
  var
    xxx: Integer;
  begin
    if n = al_Nil then begin result := al_Nil; exit; end;

    result := ALGetAtomType(n);
    if result = DT_STRING then
    begin
      // process id
      result := al_CDR[n].vi;
      if result = al_opt_psi then begin result := n; exit; end;
      if IsLocalId(result) then begin result := n; exit; end;
      //result := ALFindGlobalSymbol(eng stringPool[result], result);
      result := n;
      //WriteLn('str: "', ALGetAtomStr(result), '"');
      result := XEvalAtom(result);  // symbol value or -1
      //Write('e-str: '); ALPrintNoQ(result); WriteLn;
      if result = -1 then begin result := n; exit; end; // no such symbol

      //Write('GLB: '); ALPrintNoQ(result); WriteLn;
      // primitive: just return it
      if ALGetAtomType(result) = DT_PRIMITIVE then exit;
      {$IFDEF ALISP3_OPT_INSERT_LAMBDAS}
      // not a lambda? return original
      if ALGetAtomType(result) <> -1 then begin result := n; exit; end;
      xxx := XEvalAtom(al_CAR[result]);
      // unknown symbol? return original
      if xxx = -1 then begin result := n; exit; end;
      // lambda? just return it
      if (xxx = al_slambda) or (xxx = al_smlambda) then exit;
      {$ENDIF}
      // other things...
      result := n;
    end
    else if result = -1 then
    begin
      // pair or list
      result := al_CAR[n];
      if ALIsAtom(result) then
      begin
        result := XEvalAtom(result);
      end;

      // check for quoted list
      if (result = al_squote) or (result = al_sxquote) then
      begin
        // live quoted lists as is
        //result := n;
        result := ALCons(result, al_CDR[n].vi);
        //lineInfo[result] := lineInfo[n];
        exit;
      end;

      {$IFDEF ALISP3_OPT_LOCAL_DEFUNS}
      // check for local defun/macro
      if (result = al_sdefun) or (result = al_smacro) then
      begin
        // global defun?
        if Pos('::', ALGetAtomStr(al_CAR[al_CDR[n].vi])) <> 0 then
        begin
          result := n;
          exit;
        end;
        //WriteLn('DEFUN "', ALGetAtomStr(eng car[cdr[n]]), '"');
        //WriteLn('opt_args:'); ALPrintNoQ(opt_args); WriteLn; WriteLn;
        // try to add function local
        if AddLocalVar(al_CDR[al_CAR[al_CDR[n].vi]].vi) then
        begin
          //WriteLn('opt_args:'); ALPrintNoQ(opt_args); WriteLn; WriteLn;
          // build lambda expression
          if result = al_smacro then
            result := ALCons(al_smlambda, al_CDR[al_CDR[n].vi].vi)
          else result := ALCons(al_slambda, al_CDR[al_CDR[n].vi].vi);
          //WriteLn('NEW LAMBDA:'); ALPrintNoQ(result); WriteLn;
          //lineInfo[result] := lineInfo[cdr[n]];
          // optimize it
          ALPushCell(result);
          result := LispProcessLambda(result);
          ALPopCells(1);
          // transform to setq()
          //xxx := result;
          result := ALCons(al_squote, result);
          //lineInfo[result] := lineInfo[xxx];
          //xxx := result;
          ALPushCell(result);
          result := ALCons(al_ssetq,
            ALCons(al_CAR[al_CDR[n].vi]{name}, ALCons(result, al_Nil)));
          //lineInfo[result] := lineInfo[xxx];
          ALPopCells(1);
          exit;
        end;
      end;
      {$ENDIF}
      {$IFDEF ALISP3_OPT_NESTED_LAMBDAS}
      // check for nested lambda
      if (result = al_slambda) or (result = al_smlambda) then
      begin
        // process nested lambda
        ALPushCell(n);
        result := LispProcessLambda(n);
        ALPopCells(1);
        exit;
      end;
      {$ENDIF}
      // check for fluid and unfluid
      if (result = al_sfluid) or (result = al_sunfluid) then
      begin
        // process and drop out fluid() & unfluid()
        result := DoFluidList(al_CDR[n].vi, al_opt_fluids, result = al_sfluid);
        exit;
      end;

      result := -1;
      // process the list
      while n <> al_Nil do
      begin
        if ALIsAtom(n) then
        begin
          if result = -1 then
          begin
            result := n;
            ALPushCell(result); // push processed head
          end
          else
          begin
            al_CDR[result].vi := n;
            //result := cdr[result];
          end;
          n := al_Nil;
        end
        else
        begin
          if ALGetAtomType(al_CAR[n]) = -1 then
          begin
            // ((...))
            xxx := XEvalAtom(al_CAR[al_CAR[n]]);
            if (xxx = al_sfluid) or (xxx = al_sunfluid) then
            begin
              DoFluidList(al_CDR[al_CAR[n]].vi, al_opt_fluids, xxx = al_sfluid);
              // skip procedure
              n := al_CDR[n].vi; continue;
            end;
          end;

          if result = -1 then
          begin
            result := ALCons(ProcessList(al_CAR[n]), al_Nil);
            //lineInfo[result] := lineInfo[car[n]];
            ALPushCell(result); // push processed head
          end
          else
          begin
            if ALIsAtom(al_CAR[n]) then al_CDR[result].vi := ALCons(al_CAR[n], al_Nil)
            else al_CDR[result].vi := ALCons(ProcessList(al_CAR[n]), al_Nil);
            result := al_CDR[result].vi;
            //lineInfo[result] := lineInfo[eng_s car[n]];
          end;
          n := al_CDR[n].vi;
        end;
      end;
      if result = -1 then result := al_Nil else result := ALPopCells(1);
    end
    else result := n;
  end;

var
  lmb{, xxx}: Integer;
begin
  //WriteLn('opt_args:'); ALPrintNoQ(opt_args); WriteLn; WriteLn;
  lmb := al_CAR[n];
  if not ALIsAtom(lmb) then begin result := n; exit; end;
  lmb := XEvalAtom(lmb);
  // don't process non-lambdas
  if (lmb <> al_smlambda) and (lmb <> al_slambda) then begin result := n; exit; end;

  n := al_CDR[n].vi; // opt_args
  al_CDR[al_opt_args].vi := ALCons(al_CAR[n], al_CDR[al_opt_args].vi); // save opt_args
  n := al_CDR[n].vi; // body
  result := ProcessList(n);
  //xxx := result;
  result := ALCons(lmb, ALCons(al_CAR[al_CDR[al_opt_args].vi], result));
  //lineInfo[result] := lineInfo[xxx];
  al_CDR[al_opt_args].vi := al_CDR[al_CDR[al_opt_args].vi].vi; // drop opt_args
end;


function LispCompile (n: Integer): Integer;
// n is lambda expression, pushed by calee
{var
  oef: Boolean;}
begin
  al_opt_fluids := TStringList.Create();
  // the following line allows using XEvalAtom()
  //oef := al_ThrowSysError; al_ThrowSysError := true;
  ALPushCell(n);
  try
    al_opt_fluids.Sorted := true;
    al_opt_fluids.Duplicates := dupIgnore;
    al_opt_fluids.CaseSensitive := true;
    {$IFDEF ALISP3_OPT_SHOW_SRC}
    Write('src: '); ALPrintNoQ(n); WriteLn;
    {$ENDIF}
    al_opt_pipe := ALNewStrAtom('|');
    al_opt_psi := al_CDR[al_opt_pipe].vi;
    ALPushCell(al_opt_pipe); // save opt_pipe atom (just in case)
    al_opt_args := ALCons(al_Nil, al_Nil);
    ALPushCell(al_opt_args); // make opt_args GC root
    result := LispProcessLambda(n);
    ALPopCells(3); // drop n, opt_args & opt_pipe atom
    {$IFDEF ALISP3_OPT_SHOW_RESULT}
    Write('optimized: '); ALPrintNoQ(result); WriteLn;
    {$ENDIF}
  finally
    //al_ThrowSysError := oef;
    al_opt_fluids.Free();
  end;
end;
