// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// FloatToSpeccy() is based on the code by Martijn van der Heide
//
unit zxFloatU;

interface


function FloatToSpeccy (value: Double; out buf): Boolean;
function SpeccyToFloat (const buf): Double;


implementation


function FloatToSpeccy (value: Double; out buf): Boolean;
// takes exactly 5 bytes
var
  ex: Double;
  expon: Integer;
  mantissa: LongWord;
  intValue: Integer;
  outP: Integer;
  sign: Byte;

  function Log2 (const x: Extended): Extended;
  asm
    fld1
    fld x
    fyl2x
    fwait
  end;

  function IntPower (const Base: Extended; const Exponent: Integer): Extended;
  asm
    mov     ecx, eax
    cdq
    fld1                      { Result := 1 }
    xor     eax, edx
    sub     eax, edx          { eax := Abs(Exponent) }
    jz      @@3
    fld     Base
    jmp     @@2
  @@1:
    fmul    ST, ST            { X := Base * Base }
  @@2:
    shr     eax,1
    jnc     @@1
    fmul    ST(1),ST          { Result := Result * X }
    jnz     @@1
    fstp    st                { pop X from FPU stack }
    cmp     ecx, 0
    jge     @@3
    fld1
    fdivrp                    { Result := 1 / Result }
  @@3:
    fwait
  end;

  procedure EmitB (b: Byte);
  begin
    ASSERT(outP < 5);
    PChar(@buf)[outP] := Char(b);
    Inc(outP);
  end;

begin
  outp := 0; result := true;

  intValue := trunc(value);
  if (value = intValue) and (intvalue >= -65535) and (intvalue <= 65535) then
  begin
    // store in short integer format
    EmitB($00);
    // insert sign
    if intValue >= 0 then EmitB($00) else begin EmitB($FF); Inc(intValue, 65536); end;
    EmitB(intValue and $FF);
    EmitB(intValue shr 8);
    EmitB($00);
  end
  else
  begin
    // need to store in full floating point format
    if value < 0 then
    begin
      // sign bit is high bit of byte 2
      sign := $80;
      value := -value;
    end
    else sign := $00;

    ex := int(log2(value));
    if (ex < -129) or (ex > 126) then
    begin
      // number too big
      result := false;
      exit;
    end;

    expon := trunc(ex);
    if expon = 0 then ex := 1.0 else ex := IntPower(2.0, expon);

    // calculate mantissa
    ex := (value/ex-1.0)*2147483648.0;
    //value := int(ex); if value < 0 then value := value-1; // floor
    mantissa := LongWord(trunc(ex));

    // emit number bytes
    EmitB(expon+$81);  // insert exponent
    EmitB(((mantissa shr 24) and $7F) or sign); // insert mantissa
    EmitB((mantissa shr 16) and $FF); // (big endian!)
    EmitB((mantissa shr 8) and $FF);
    EmitB(mantissa and $FF);
  end;

  ASSERT(outP = 5);
end;

function SpeccyToFloat (const buf): Double;
var
  ba: packed array [0..4] of Byte absolute buf;
  exp: Integer;
  mant: LongWord;
  ex: Double;

  function IntPower (const Base: Extended; const Exponent: Integer): Extended;
  asm
    mov     ecx, eax
    cdq
    fld1                      { Result := 1 }
    xor     eax, edx
    sub     eax, edx          { eax := Abs(Exponent) }
    jz      @@3
    fld     Base
    jmp     @@2
  @@1:
    fmul    ST, ST            { X := Base * Base }
  @@2:
    shr     eax,1
    jnc     @@1
    fmul    ST(1),ST          { Result := Result * X }
    jnz     @@1
    fstp    st                { pop X from FPU stack }
    cmp     ecx, 0
    jge     @@3
    fld1
    fdivrp                    { Result := 1 / Result }
  @@3:
    fwait
  end;
begin
  if (ba[0] = 0) and (ba[4] = 0) and (ba[1] = 0) or (ba[1] = $FF) then
  begin
    // integer form
    result := ba[2]+ba[3]*256;
    if ba[1] = $FF then result := result-65536;
  end
  else
  begin
    // float form
    exp := ba[0]-$81;
    if exp = 0 then ex := 1.0 else ex := IntPower(2.0, exp);

    mant := (LongWord(ba[1] and $7F) shl 24);
    mant := mant or (LongWord(ba[2]) shl 16);
    mant := mant or (LongWord(ba[3]) shl 8);
    mant := mant or ba[4];

    result := mant;
    result := (result*ex/2147483648.0)+ex;
    if (ba[1] and $80) <> 0 then result := -result;
  end;
end;


end.
