// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// GNU GPL
//
// the code was taken from Yase2 by Christian Hackbart
// note from Ketmar:
//   i still don't fully understand this code, but it seems to work for now.
//   i'll comment (and maybe rewrite) it in the future
//   for now i've changed Hobeta loader a little, added some tests, fixed
//   some bugs (i even don't remember where they were %-) & so on.
//   also TR-DOS disk management (TRxxx) procedures was added.
//
//   default one-sector boot by POLTERGEIST/SPEED CO.
//
unit dspBetaU;
// fDiskStatus:
//   $04:
//   $40: -1: r/o
//   $80:

interface

uses
  Windows, Messages, CommDlg, SysUtils, Classes, dspEmuMouseU, dspDialogsU,
  xpackU, wdxpackU, dspUtilsU;


type
  PByteArrayTRD = ^TByteArrayTRD;
  TByteArrayTRD = packed array [0..maxInt-2] of Byte;

  TSCLHeader = packed record
    fileName: packed array [0..7] of Char;
    fileType: Byte;
    start: Word;
    len: Word;
    sizeSec: Byte;
  end;

  THobHeader = packed record
    name: packed array [0..7] of Char;
    ftype: Char;
    start: Word;
    len: Word;
    zero: Byte;  // always zero
    lenSec: Byte; // length in sectors
    sum: Word;
  end;


  TDiskImage = class
  private
    fDisk: PByteArrayTRD;
    fSize: LongWord;
    fPosition: LongWord;
    fFileName: string;
    fChanged: Boolean;

    procedure SetFileName (const value: string);
    procedure ReadSCL (const pFileName: string);
    procedure ReadHob (const pFileName: string);
    procedure AddSclFile (const hdr: TSCLHeader; const data);

  public
    constructor Create ();
    destructor Destroy (); override;
    procedure EmptyDisk ();
    procedure Eject ();
    procedure Read (out buffer; size: LongWord; out numRead: LongWord);
    procedure Write (const buffer; size: LongWord; out numWritten: LongWord);
    procedure Seek (position: LongWord);

    procedure ChangeFileName (const fName: string);
    procedure SaveToFile (const fName: string);

    procedure SetDirty (flagOn: Boolean);

    procedure AddStdBoot (forced: Boolean);

    function IsTRDisk (): Boolean;
    function TRFileCount (): Integer;
    function TRFreeSectors (): Integer;
    function TRFileName (index: Integer): string;
    function TRIsFileDeleted (index: Integer): Boolean;
    function TRFileExt (index: Integer): string;
    function TRFileSecSize (index: Integer): Integer;
    function TRFileStart (index: Integer): Integer;
    function TRFileSize (index: Integer): Integer;
    function TRFileData (index: Integer): Pointer;
    function TRFindFile (name: string; ext: Char; from: Integer = 0): Integer;
    procedure TRDeleteFile (index: Integer);
    procedure TRFileSaveHob (index: Integer; dosName: string);
    function LogSectorData (index: Integer): Pointer;


    property FilePos: LongWord read fPosition;
    property FileSize: LongWord read fSize;
    property FileName: string read fFileName write SetFileName;
    property Dirty: Boolean read fChanged;
  end;

  TBetaDisk = class
  private
    fRegStatus: Byte;
    fRegCom: Byte;
    fRegTrack: ShortInt;
    fRegSect: Byte;
    fRegData: Byte;
    fSystem: Byte;
    fStepDirect: ShortInt;
    fTrackReal: array [0..3] of Byte;
    fDiskBuf: array [0..$FF] of Byte;
    fCurrentDisk: Byte;
    fDskIndexCounter: Byte;
    fDskCountDatLost: Byte;
    fCntData,DskDataSize: Integer;
    fCntReady: Word;

    fDiskStatus: Byte;
    fReady: Boolean;
    fReadIndexTrack: Boolean;

    function DiskPosition (sect: Word): LongInt;
    procedure SetStateStep ();
    procedure OpCom ();
    procedure StepRear ();
    procedure OpReadSector ();
    procedure OpWriteSector ();

  public
    disks: array [0..3] of TDiskImage;

    constructor Create ();
    destructor Destroy (); override;

    procedure DoDiskVG (operIO: Byte; var dataIO: Byte);
  end;


var
  bootName: string = '';
  sclAutoboot: Boolean = true;
  trdAutoboot: Boolean = true;
  trxCheckSum: Boolean = true;


implementation


{$INCLUDE boot1s.inc}

const
  formatDatLen = 1;
  dskIndex: array [0..4] of Byte = (0, 1, 1, 0, 0);


function CalcHobSum (const head): Word;
var
  p: PChar;
  f: Integer;
begin
  result := 0; p := @head;
  for f := 0 to 14 do Inc(result, Byte(p[f])*257+f);
end;


{ TDiskImage }
constructor TDiskImage.Create ();
begin
  fSize := 655360;//256*4096;
  GetMem(fDisk, fSize);
  EmptyDisk();
end;

destructor TDiskImage.Destroy ();
begin
  Eject();
  FreeMem(fDisk);
end;

function TDiskImage.IsTRDisk (): Boolean;
begin
  result :=
    ((fDisk^[$08E3] = $16) or  // 80T,DS
    (fDisk^[$08E3] = $17)) and  // 80T,SS
    (fDisk^[$08E7] = $10);     // trdos flag
end;

function TDiskImage.TRFileCount (): Integer;
begin
  result := fDisk^[$08E4]; if result > 128 then result := 128;
end;

function TDiskImage.TRIsFileDeleted (index: Integer): Boolean;
begin
  if (index < 0) or (index > 127) then result := true
  else result := fDisk^[index*16] < 2;
end;

function TDiskImage.TRFileName (index: Integer): string;
begin
  if (index < 0) or (index > 127) then result := '        '
  else
  begin
    SetLength(result, 8);
    Move(fDisk^[index*16], result[1], 8);
    if result[1] = #0 then result := '';
  end;
end;

function TDiskImage.TRFileExt (index: Integer): string;
begin
  if (index < 0) or (index > 127) then result := ' '
  else result := char(fDisk^[index*16+8]);
end;

function TDiskImage.TRFileSecSize (index: Integer): Integer;
begin
  if (index < 0) or (index > 127) then result := 0
  else result := fDisk^[index*16+13];
end;

function TDiskImage.TRFileStart (index: Integer): Integer;
begin
  if (index < 0) or (index > 127) then result := 0
  else result := fDisk^[index*16+9]+fDisk^[index*16+10]*256;
end;

function TDiskImage.TRFileSize (index: Integer): Integer;
begin
  if (index < 0) or (index > 127) then result := 0
  else result := fDisk^[index*16+11]+fDisk^[index*16+12]*256;
end;

function TDiskImage.TRFileData (index: Integer): Pointer;
begin
  if (index < 0) or (index > 127) then result := nil
  else result := @(fDisk^[LongWord(Word(fDisk^[index*16+15])*16+fDisk^[index*16+14])*256]);
end;

function TDiskImage.TRFreeSectors (): Integer;
begin
  result := fDisk^[$08E5]+256*fDisk^[$08E6];
end;

procedure TDiskImage.TRDeleteFile (index: Integer);
begin
  if (index < 0) or (index > 127) then exit;
  if index >= TRFileCount() then exit;
  if index = TRFileCount()-1 then
  begin
    // the last file
    // patch directory info
    fDisk^[index*16] := 0; // "no file" flag
    // increment # of free sectors
    Inc(PWord(@(fDisk^[$08E5]))^, TRFileSize(index));
  end
  else
  begin
    // patch directory info
    fDisk^[index*16] := 1; // "erased file" flag
    // increment erased files counter
    Inc(fDisk^[$08F4]);
  end;
end;

function TDiskImage.TRFindFile (name: string; ext: Char; from: Integer = 0): Integer;
begin
  name := Copy(name+'        ', 1, 8);
  result := from;
  while (result < TRFileCount()) and
        ((TRFileName(result) <> name) or (TRFileExt(result) <> ext)) do Inc(result);
  if result >= TRFileCount() then result := -1;
end;

procedure TDiskImage.TRFileSaveHob (index: Integer; dosName: string);
var
  fo: TFileStream;
  head: THobHeader;
  trData: LongWord;
begin
  if (index < 0) or (index > 127) or (fDisk^[index*16+14] > 15) then exit;

  trData := LongWord(Word(fDisk^[index*16+15])*16+fDisk^[index*16+14])*256;
  if trData >= fSize then exit;

  // prepare HoBeta header
  Move(fDisk^[index*16], head, 13);
  if head.name[0] = #1 then head.name[0] := ' '; // was deleted
  head.zero := 0;
  head.lenSec := fDisk^[index*16+13];
  head.sum := CalcHobSum(head);

  // check if the whole file on the disk
  if trData+head.lenSec*256 > fSize then exit;

  fo := CreateFileStream(dosName);//TFileStream.Create(dosName, fmCreate);
  try
    try
      fo.WriteBuffer(head, SizeOf(head));
      fo.WriteBuffer(fDisk^[trData], head.lenSec*256);
    finally
      fo.Free();
    end;
  except
    DeleteFile(dosName);
    raise;
  end;
end;

function TDiskImage.LogSectorData (index: Integer): Pointer;
begin
  if (index < 0) or (LongWord(index) >= fSize div 256) then result := nil
  else result := @(fDisk^[index*256]);
end;

procedure TDiskImage.ChangeFileName (const fName: string);
begin
  fFileName := fName;
end;

procedure TDiskImage.SetDirty (flagOn: Boolean);
begin
  fChanged := flagOn;
end;

procedure TDiskImage.SaveToFile (const fName: string);
var
  fs: TFileStream;
  s: string;
  f, fc, a: Integer;
  hdr: TSCLHeader;
  //bf: TXArkBlockFile;
  wbf: TWDXBlockFile;
  sclSum: Integer;

  procedure WrBuff (const buf; count: Integer);
  var
    p: PChar;
  begin
    //if bf = nil then fs.WriteBuffer(buf, count)
    //else bf.WriteBuffer(buf, count);
    if wbf <> nil then wbf.WriteBuffer(buf, count)
    else fs.WriteBuffer(buf, count);
    // calc SCL sum
    p := @buf;
    while count > 0 do
    begin
      Inc(sclSum, Byte(p^)); Inc(p); Dec(count);
    end;
  end;

begin
  fs := CreateFileStream(fName);//TFileStream.Create(fName, fmCreate);
  {bf := nil; }wbf := nil; sclSum := 0;
  try
    try
      if (LowerCase(ExtractFileExt(fName)) = '.scl') or
         (LowerCase(ExtractFileExt(fName)) = '.scz') then
      begin
        // save scl
        if LowerCase(ExtractFileExt(fName)) = '.scz' then
          wbf := TWDXBlockFile.CreateNew(fs); //bf := TXArkBlockFile.CreateNew(fs);
        s := 'SINCLAIR';
        WrBuff(s[1], Length(s));
        fc := fDisk^[$08E4];
        WrBuff(fc, 1);

        // write dir
        a := 0;
        for f := 1 to fc do
        begin
          WrBuff(fDisk^[a], 14); Inc(a, 16);
        end;

        // write data
        for f := 0 to fc-1 do
        begin
          a := f*16;
          Move(fDisk^[a], hdr, 14); Inc(a, 14);
          a := fDisk^[a]+fDisk^[a+1]*16;
          WrBuff(fDisk^[a*256], hdr.sizeSec*$100);
        end;

        // now we have to write the sum
        WrBuff(sclSum, 4);
      end
      else
      begin
        if LowerCase(ExtractFileExt(fName)) = '.trz' then
        begin
          //bf := TXArkBlockFile.CreateNew(fs);
          //bf.WriteBuffer(fDisk^, fSize);
          wbf := TWDXBlockFile.CreateNew(fs);
          wbf.WriteBuffer(fDisk^, fSize);
        end
        else fs.WriteBuffer(fDisk^, fSize);
      end;
    finally
      //bf.Free();
      wbf.Free();
      fs.Free();
    end;
  except
    DeleteFile(fName);
    raise;
  end;
end;

procedure TDiskImage.Eject ();
begin
  if fChanged then
  begin
    repeat
      if fFileName = '' then
      begin
        case MessageBox(emuHWnd,
               PChar('Do you want to save the changed disk ?'), 'Disk changed',
               MB_YESNOCANCEL or MB_ICONQUESTION) of
          ID_NO, ID_CANCEL: break;
          ID_YES:
            begin
              fFileName := EmuDialogGetDiskNameToSave();
              if fFileName = '' then break;
            end;
        end;
      end
      else
      begin
        EmuUnlockMouse();
        case MessageBox(emuHWnd,
               PChar('Do you want to save the disk ("'+fFileName+'") ?'), 'Disk changed',
               MB_YESNOCANCEL or MB_ICONQUESTION) of
          ID_NO: break;
          ID_CANCEL:
            begin
              fFileName := EmuDialogGetDiskNameToSave();
              if fFileName = '' then break;
            end;
        end;
      end;
      try
        SaveToFile(fFileName);
        fChanged := false;
      except
      end;
    until true;
  end;
end;

procedure TDiskImage.EmptyDisk ();
begin
  fChanged := false; {fPosition := 0;} fFileName := '';

  ASSERT(fDisk <> nil);
  FillChar(fDisk^, fSize, 0);
  fDisk^[$08E2] := $1;   // first free track
  fDisk^[$08E3] := $16;  // 80T,DS
  fDisk^[$08E5] := $F0; // free sec
  fDisk^[$08E6] := $09; // free sec
  fDisk^[$08E7] := $10;  // trdos flag
  //!!FillChar(fDisk^[$08E9], 10, 32); // ???
  //FillChar(fDisk^[$08F5], 8, 32); // disk name
  // disk name
  fDisk^[$08F5] := ord('k');
  fDisk^[$08F6] := ord('8');
  fDisk^[$08F7] := ord('D');
  fDisk^[$08F8] := ord('S');
  fDisk^[$08F9] := ord('P');
  fDisk^[$08FA] := ord('e');
  fDisk^[$08FB] := ord('m');
  fDisk^[$08FC] := ord('u');

  WriteLn('disk: empty disk created');
end;

procedure TDiskImage.AddSclFile (const hdr: TSCLHeader; const data);
// add new file to disk
var
  free, pos, fipos: Integer;
begin
  WriteLn('disk: adding file: "', hdr.fileName, '"');

  free := fDisk^[$08E5]+256*fDisk^[$08E6];
  WriteLn(free, ' free sectors');
  WriteLn('file takes ', hdr.sizeSec, ' sectors');

  if fDisk^[$08E4] >= 127 then exit; // too many files
  if free < hdr.sizeSec then exit; // disk full

  // calc first free pos
  pos := fDisk^[$08E1]+16*fDisk^[$08E2];

  // store file data
  Move(data, fDisk^[pos*$100], hdr.sizeSec*$100);

  // set file info (w/o pos)
  fipos := fDisk^[$08E4]*16;
  Move(hdr, fDisk^[fipos], 14);

  // set file info (pos)
  fDisk^[fipos+14] := fDisk^[$08E1]; // starting sector
  fDisk^[fipos+15] := fDisk^[$08E2]; // starting track

  // correct free start
  Inc(pos, hdr.sizeSec);
  fDisk^[$08E1] := pos mod 16; // first free sector
  fDisk^[$08E2] := pos div 16; // first free track

  // one more file appears
  Inc(fDisk^[$08E4]);

  // decrease # of free sectors
  Dec(free, hdr.sizeSec);

  fDisk^[$08E5] := (free and $FF);
  fDisk^[$08E6] := ((free shr 8) and $FF);
  WriteLn(free, ' free sectors left');
  WriteLn(hdr.fileName, '.', chr(hdr.fileType), ' at #', IntToHex(hdr.start, 4),
    ', len=#', IntToHex(hdr.len, 4));
  WriteLn(hdr.fileName, '.', chr(hdr.fileType), ' at ', IntToStr(hdr.start):5,
    ', len=', IntToStr(hdr.len));

end;

procedure TDiskImage.AddStdBoot (forced: Boolean);
var
  hd: THobHeader;
  hdr: TSCLHeader;
  fn: string;
  f, c: Integer;
begin
  // is there any "boot"?
  // patch directory info
  f := TRFindFile('boot', 'B');
  if f <> -1 then
  begin
    if not forced then exit; // do nothing %-)
    // rename old boot
    c := ord('0');
    while c <= ord('z') do
    begin
      fn := chr(c)+'oot';
      if TRFindFile(fn, 'B') = -1 then break;
      Inc(c);
      if c > ord('9') then Inc(c, ord('a')-ord('9')-1);
    end;
    fDisk^[f*16] := c;
  end;

  if bootName <> '' then
  begin
    fn := ExtractFilePath(ParamStr(0))+bootName;
    if not FileExists(fn) then fn := bootName;
    try
      ReadHob(fn);
      exit;
    except
    end;
  end;

  Move(stdBoot[0], hd, SizeOf(hd));
  Move(hd, hdr, 13); hdr.sizeSec := hd.lenSec;
  AddSclFile(hdr, stdBoot[17]);
  fChanged := false;
end;


procedure TDiskImage.ReadSCL (const pFileName: string);
var
  fs: TFileStream;
  temp: PChar;
  sz: LongWord;
  id: packed array [0..7] of Char;
  data: LongWord;
  files, f: Integer;
  hdr: TSCLHeader;
  //autoBoot: Boolean;
  bf: TXArkBlockFile;
  wbf: TWDXBlockFile;
begin
  temp := nil; sz := 0; bf := nil; wbf := nil;
  try
    fs := TFileStream.Create(pFileName, fmOpenRead or fmShareDenyWrite);
  except exit;
  end;
  try
    try
      WriteLn('checking WDX...');
      if WDXIsBlockStream(fs) then
      begin
        wbf := TWDXBlockFile.Create(fs, true);
        WriteLn('WDX FILE!');
        sz := wbf.Size;
        GetMem(temp, sz);
        wbf.ReadBuffer(temp^, sz);
      end
      else if LowerCase(ExtractFileExt(pFileName)) = '.scz' then
      begin
        WriteLn('XPACK FILE!');
        bf := TXArkBlockFile.Create(fs, true);
        sz := bf.Size;
        GetMem(temp, sz);
        bf.ReadBuffer(temp^, sz);
      end
      else
      begin
        WriteLn('normal file');
        sz := fs.Size;
        GetMem(temp, sz);
        fs.ReadBuffer(temp^, sz);
      end;
    finally
      bf.Free(); wbf.Free();
      fs.Free();
    end;
  except
    ReallocMem(temp, 0);
  end;
  EmptyDisk();
  if temp = nil then exit;

  //autoBoot := sclAutoboot;
  fFileName := ChangeFileExt(pFileName, '.trd');
  try
    Move(temp^, id, 8);// SINCLAIR
    if id <> 'SINCLAIR' then exit;
    files := Byte(temp[8]);
    data := 9+14*files; // data (follows dir)
    if data > sz then exit;;
    for f := 0 to files-1 do
    begin
      Move(temp[9+14*f], hdr, 14);
      if data+hdr.sizeSec*$100 > sz then break;
      AddSclFile(hdr, temp[data]);
      Inc(data, hdr.sizeSec*$100);
      //if hdr.fileName = 'boot    B' then autoBoot := false;
    end;

    //if autoBoot then
    AddStdBoot(false);
  finally
    FreeMem(temp);
  end;
end;

procedure TDiskImage.ReadHOB (const pFileName: string);
var
  fs: TFileStream;
  temp: PChar;
  hdr: TSCLHeader;
  sz: LongWord;
  hd: THobHeader;
begin
  temp := nil; sz := 0;

  try
    fs := TFileStream.Create(pFileName, fmOpenRead or fmShareDenyWrite);
  except exit;
  end;
  try
    try
      sz := fs.Size;
      if sz < SizeOf(hd)+256 then
      begin
        WriteLn('invalid HoBeta file "', pFileName, '"');
        exit;
      end;
      GetMem(temp, sz);
      fs.ReadBuffer(temp^, sz);
    finally
      fs.Free();
    end;
  except
    ReallocMem(temp, 0);
  end;
  if temp = nil then exit;

  try
    //temp[13] := chr(ord(temp[14]) div 256);
    //temp[13] := temp[14];
    Move(temp^, hd, SizeOf(hd));
    if not trxCheckSum then hd.sum := CalcHobSum(hd);
    if (hd.zero <> 0) or (hd.lenSec = 0) or (hd.lenSec*256+SizeOf(hd) > sz) or
       (CalcHobSum(hd) <> hd.sum) then
    begin
      WriteLn('invalid HoBeta file "', pFileName, '"');
      WriteLn('zero: ', hd.zero);
      WriteLn('lenSec: ', hd.lenSec);
      WriteLn('full size: ', hd.lenSec*256, ' | ', sz);
      WriteLn('check sum: ', IntToHex(CalcHobSum(hd), 4), ' | ', IntToHex(hd.sum, 4));
      exit;
    end;
    Move(hd, hdr, 13); hdr.sizeSec := hd.lenSec;
    AddSclFile(hdr, temp[17]);
    fChanged := true;
  finally
    FreeMem(temp);
  end;
end;

procedure TDiskImage.SetFileName (const value: string);
var
  fs: TFileStream;
  sz: Int64;
  buf: array [0..65536] of Byte;
  head: TSCLHeader;
  s: string;
  f: Integer;
  bf: TXArkBlockFile;
  wbf: TWDXBlockFile;
begin
  if fFileName = value then exit;
  if (LowerCase(ExtractFileExt(value)) = '.trd') then Eject();
  if (LowerCase(ExtractFileExt(value)) = '.trz') then Eject();
  fChanged := false;
  fPosition := 0;

  if (value = '') or not FileExists(value) then
  begin
    WriteLn('no file "', value, '"');
    exit;
    EmptyDisk();
    fFileName := value;
    if fFileName <> '' then ChangeFileExt(fFileName, '.trd');
    if trdAutoboot then AddStdBoot(false);
  end
  else
  begin
    if (LowerCase(ExtractFileExt(value)) = '.scl') or
       (LowerCase(ExtractFileExt(value)) = '.scz') then
    begin
      ReadSCL(value);
      if fFileName = '' then
      begin
        fFileName := ChangeFileExt(value, '.trd');
      end
      else fFileName := '';
      exit;
    end;

    if Pos('$', ExtractFileExt(value)) > 0 then
    begin
      ReadHOB(value);
      exit;
    end;

    if (LowerCase(ExtractFileExt(value)) = '.trd') or
       (LowerCase(ExtractFileExt(value)) = '.trz') then
    begin
      // read trd
      EmptyDisk();
      fs := TFileStream.Create(value, fmOpenRead or fmShareDenyWrite);
      try
        bf := nil; wbf := nil;
        try
          if WDXIsBlockStream(fs) then
          begin
            wbf := TWDXBlockFile.Create(fs);
            sz := wbf.Size; if sz > fSize then sz := fSize;
            wbf.ReadBuffer(fDisk^, sz);
          end
          else if LowerCase(ExtractFileExt(value)) = '.trd' then
          begin
            sz := fs.Size; if sz > fSize then sz := fSize;
            fs.ReadBuffer(fDisk^, sz);
          end
          else
          begin
            bf := TXArkBlockFile.Create(fs);
            sz := bf.Size; if sz > fSize then sz := fSize;
            bf.ReadBuffer(fDisk^, sz);
          end;
          fFileName := value;
          WriteLn('disk "', fFileName, '" loaded');
        finally
          bf.Free(); wbf.Free();
          fs.Free();
        end;
      except
        EmptyDisk();
        if trdAutoboot then AddStdBoot(false);
        raise;
      end;

      if trdAutoboot then
      begin
        {for f := 0 to 127 do
        begin
          s := TRFileName(f);
          if s = '' then break;
          if (s = 'boot    ') and (TRFileExt(f) = 'B') then exit;
        end;}
        AddStdBoot(false);
      end;
      exit;
    end;

    // add raw file
    fs := TFileStream.Create(value, fmOpenRead or fmShareDenyWrite);
    try
      sz := fs.Size;
      if sz > 65536-256 then
      begin
        WriteLn('file too big!');
        exit;
      end;
      if sz = 0 then
      begin
        WriteLn('empty file!');
        exit;
      end;

      FillChar(buf[0], SizeOf(buf), 0);
      fs.ReadBuffer(buf[0], sz);

      s := Copy(ExtractFileName(value), 1, 8);
      for f := 1 to Length(s) do if s[f] > #127 then s[f] := '_';
      while Length(s) < 8 do s := s+' ';
      Move(s[1], head.fileName, 8);

      s := Copy(ExtractFileExt(value), 2, 1);
      if s = '' then s := ' ';
      if s[1] > #127 then s[1] := '_';
      head.fileType := Byte(s[1]);

      head.start := $6000;
      head.len := sz;
      head.sizeSec := sz div 256;
      if head.len mod 256 <> 0 then Inc(head.sizeSec);
      AddSclFile(head, buf[0]);

      WriteLn('file "', value, '" disked');
    finally
      fs.Free();
    end;
    exit;
  end;
end;

procedure TDiskImage.Read (out buffer; size: LongWord; out numRead: LongWord);
begin
  if fPosition+size > fSize then numRead := fSize-fPosition else numRead := size;
  Move(Pointer(LongWord(fDisk)+fPosition)^, buffer, numRead);
end;

procedure TDiskImage.Write (const buffer; size: LongWord; out numWritten: LongWord);
begin
 if fPosition+size > fSize then numWritten := fSize-fPosition else numWritten := size;
 Move(buffer, Pointer(LongWord(fDisk)+fPosition)^, numWritten);
 fChanged := true;
end;

procedure TDiskImage.Seek (position: LongWord);
begin
 fPosition := position; if fPosition > fSize then fPosition := fSize;
end;


{ TBetaDisk }
constructor TBetaDisk.Create ();
var
  f: Integer;
begin
  fDiskStatus := 0; fReady := true; fReadIndexTrack := false;
  fCurrentDisk := 0;
  for f := Low(disks) to High(disks) do disks[f] := TDiskImage.Create();
end;

destructor TBetaDisk.Destroy ();
var
  f: Integer;
begin
  for f := Low(disks) to High(disks) do disks[f].Free();
end;

function TBetaDisk.DiskPosition (sect: Word): LongInt;
begin
  result := ((fRegTrack shl 5) or (fSystem and $10 xor $10) or sect) shl $08;
end;

procedure TBetaDisk.SetStateStep ();
// set VG state after "step" commands
begin
  if fRegTrack < 0 then
  begin
    fTrackReal[fSystem and $03] := 0;
    fRegTrack := 0;
  end;
  if fRegTrack>$4F then
  begin
    fTrackReal[fSystem and $03] := $4F;
    fRegTrack := $4F;
  end;

  fRegStatus := fDiskStatus and $40;
  if (fRegCom and $08) > 0 then fRegStatus := fRegStatus or $20;
  if fRegTrack = 0 then fRegStatus := fRegStatus or $04;
end;

procedure TBetaDisk.OpCom ();
begin
  fDskCountDatLost := $00; fCntData := $0000; fRegStatus := $03;
  if not fReady then
  begin
    fRegStatus := $01;
    fCntReady := $10;
  end;
end;

procedure TBetaDisk.StepRear ();
// cmd: sector step backward
begin
  if fRegTrack>0 then
  begin
    Dec(fRegTrack);
    fTrackReal[fSystem and $03] := fRegTrack;
  end;
  fStepDirect := -1;
  SetStateStep();
end;

procedure TBetaDisk.OpReadSector ();
var
  numRead: LongWord;
  position: LongInt;
begin
  if (fRegSect > $10) or (Byte(fRegTrack) <> fTrackReal[fSystem and $03]) or
     ((fDiskStatus and $04) = 0) then
  begin
    fRegStatus := $10; {OpNotFound}
    exit;
  end;

  position := DiskPosition(fRegSect-1);

  disks[fCurrentDisk].Seek(position);
  disks[fCurrentDisk].Read(fDiskBuf, SizeOf(fDiskBuf), numRead);

  if position = 2048 then fDiskBuf[227] := 16; //BUG !!! at pos $8e3 must be $10??

  if numRead <> SizeOf(fDiskBuf) then fRegStatus := $08 //OpContSum
  else
  begin
    DskDataSize := $FF;
    fRegData := fDiskBuf[0];
    OpCom();
  end;
end;

procedure TBetaDisk.OpWriteSector ();
begin
  if (fDiskStatus and $40) > 0 then
  begin
    fRegStatus := $40; //OpWrProt
    exit;
  end;
  if (fRegSect > $10) or (Byte(fRegTrack) <> fTrackReal[fSystem and $03]) or
    ((fDiskStatus and $04) = 0) then fRegStatus := $10 //OpNotFound
  else
  begin
    DskDataSize := $0100;
    OpCom();
  end;
end;

procedure TBetaDisk.DoDiskVG (operIO: Byte; var dataIO: Byte);
var
  numWritten: LongWord;
  f: Integer;
  //attr: Integer;
  intrqDrq: Word;
begin
  if (fRegStatus and $01) <> 0 then
  begin
    // command is executing
    Dec(fDskCountDatLost);
    if fDskCountDatLost = 0 then fRegStatus := $04; // data lost
  end;
  if fCntReady > 0 then
  begin
    Dec(fCntReady);
    if fCntReady = 0 then fRegStatus := $03; // executing, data query
  end;

  case operIO of
    $0: // port #1F (cmd), read, ???
      begin
        dataIO := fRegStatus;
        if (fRegCom and $80) > 0 then exit;
        Dec(fDskIndexCounter);
        if (fDskIndexCounter and $0E) > 0 then dataIO := dataIO or $02;
      end;
    $1: // port #1F (cmd), write: do command
      begin
        if (fRegStatus and $01) = 0 then
        begin
          if (dataIO and $F0) = $D0 then exit;
          fRegCom := dataIO;
          case fRegCom shr 4 of
            $0: // init, ignore all params
              begin
                fTrackReal[fSystem and $03] := 0;
                fRegTrack := 0; fStepDirect := -1;
                SetStateStep();
              end;
            $1: // seek, ignore all params
              begin
                fTrackReal[fSystem and $03] := fRegData;
                fRegTrack := fRegData;
                if fRegData-fRegTrack < 0 then fStepDirect := -1;
                if fRegData-fRegTrack > 0 then fStepDirect := 1;
                SetStateStep();
              end;
            $2..$5:
              begin
                // 2, 3: step in prev. dir
                // 4, 5: step forward
                if ((fRegCom shr 4) in [2, 3]) and (fStepDirect = -1) then
                begin
                  // prev. was back
                  StepRear(); exit;
                end;
                // step forward
                if fRegTrack > 0 then
                begin
                  Inc(fRegTrack);
                  fTrackReal[fSystem and $03] := fRegTrack;
                end;
                fStepDirect := 1;
                SetStateStep();
              end;
            $6, $7: StepRear();
            $8, $9: OpReadSector();
            $0A, $0B: OpWriteSector();
            $0C: // read address
              begin
                // data: track #, side #, sector #, sector length, crc(?)
                DskDataSize := $0005;
                fRegData := 1;
                if fReadIndexTrack then fRegData := fRegTrack;
                OpCom();
              end;
            $0E: ; // read the whole track (& loose sync after this %-)
            $0F: // write (format) the whole track
              begin
                if (fDiskStatus and $40) > 0 then fRegStatus := $40 //OpWrProt
                else
                begin
                  DskDataSize := formatDatLen;
                  OpCom();
                end;
              end;
          end;
        end;
      end;
    $2: // port #3F, read: get current track
      dataIO := fRegTrack;
    $3: // port #3F, write: set current track
      if (fRegStatus and $01) = 0 then fRegTrack := dataIO;
    $4: // port #5F, read: get current sector
      dataIO := fRegSect;
    $5: // port #5F, write: set current sector
      if (fRegStatus and $01 = 0) then fRegSect := dataIO;
    $6: // port #7F, read: common data reg
      begin
        if (fRegStatus and $01) > 0 then
        begin
          case fRegCom shr 4 of
            $8, $9: // reading "read sector" bytes
              begin
                dataIO := fRegData;
                fDskCountDatLost := $10;
                if fCntData < DskDataSize then
                begin
                  Inc(fCntData);
                  fRegData := fDiskBuf[fCntData];
                end
                else
                begin
                  if (fRegCom and $10) > 0 then
                  begin
                    Inc(fRegSect);
                    OpReadSector();
                  end
                  else fRegStatus := $00; //OpOk
                end
              end;
            $C: // reading "read address" bytes
               begin
                 dataIO := fRegData;
                 fDskCountDatLost := $10;
                 if fCntData < DskDataSize then
                 begin
                   fRegData := dskIndex[fCntData];
                   Inc(fCntData);
                 end
                 else fRegStatus := $00; //OpOk
               end;
          end;
        end
        else dataIO := fRegData;
      end;
    $7: // port #7F, write: common data reg
      begin
        if (fRegStatus and $01) > 0 then
        begin
          case fRegCom shr 4 of
            $0A, $0B: // write "write sector" data
              begin
                fDiskBuf[fCntData] := dataIO;
                fRegData := dataIO; Inc(fCntData);
                fDskCountDatLost := $10;
                if fCntData >= DskDataSize then
                begin
                  disks[fCurrentDisk].Seek(DiskPosition(fRegSect-1));
                  disks[fCurrentDisk].Write(fDiskBuf, SizeOf(fDiskBuf), numWritten);
                  if numWritten <> SizeOf(fDiskBuf) then
                  begin
                    fRegStatus := $08; //OpContSum
                    exit;
                  end;
                  if (fRegCom and $10) > 0 then
                  begin
                    Inc(fRegSect);
                    OpWriteSector();
                  end
                  else fRegStatus := $00; //OpOk
                end;
              end;
            $0F: // write "format track" data
              begin
                fRegData := dataIO;
                fDskCountDatLost := $10;
                Inc(fCntData);
                if fCntData >= DskDataSize then
                begin
                  FillChar(fDiskBuf, SizeOf(fDiskBuf), 0);
                  if (fDiskStatus and $04) > 0 then
                  begin
                    disks[fCurrentDisk].Seek(DiskPosition(0));
                    for f := 0 to $10-1 do
                      disks[fCurrentDisk].Write(fDiskBuf, SizeOf(fDiskBuf), numWritten);
                  end
                  else
                  begin
                    // disks[fCurrentDisk].EmptyDisk;
                    // Problem!!!!!!!!11
                    for f := 0 to $A00-1 do
                      disks[fCurrentDisk].Write(fDiskBuf, SizeOf(fDiskBuf), numWritten);
                    fDiskStatus := fDiskStatus or $04;
                  end;
                  fRegStatus := $00; //OpOk
                end;
              end;
          end
        end
        else fRegData := dataIO;
      end;
    $8: // port #FF, read system register
      begin
        {if (fRegStatus and 32) > 0 then DRQ := $40 else DRQ := 0;
        if (fRegStatus and 64) > 0 then Intrq := $80 else Intrq := 0;
        dataIO := DRQ or intrq;}
        intrqDrq := $BF;
        if (fRegStatus and $01) > 0 then intrqDrq := $3F;
        if (fRegStatus and $02) > 0 then intrqDrq := $7F;
        dataIO := intrqDrq;
      end;
    $9: // port #FF, write to system register
      begin
        if ((fRegStatus and $01) > 0) and ((fRegCom and $80) > 0) then
        begin
          fRegStatus := $08; //OpContSum
          exit;
        end;
        fSystem := dataIO;
        if ((fSystem xor fDiskStatus) and $03) > 0 then
        begin
          DoDiskVG($0B, dataIO);
          DoDiskVG($0A, dataIO);
        end;
      end;
    $0A: // set current disk
      begin
        if (fDiskStatus and $80) > 0 then exit;
        fDiskStatus := fSystem and $03;
        fCurrentDisk := fDiskStatus;

        (*attr := FileGetAttr(disks[fCurrentDisk].FileName);
        if attr <> -1 then
        begin
          attr := attr and $01;
          fDiskStatus := fDiskStatus or (attr shl 6); // r/o?
          //if ioresult=0 then fDiskStatus:=fDiskStatus or $04;
        end;*)
        disks[fCurrentDisk].seek(0); // Reset
        fDiskStatus := fDiskStatus or $80;
        fDiskStatus := fDiskStatus or $04;
      end;
    $0B:
      begin
        if (fDiskStatus and $80) = 0 then exit;
        // if fDiskStatus and $04>0 then disks[fCurrentDisk].eject;//close(Str); !!!!
        fDiskStatus := $00;
      end;
    $0C:
      begin
        fRegStatus := $24;
        fRegCom := $00;
        fRegTrack := $00;
        fRegSect := $01;
        fRegData := $00;
        fSystem := $3C;
        fStepDirect := ShortInt($FF);
        if ((fDiskStatus and $80) > 0) and ((fDiskStatus and $03) > 0) then
        begin
          DoDiskVG($0B, dataIO);
          DoDiskVG($0A, dataIO);
        end;
      end;
  end;
end;


end.
