// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// variables & routines
// the idea of tape handling was taken from Unreal Spectrum by SMT
//
{$INCLUDE options.inc}
unit dspEmuMiscU;

interface

uses
  MMSystem,
  win32U, lispCon3U, xpackU,
  dspEmuVarsU,
  z80U,
  dspBetaU,
  dspVScrU, dspZXDrawU,
  dspAYEngineU,
  dspTZXU,
  dzsLibU,
  wpackU,
  {$IFDEF USE_HP_TIMER}hpTImerU,{$ELSE}qCounterU,{$ENDIF}
  dspMachConstU,
  dspConDriverU,
  dspEmuSoundU,
  dspEmuMouseU,
  dspWinU,
  Windows, SysUtils, Classes;


const
  //emuInfoMsgY = 12;
  emuInfoMsgY = 240-22;
  //emuInfoMsgColor = 11;
  emuInfoMsgColor = 13;
  emuInfoMsgOutColor = 0;


const
  // numberz for tape counter
  cntNumz: packed array [0..87] of Byte = (
    $FE,$C6,$BA,$BA,$BA,$BA,$C6,$FE, {0}
    $FE,$EE,$CE,$EE,$EE,$EE,$C6,$FE, {1}
    $FE,$C6,$BA,$FA,$C6,$BE,$82,$FE, {2}
    $FE,$C6,$BA,$E6,$FA,$BA,$C6,$FE, {3}
    $FE,$F6,$E6,$D6,$B6,$82,$F6,$FE, {4}
    $FE,$82,$BE,$86,$FA,$BA,$C6,$FE, {5}
    $FE,$C6,$BE,$86,$BA,$BA,$C6,$FE, {6}
    $FE,$82,$FA,$F6,$EE,$EE,$EE,$FE, {7}
    $FE,$C6,$BA,$C6,$BA,$BA,$C6,$FE, {8}
    $FE,$C6,$BA,$BA,$C2,$FA,$C6,$FE, {9}
    $FE,$C6,$BA,$BA,$BA,$BA,$C6,$FE);{0}


var
  conLogo: PByte = nil;
  joyDebugInfo: Boolean = false;


procedure EmuDrawWCon ();


function EmuTapeBit (): Byte;
procedure EmuRewindTape ();
  // go to the first palyable block (stop tape if no such blocks)
  // tape can be in the "playing state"
procedure EmuStartTape (user: Boolean = false); // "play" loaded tape
procedure EmuStopTape (user: Boolean = true);
  // "stop" loaded tape
  // if not <user> -- do autopause
function EmuTapeBlockCount (): Integer;
function EmuTapeCurBlock (): Integer;
procedure EmuTapeGoToBlock (num: Integer);
procedure EmuTapeSkipBlocks (num: Integer);
procedure EmuDrawTapeCounter (x, y: Integer);


procedure EmuRealizeState (state: PDZSZ80State);
function EmuGetState (): PDZSZ80State;

procedure DrawText (x, y: Integer; const s: string; fc, bc: Byte);
procedure DrawTextOutline (x, y: Integer; const s: string; tc, oc: Byte);

procedure EmuRenderScreen ();
procedure EmuRenderScreenNoInv ();
procedure EmuDrawScreen ();

procedure EmuInfoMessage (const msg: string; const timeOut: LongWord = 2000);

procedure EmuResetTimers ();
procedure EmuFixShiftsState ();

procedure EmuSetMouseXY (x, y: Integer);
procedure DoPCMouse ();
procedure PCKeys2ZX ();
procedure PCJoy2ZX ();

procedure PCJoyReset ();
procedure PCJoyScan ();


implementation


var
  outFont: packed array [Char, 0..9, 0..7] of Byte;


procedure BuildCharOutline (ch: Char);
// buf: 8x10 bytes
// values:
//  0: empty
//  1: fore
//  2: outline
var
  dx, dy: Integer;
  b: Byte;
begin
  for dy := 1 to 8 do
  begin
    b := dspFont[ch, dy-1];
    for dx := 1 to 6 do
    begin
      if (b and $80) <> 0 then outFont[ch, dy, dx] := 1;
      b := b shl 1;
    end;
  end;
  for dy := 1 to 8 do
  begin
    for dx := 1 to 6 do
    begin
      if outFont[ch, dy, dx] = 1 then
      begin
        // y-1
        if outFont[ch, dy-1, dx-1] = 0 then outFont[ch, dy-1, dx-1] := 2;
        if outFont[ch, dy-1, dx  ] = 0 then outFont[ch, dy-1, dx  ] := 2;
        if outFont[ch, dy-1, dx+1] = 0 then outFont[ch, dy-1, dx+1] := 2;
        // y
        if outFont[ch, dy, dx-1] = 0 then outFont[ch, dy, dx-1] := 2;
        if outFont[ch, dy, dx+1] = 0 then outFont[ch, dy, dx+1] := 2;
        // y+1
        if outFont[ch, dy+1, dx-1] = 0 then outFont[ch, dy+1, dx-1] := 2;
        if outFont[ch, dy+1, dx  ] = 0 then outFont[ch, dy+1, dx  ] := 2;
        if outFont[ch, dy+1, dx+1] = 0 then outFont[ch, dy+1, dx+1] := 2;
      end;
    end;
  end;
end;

procedure BuildOutFont ();
var
  ch: Char;
begin
  FillChar(outFont, SizeOf(outFont), 0);
  for ch := #0 to #255 do BuildCharOutline(ch);
end;


procedure SDrawChar (x, y: Integer; ch: Char; fc: Byte; oc: Byte = 255);
var
  dx, dy, b: Integer;
  c: Byte;
begin
  Dec(x); Dec(y);
  for dy := 0 to 9 do
  begin
    for dx := 0 to 7 do
    begin
      b := outFont[ch, dy, dx];
      case b of
        1: c := fc;
        2: c := oc;
        else c := 255;
      end;
      if c <> 255 then VPutPixel(x+dx, y, c);
    end;
    Inc(y);
  end;
end;

procedure SDrawStringPartSS (x, y: Integer; const s: ShortString; stPos, len: Integer;
  fc: Byte; oc: Byte = 255);
var
  realLen: Integer;
begin
  if (fc = 255) and (oc = 255) then exit;
  if stPos < 1 then stPos := 1;
  realLen := Length(s)-stPos+1;
  if realLen > len then realLen := len;
  if (realLen < 1) or (len < 1) then exit;
  while stPos <= realLen do
  begin
    SDrawChar(x, y, s[stPos], fc, oc);
    Inc(stPos); Inc(x, 6);
  end;
end;

procedure SDrawStringPart (x, y: Integer; const s: string; stPos, len: Integer;
  fc: Byte; oc: Byte = 255);
var
  realLen: Integer;
begin
  if (fc = 255) and (oc = 255) then exit;
  if stPos < 1 then stPos := 1;
  realLen := Length(s)-stPos+1;
  if realLen > len then realLen := len;
  if (realLen < 1) or (len < 1) then exit;
  while stPos <= realLen do
  begin
    SDrawChar(x, y, s[stPos], fc, oc);
    Inc(stPos); Inc(x, 6);
  end;
end;

procedure SDrawStringSS (x, y: Integer; const s: ShortString; fc: Byte; bc: Byte = 255);
begin
  SDrawStringPartSS(x, y, s, 1, Length(s), fc, bc);
end;

procedure SDrawString (x, y: Integer; const s: string; fc: Byte; bc: Byte = 255);
begin
  SDrawStringPart(x, y, s, 1, Length(s), fc, bc);
end;

procedure DrawText (x, y: Integer; const s: string; fc, bc: Byte);
begin
  VDrawString(x, y, s, fc, bc);
end;

procedure DrawTextOutline (x, y: Integer; const s: string; tc, oc: Byte);
{var
  dx, dy: Integer;}
begin
  {if oc <> 255 then
  begin
    for dx := -1 to 1 do
      for dy := -1 to 1 do
        if not ((dx = 0) and (dy = 0)) then VDrawString(x+dx, y+dy, s, oc, 255);
    if tc = 255 then VDrawString(x, y, s, oc, 255);
  end;
  if tc <> 255 then VDrawString(x, y, s, tc, 255);}
  SDrawString(x, y, s, tc, oc);
end;

{$INCLUDE conlogo.inc}

procedure LoadConLogo ();
var
  sz: Integer;
  w, h: Word;
  cl: PChar;
begin
  Move(conLogoWDP[0], w, 2); Move(conLogoWDP[2], h, 2);
  sz := w*h+4;
  ReallocMem(conLogo, sz); cl := Pointer(conLogo);
  WDXUnpack(conLogoWDP[4], SizeOf(conLogoWDP)-4, cl[4], sz-4);
  Move(conLogoWDP[0], conLogo^, 4);
  //ReallocMem(conLogo, 0);
  {rs := TResourceStream.Create(hInstance, 'CONLOGO', RT_RCDATA);
  try
    ReallocMem(conLogo, rs.Size);
    rs.ReadBuffer(conLogo^, rs.Size);
  finally
    rs.Free();
  end;}
end;

procedure PCJoyReset ();
var
  c: Integer;
  ji: TJoyInfo;
  jc: TJoyCaps;
begin
  FillChar(pcJoyXCen, SizeOf(pcJoyXCen), 0);
  FillChar(pcJoyYCen, SizeOf(pcJoyYCen), 0);
  FillChar(pcJoyXMinMax, SizeOf(pcJoyXMinMax), 0);
  FillChar(pcJoyYMinMax, SizeOf(pcJoyYMinMax), 0);
  pcJoyCount := joyGetNumDevs();
  pcJoyActive := 0;
  if pcJoyCount > 2 then pcJoyCount := 2;
  if joyDebugInfo then
  begin
    WriteLn('======PCJoyReset======');
    WriteLn('devices: ', pcJoyCount);
  end;
  for c := 0 to pcJoyCount-1 do
  begin
    if joyGetPos(c, @ji) = MMSYSERR_NOERROR then
    begin
      if joyDebugInfo then
      begin
        WriteLn(' joy', c, ': xcen=', ji.wXPos, '; ycen=', ji.wYPos);
      end;
      pcJoyXCen[c] := ji.wXPos;
      pcJoyYCen[c] := ji.wYPos;
      pcJoyActive := pcJoyActive or (1 shl c);
    end
    else continue;
    if joyGetDevCaps(c, @jc, SizeOf(jc)) = MMSYSERR_NOERROR then
    begin
      pcJoyXMinMax[c, 0] := jc.wXMin; pcJoyXMinMax[c, 1] := jc.wXMax;
      pcJoyYMinMax[c, 0] := jc.wYMin; pcJoyYMinMax[c, 1] := jc.wYMax;
      if joyDebugInfo then
      begin
        WriteLn(' joy', c, ': xmin=', jc.wXMin, '; xmax=', jc.wXMax);
        WriteLn(' joy', c, ': ymin=', jc.wYMin, '; ymax=', jc.wYMax);
      end;
    end
    else pcJoyActive := pcJoyActive and (not (1 shl c));
  end;
  if joyDebugInfo then WriteLn('======PCJoyReset======');
  PCJoyScan();
end;

procedure PCJoyScan ();
const
  joySens = 0.5;
var
  c, nx, ny, t, x, y: Integer;
  ts: Single;
  ji: TJoyInfo;
begin
  FillChar(pcJoys, SizeOf(pcJoys), 0);
  if not config.usePCJoystick then exit;
  for c := 0 to pcJoyCount-1 do
  begin
    if (pcJoyActive and (1 shl c)) = 0 then continue;
    nx := pcJoyXMinMax[c, 1]-pcJoyXMinMax[c, 0];
    if nx <= 0 then continue;
    ny := pcJoyYMinMax[c, 1]-pcJoyYMinMax[c, 0];
    if ny <= 0 then continue;
    if joyGetPos(c, @ji) = MMSYSERR_NOERROR then
    begin
      if (ji.wButtons and JOY_BUTTON1) <> 0 then pcJoys[c, 4] := true;
      if (ji.wButtons and JOY_BUTTON2) <> 0 then pcJoys[c, 5] := true;
      if (ji.wButtons and JOY_BUTTON3) <> 0 then pcJoys[c, 6] := true;
      if (ji.wButtons and JOY_BUTTON4) <> 0 then pcJoys[c, 7] := true;
      x := ji.wXPos; y := ji.wYPos;
      Dec(x, pcJoyXCen[c]); Dec(y, pcJoyYCen[c]);

      if joyDebugInfo and ((x <> 0) or (y <> 0)) then
      begin
        WriteLn('======PCJoyScan======');
        WriteLn(' joy', c, ': x=', x, '; y=', y);
        WriteLn('======PCJoyScan======');
      end;
      // left?
      if x < 0 then
      begin
        t := pcJoyXCen[c]-pcJoyXMinMax[c, 0]; if t < 1 then ts := 0.0
        else ts := (-x)/t;
        if ts >= joySens then pcJoys[c, 0] := true;
      end;
      // right?
      if x > 0 then
      begin
        t := pcJoyXMinMax[c, 1]-pcJoyXCen[c]; if t < 1 then ts := 0.0
        else ts := x/t;
        if ts >= joySens then pcJoys[c, 1] := true;
      end;
      // up?
      if y < 0 then
      begin
        t := pcJoyYCen[c]-pcJoyYMinMax[c, 0]; if t < 1 then ts := 0.0
        else ts := (-y)/t;
        if ts >= joySens then pcJoys[c, 2] := true;
      end;
      // down?
      if y < 0 then
      begin
        t := pcJoyYMinMax[c, 1]-pcJoyYCen[c]; if t < 1 then ts := 0.0
        else ts := y/t;
        if ts >= joySens then pcJoys[c, 3] := true;
      end;
    end;
  end;
end;

procedure PCKeys2ZX ();
// partially taken from UnrealSpectrum
var
  done: Boolean;
  shifted, sc: Boolean;
  f, k, j: Integer;
begin
  if zxWinList.Count > 0 then exit; //!

  // clear keyboard masks
  emuKeybMask[0] := $FF; emuKeybMask[1] := $FF; emuKeybMask[2] := $FF;
  emuKeybMask[3] := $FF; emuKeybMask[4] := $FF; emuKeybMask[5] := $FF;
  emuKeybMask[6] := $FF; emuKeybMask[7] := $FF;

  if emuKeysLocked then begin emuJoyMask := 0; exit; end;
  emuJoyMask := $FF;

  shifted := pcKeys[VK_SHIFT] or pcKeys[VK_LSHIFT] or pcKeys[VK_RSHIFT];
  for f := 0 to 255 do
  begin
      // autofire
    if f = 0 then
    begin
      pcKeys[0] := config.autofiring and (emuAutofireDelay <= 0);
    end;

    if pcKeys[f] then
    begin
      sc := shifted; if shifted and (pcKeyMap[true, f].masks[0] = $FF) then sc := false;
      pcKeyMap[sc, f].ports[0]^ := pcKeyMap[sc, f].ports[0]^ and pcKeyMap[sc, f].masks[0];
      pcKeyMap[sc, f].ports[1]^ := pcKeyMap[sc, f].ports[1]^ and pcKeyMap[sc, f].masks[1];
    end;
  end;
  emuJoyMask := emuJoyMask xor $FF;

  if not config.emulateMatrixEffect then exit;
  // emulate keyboard matrix effect (100% of code from Unreal Spectrum)
  repeat
    done := true;
    for k := 0 to 6 do
    begin
      for j := k+1 to 7 do
      begin
        if ((emuKeybMask[k] or emuKeybMask[j]) <> $FF) and
           (emuKeybMask[k] <> emuKeybMask[j]) then
        begin
          emuKeybMask[j] := (emuKeybMask[k] and emuKeybMask[j]);
          emuKeybMask[k] := emuKeybMask[j];
          done := false;
        end;
      end;
    end;
  until done;
end;

procedure PCJoy2ZX ();
var
  f, c: Integer;
begin
  // clear keyboard masks
  {emuKeybMask[0] := $FF; emuKeybMask[1] := $FF; emuKeybMask[2] := $FF;
  emuKeybMask[3] := $FF; emuKeybMask[4] := $FF; emuKeybMask[5] := $FF;
  emuKeybMask[6] := $FF; emuKeybMask[7] := $FF;
  emuJoyMask := $FF;}
  if not config.usePCJoystick then exit;
  emuJoyMask := emuJoyMask xor $FF;
  for c := 0 to 1 do
  begin
    if (pcJoyActive and (1 shl c)) = 0 then continue;
    for f := 0 to 7 do
    begin
      if pcJoys[c, f] then
      begin
        pcJoyMap[c, f].ports[0]^ := pcJoyMap[c, f].ports[0]^ and pcJoyMap[c, f].masks[0];
        pcJoyMap[c, f].ports[1]^ := pcJoyMap[c, f].ports[1]^ and pcJoyMap[c, f].masks[1];
      end;
    end;
  end;
  emuJoyMask := emuJoyMask xor $FF;
end;

procedure DoPCMouse ();
var
  mp: TPoint;
  d: Integer;
  r: TRect;
begin
  GetCursorPos(mp);
  d := emuMousePos.x+(mp.x-winMousePos.x) div emuScale;
  d := d and $FF; emuMousePos.x := d;
  d := emuMousePos.y+(winMousePos.y-mp.y) div emuScale;
  d := d and $FF; emuMousePos.y := d;
  if emuMouseLocked and GetClipCursor(r) then
  begin
    Inc(r.Left, (r.Right-r.Left) div 2);
    Inc(r.Top, (r.Bottom-r.Top) div 2);
    SetCursorPos(r.Left, r.Top);
    GetCursorPos(winMousePos);
  end
  else winMousePos := mp;
end;

procedure EmuSetMouseXY (x, y: Integer);
var
  r: TRect;
begin
  GetWindowRect(emuHWnd, r);
  Inc(r.Left, GetSystemMetrics(SM_CXFRAME));
  Inc(r.Top, GetSystemMetrics(SM_CYFRAME)+GetSystemMetrics(SM_CYCAPTION));
  SetCursorPos(r.Left+x*emuScale, r.Top+y*emuScale);
  GetCursorPos(winMousePos);
  GetCursorPos(emuMousePos);
  DoPCMouse();
end;

procedure EmuFixShiftsState ();
begin
  emuShiftsState := 0;
  if (GetKeyState(VK_MENU) and $80) <> 0 then emuShiftsState := emuShiftsState or emuSSAlt;
  if (GetKeyState(VK_CONTROL) and $80) <> 0 then emuShiftsState := emuShiftsState or emuSSCtrl;
  if (GetKeyState(VK_SHIFT) and $80) <> 0 then emuShiftsState := emuShiftsState or emuSSShift;
end;

procedure EmuResetTimers ();
begin
  emuFPS := 0;
  {$IFDEF USE_HP_TIMER}
  HPTimerOn(emuHQCounter);
  HPTimerOn(emuFPSHQCounter);
  emuFPSHQCounterCounting := false;
  {$ELSE}
  emuHQCounter.Stop(); if not emuPaused and not emuTempPaused then emuHQCounter.Start();
  emuFPSHQCounter.Stop();
  {$ENDIF}
end;

procedure EmuPause (immed: Boolean = false);
begin
  if emuPaused then exit;
  emuFramesToSkip := -1;
  emuZ80.breaked := true;
  emuPaused := true;
  if emuSoundEnabled then EmuSoundDriverPause(true);
  EmuUnlockMouse();
  WriteLn('dsp paused');
  EmuResetTimers();
end;

procedure EmuUnpause ();
begin
  if not emuPaused and not emuTempPaused then exit;
  emuFramesToSkip := -1;
  if emuShowKeyHelp then emuBorderColor := emuSavedBorderColor;
  emuShowKeyHelp := false;
  emuPaused := false;
  emuTempPaused := false;
  if emuSoundEnabled then EmuSoundDriverPause(false);
  WriteLn('dsp unpaused');
  EmuResetTimers();
end;

procedure EmuTempPause ();
begin
  if emuTempPaused then exit;
  emuZ80.breaked := true;
  emuTempPaused := true;
  if emuSoundEnabled then EmuSoundDriverPause(true);
  EmuUnlockMouse();
  if not emuPaused then WriteLn('dsp temp paused');
  EmuResetTimers();
end;

procedure EmuTempUnpause ();
begin
  if not emuTempPaused then exit;
  emuTempPaused := false;
  if not emuPaused and emuSoundEnabled then EmuSoundDriverPause(false);
  if not emuPaused then WriteLn('dsp temp unpaused');
  EmuResetTimers();
end;

procedure EmuDrawScreen ();
var
  sve: packed array [0..9, 0..31] of Byte;
  f: Integer;
  rest: Boolean;
  s: string;
  tm: TSystemTime;
begin
  if (config.showClock = 2) or
     ((config.showClock = 1) and emuFullscreen) then
  begin
    rest := true;

    GetLocalTime(tm);
    Str(tm.wHour:2, s);
    if tm.wSecond mod 2 = 0 then s := s+' ' else s := s+':';
    if tm.wMinute < 10 then s := s+'0';
    s := s+IntToStr(tm.wMinute);

    // save screen
    for f := 1 to 10 do Move(dspVScreenRam[f, 287], sve[f-1], 32);
    DrawTextOutline(288, 2, s, 4, 0);
  end
  else rest := false;
  if emuScale = 1 then VRepaint() else VRepaint2x();
  if rest then for f := 1 to 10 do Move(sve[f-1], dspVScreenRam[f, 287], 32);
end;

procedure EmuInfoMessage (const msg: string; const timeOut: LongWord = 2000);
begin
  emuInfoMsgHideTime := GetTickCount()+timeOut;
  if length(msg) > 53 then emuInfoMsg := Copy(msg, 1, 50)+'...' else emuInfoMsg := msg;
  emuInfoMsgX := (320-Length(emuInfoMsg)*6) div 2;
end;


procedure EmuRenderScreenNoInv ();
var
  y: Integer;
begin
  VResetClip();
  if emuShowKeyHelp then
  begin
    for y := 0 to 239 do ZXUpdateLine(y, @(emuKeyScreen[0]));
  end
  else
  begin
    for y := 0 to 239 do ZXUpdateLine(y, @(emuMemory[emuScreenPage]));
    if emuTempPaused or emuPaused then DrawTextOutline(4, 230, 'paused', 15, 0);
  end;
  VResetOrigin(); VResetClip();
  ZXWindowsDraw();
  VResetOrigin(); VResetClip();
  EmuDrawWCon();
  if emuInfoMsgHideTime > GetTickCount() then
    DrawTextOutline(emuInfoMsgX, emuInfoMsgY, emuInfoMsg, emuInfoMsgColor, emuInfoMsgOutColor);
end;

procedure EmuRenderScreen ();
begin
  EmuRenderScreenNoInv();
  InvalidateRect(emuHWnd, nil, false);
end;

procedure EmuRealizeState (state: PDZSZ80State);
var
  f, c: Integer;
  blk: PDZSMemBlock;
begin
  if state.model <= dzs48k then EmuSetModel(zxm48k)
  else if state.model = dzsPentagon then EmuSetModel(zxmPentagon128k)
  else EmuSetModel(zxm128k);

  emuZ80.im := state.im;
  //WriteLn('im: ', emuZ80.im);
  emuZ80.iff1 := state.iff1;
  emuZ80.iff2 := state.iff2;
  //WriteLn('iff1: ', emuZ80.iff1);
  //WriteLn('iff2: ', emuZ80.iff2);

  Z80SetAF(emuZ80, state.regs.af[0].w);
  Z80SetBC(emuZ80, state.regs.bc[0].w);
  emuZ80.regDE := state.regs.de[0].w;
  emuZ80.regHL := state.regs.hl[0].w;
  emuZ80.regAFx := state.regs.af[1].w;
  emuZ80.regBCx := state.regs.bc[1].w;
  emuZ80.regDEx := state.regs.de[1].w;
  emuZ80.regHLx := state.regs.hl[1].w;
  emuZ80.regIX := state.regs.ix.w;
  emuZ80.regIY := state.regs.iy.w;
  emuZ80.regSP := state.regs.sp.w;
  emuZ80.regPC := state.regs.pc.w;
  emuZ80.regI := state.regs.ir.i;
  emuZ80.regR := state.regs.ir.r and $80;
  emuZ80.regRTemp := state.regs.ir.r and $7F;
  emuZ80.halted := state.halted;

  for f := 0 to 15 do emuAY.Registers[f] := state.psg[f];
  emuAYRegIndex := state.lastFFFD;
  emuAYEnabled := state.ayType <> dzsNoAY;

  //WriteLn('sp=', emuZ80.regSP);
  //WriteLn('pc=', emuZ80.regPC);

  if state.last7FFD <> $FF then Z80PortOut(emuZ80, $7FFD, state.last7FFD);
  Z80PortOut(emuZ80, $FE, state.border);
  emuZ80.tStates := 0;

  if state.trDosIn then emuPaging[0] := 10;

  for f := 0 to 7 do
  begin
    blk := DZSFindPage(state, f);
    FillChar(emuMemory[f], 16384, 0);
    if (blk = nil) or (Length(blk.data) < 1) then continue;
    c := Length(blk.data); if c > 16384 then c := 16384;
    Move(blk.data[0], emuMemory[f], c);
  end;

  if config.pent512 then
  begin
    for f := 8 to 31 do
    begin
      blk := DZSFindPage(state, f);
      FillChar(emuMemory[f+3], 16384, 0);
      if (blk = nil) or (Length(blk.data) < 1) then continue;
      c := Length(blk.data); if c > 16384 then c := 16384;
      Move(blk.data[0], emuMemory[f+3], c);
    end;
  end;

  EmuRenderScreen();
  EmuDrawScreen();//VRepaint();
end;

function EmuGetState (): PDZSZ80State;
var
  f: Integer;
begin
  result := DZSNewStateRec();
  try
    if emuModel = zxm48k then result.model := dzs48k
    else if emuModel = zxmPentagon128k then result.model := dzsPentagon
    else result.model := dzs128k;

    result.im := emuZ80.im;
    result.iff1 := emuZ80.iff1;
    result.iff2 := emuZ80.iff2;

    result.regs.af[0].w := Z80GetAF(emuZ80);
    result.regs.bc[0].w := Z80GetBC(emuZ80);
    result.regs.de[0].w := emuZ80.regDE;
    result.regs.hl[0].w := emuZ80.regHL;
    result.regs.af[1].w := emuZ80.regAFx;
    result.regs.bc[1].w := emuZ80.regBCx;
    result.regs.de[1].w := emuZ80.regDEx;
    result.regs.hl[1].w := emuZ80.regHLx;
    result.regs.ix.w := emuZ80.regIX;
    result.regs.iy.w := emuZ80.regIY;
    result.regs.sp.w := emuZ80.regSP;
    result.regs.pc.w := emuZ80.regPC;
    result.regs.ir.i := emuZ80.regI;
    result.regs.ir.r := (emuZ80.regR and $80) or (emuZ80.regRTemp and $7F);

    for f := 0 to 15 do result.psg[f] := emuAY.Registers[f];
    result.lastFFFD := emuAYRegIndex;
    if emuAYEnabled then result.ayType := dzsAY else result.ayType := dzsNoAY;

    result.last7FFD := emuLastOut7FFD;
    result.border := emuBorderColor and $07;
    result.tStates := 0;

    result.trDosIn := (emuPaging[0] = 10);

    result.last1FFD := 0;

    result.halted := emuZ80.halted;
    result.lastWasEI := false;  //!!!FIXME

    for f := 0 to 7 do DZSNewPageWithData(result, f, emuMemory[f], 16384);
    if config.pent512 then
    begin
      for f := 8 to 31 do DZSNewPageWithData(result, f, emuMemory[f+3], 16384);
    end;
  except
    DZSFreeStateRec(result);
    raise;
  end;
end;

procedure EmuDrawTapeCounter (x, y: Integer);
const
  maxT = 9999*8;
var
  f, pos, max: Integer;
  digs, dofs: array [0..3] of Integer;

  procedure DrawDigit (x, y, d, ofs: Integer);
  var
    dx, dy: Integer;
    b: Byte;
  begin
    y := 239-y; Inc(ofs, d*8);
    for dy := 0 to 7 do
    begin
      b := cntNumz[ofs]; Inc(ofs);
      for dx := 0 to 7 do
      begin
        if (b and $80) <> 0 then dspVScreenRam[y, x] := 15
        else dspVScreenRam [y, x] := 0;
        b := b shl 1;
        Inc(x);
      end;
      Dec(x, 8); Dec(y);
    end;
  end;

begin
  max := emuTapeInfo.tape.Count;
  if max < 1 then pos := 0
  else
  begin
    pos := max-emuTapeInfo.tape.Position;
    pos := Int64(pos)*maxT div max;
    if pos > maxT then pos := maxT;
  end;
  if pos = 0 then exit;

  VHLine(x, y, 4*8+3, 15);
  VHLine(x, y+11, 4*8+3, 15);
  VVLine(x, y+1, 10, 15);
  VVLine(x+4*8+2, y+1, 10, 15);

  VHLine(x+1, y+1, 4*8+1, 0);
  VHLine(x+1, y+10, 4*8+1, 0);
  VVLine(x+1, y+2, 8, 0);
  //VVLine(x+4*8+1, y+2, 8, 0);
  Inc(x, 2); Inc(y, 2);

  //Write('pos=', pos, '; ');

  dofs[3] := pos mod 8; // digit offset
  pos := pos div 8;
  digs[3] := pos mod 10;  // digit number
  pos := pos div 10;
  dofs[2] := 0;
  digs[2] := pos mod 10;  // digit number
  pos := pos div 10;
  dofs[1] := 0;
  digs[1] := pos mod 10;  // digit number
  pos := pos div 10;
  dofs[0] := 0;
  digs[0] := pos mod 10;  // digit number

  if digs[3] = 9 then dofs[2] := dofs[3];
  if digs[2] = 9 then dofs[1] := dofs[2];
  if digs[1] = 9 then dofs[0] := dofs[1];

  for f := 0 to 3 do
  begin
    //Write(f, ':', digs[f], '(', dofs[f], ')');
    //if f <> 3 then Write('; ');
    DrawDigit(x, y, digs[f], dofs[f]);
    Inc(x, 8);
  end;
  //WriteLn;
end;

procedure EmuRewindTape ();
begin
  EmuStopTape(false);
  WriteLn('rewinding tape...');
  emuTapeInfo.ticksPassed := 0;
  emuTapeInfo.tape.Rewind();
end;

function EmuTapeBlockCount (): Integer;
var
  f: Integer;
begin
  result := 0;
  for f := 0 to emuTapeInfo.tape.Count-1 do
  begin
    case emuTapeInfo.tape[f] of
      -660, -3, -2, -1: Inc(result);
    end;
  end;
end;

function EmuTapeCurBlock (): Integer;
var
  f, c: Integer;
begin
  result := 1; f := 0;
  while f < emuTapeInfo.tape.Position do
  begin
    c := emuTapeInfo.tape[f];
    case c of
      -660, -3, -2, -1: Inc(result);
    end;
    Inc(f);
  end;
end;

procedure EmuTapeGoToBlock (num: Integer);
var
  c: Integer;
begin
  EmuStopTape(false);
  WriteLn('going to the block ', num);
  emuTapeInfo.ticksPassed := 0;
  emuTapeInfo.tape.Rewind();
  //Dec(num);
  //if num < 1 then exit;
  while (num >= 0) and (emuTapeInfo.tape.Position < emuTapeInfo.tape.Count) do
  begin
    c := emuTapeInfo.tape.Current;
    case c of
      -660, -3, -2, -1:
        begin
          Dec(num);
          //if num <= 1 then break;
          //WriteLn(num, ': ', emuTapeInfo.tape.Position);
        end;
    end;
    if c > 0 then Inc(emuTapeInfo.ticksPassed, c);
    emuTapeInfo.tape.Next();
  end;
end;

procedure EmuTapeSkipBlocks (num: Integer);
begin
  num := EmuTapeCurBlock()+num;
  EmuTapeGoToBlock(num);
end;

procedure EmuStopTape (user: Boolean = true);
begin
  if not emuTapeInfo.playing then exit;
  emuTapeInfo.playing := false;
  emuTapeInfo.ticksLeft := emuZ80.tStates-emuTapeInfo.nextEdgeT;
  if emuTapeInfo.ticksLeft < 0 then emuTapeInfo.ticksLeft := 0;
  WriteLn('tape stopped...');
  if not user and config.tapeAutopause then ConExecute('pause tan');
  if config.tapeAutomax then ConExecute('speed max off');
end;

procedure EmuStartTape (user: Boolean = false);
begin
  if emuTapeInfo.playing then exit;
  emuTapeInfo.nextEdgeT := emuZ80.tStates+emuTapeInfo.ticksLeft+16; //!!! 16
  if emuTapeInfo.nextEdgeT < emuZ80.tStates then emuTapeInfo.nextEdgeT := emuZ80.tStates;
  emuTapeInfo.ticksLeft := 0;
  emuTapeInfo.playing := true;
  WriteLn('tape started...');
  if not user and config.tapeAutomax then ConExecute('speed max on');
end;

function EmuTapeBit (): Byte;
var
  c, pp: Integer;
begin
  result := emuTapeInfo.tapeBit;
  if not emuTapeInfo.playing then exit;

  while (emuTapeInfo.nextEdgeT < emuZ80.tStates) and emuTapeInfo.playing do
  begin
    //c := emuTapeInfo.tape.Current; emuTapeInfo.tape.Next();
    pp := emuTapeInfo.tape.Position;
    c := emuTapeInfo.tape.ReadPulse();
    case c of
      -667: WriteLn('tape: pause block');
      -666: WriteLn('tape: standard block');
      -665: WriteLn('tape: direct block');
      -664: WriteLn('tape: pure data block');
      -663: WriteLn('tape: pulses block');
      -662: WriteLn('tape: pure tone block');
      -661: WriteLn('tape: turbo block');
      -3: WriteLn('tape: stop if 128K');
      -2: WriteLn('tape: stop');
      -1: WriteLn('tape: stop if 48K');
    end;

    if c <= -1000 then
    begin
      c := (-c)-1000;
      if c < emuTapeInfo.tape.Messages.Count then
      begin
        WriteLn('tape msg: ', emuTapeInfo.tape.Messages[c]);
        EmuInfoMessage(emuTapeInfo.tape.Messages[c], 4000);
      end;
      continue;
    end;

    if (c = -2) or ((c = -1) and (emuModel = zxm48k)) or
       ((c = -3) and (emuModel <> zxm48k)) then
    begin
      emuTapeInfo.tapeBit := not emuTapeInfo.tapeBit; //!!! <SMT
      EmuStopTape(false);
      emuTapeInfo.playing := false;
      break;
    end;

    if c < 0 then continue; // skip a shit %-)

    emuTapeInfo.tapeBit := not emuTapeInfo.tapeBit;

    Inc(emuTapeInfo.nextEdgeT, c);
    Inc(emuTapeInfo.ticksPassed, c);

    if config.tapePrepause and config.tapeAutopause and
       (pp <> emuTapeInfo.tape.Position) and
       (emuTapeInfo.tape.Position >= emuTapeInfo.tape.Count) then
    begin
      {if config.tapeAutopause then }ConExecute('pause tan');
      if config.tapeAutomax then ConExecute('speed max off');
    end;
  end;

  result := emuTapeInfo.tapeBit;
end;


{$INCLUDE keyhelpscr.inc}

procedure LoadKeys ();
begin
  WDXUnpack(keyHelpScr[0], SizeOf(keyHelpScr), emuKeyScreen[0], 6912);
end;

function XStrAnsiToOem (const s: ShortString): ShortString;
begin
  result := s;
  if result <> '' then
  begin
    //UniqueString(result);
    CharToOemBuff(@(result[1]), @(result[1]), Length(result));
  end;
end;

procedure EmuDrawWCon ();
var
  f, y, h, lout: Integer;
  t: LongWord;
  ll: Int64;
  slideDir: Integer;
  stP: Integer;
  //cx, cy, dx, dy, wx, wy: Integer;
  //pb: PByte;
begin
  if conScrollDelta > wconBufSize+40 then conScrollDelta := wconBufSize+40;
  if conScrollDelta < 0 then conScrollDelta := 0;

  // check if something new appeared on the console
  if wconOutputCount > 0 then
  begin
    conScrollDelta := 0;
    wconOutputCount := 0;
    Inc(emuWConDumpLinesLeft, wconScrollCount); wconScrollCount := 0;

    if emuWConDumpLinesLeft < 1 then emuWConDumpLinesLeft := 1
    else if emuWConDumpLinesLeft > config.conLinesToDump then
      emuWConDumpLinesLeft := config.conLinesToDump;

    emuWConDumpScroll := GetTickCount()+config.conDumpDelay;
  end;

  // slide the console?
  slideDir := 0;
  if emuWConVisible then
  begin
    if config.conSlide and not emuMaxSpeed then
    begin
      if emuWConHeight = 0 then emuWConNextSlide := GetTickCount();
      if (emuWConHeight > config.conHeight) then emuWConNextSlide := GetTickCount();
      if emuWConHeight < config.conHeight then slideDir := 1 else
      if emuWConHeight > config.conHeight then slideDir := -1;
    end
    else emuWConHeight := config.conHeight;
  end
  else
  begin
    if emuWConHeight > 0 then emuWConDumpLinesLeft := 0;
    if config.conSlide and not emuMaxSpeed then
    begin
      if emuWConHeight = config.conHeight then emuWConNextSlide := GetTickCount();
      if emuWConHeight > 0 then slideDir := -1;
    end
    else emuWConHeight := 0;
  end;

  // do sliding
  if slideDir <> 0 then
  begin
    t := GetTickCount();
    if t >= emuWConNextSlide then
    begin
      Dec(t, emuWConNextSlide);
      ll := (Int64(t) div config.conSlideDelay)+1;
      Inc(emuWConHeight, ll*slideDir);
      if emuWConHeight < 0 then emuWConHeight := 0
      else if slideDir = 1 then if emuWConHeight > config.conHeight then emuWConHeight := config.conHeight;
      f := config.conSlideDelay-(Int64(t) mod config.conSlideDelay);
      Inc(t, emuWConNextSlide+LongWord(f));
      emuWConNextSlide := t;
    end;
  end;

  if emuWConHeight > 0 then
  begin
    y := 0; h := emuWConHeight*8+8;
    if emuWConHeight < config.conHeight then begin Dec(y, 2); Inc(h, 4); end;

    // draw console back
    VFillRect(0, y, 320, h, 255, 0);
    VFillRect(1, y+1, 318, h-2, 255, 12);
    VFillRect(2, y+2, 316, h-4, 0, 0);

    // draw console logo
    (*if (conLogo <> nil) and (PWord(Integer(conLogo)+2)^ < h-16) then
    begin
      pb := conLogo;
      wx := PWord(pb)^; Inc(pb, 2);
      wy := PWord(pb)^; Inc(pb, 2);
      cx := (320-wx) div 2; cy := (h-wy) div 2;
      for dy := 0 to wy-1 do
      begin
        for dx := 0 to wx-1 do
        begin
          if pb^ <> 0 then VPutPixel(cx+dx, cy+dy, 24);
          Inc(pb);
        end;
      end;
    end;*)

    f := wconBufSize-emuWConHeight+2; if f < 0 then f := 0;
    lout := emuWConHeight-1;

    // draw input line
    stP := Length(emuConInputLine)-50;
    if stP < 1 then stP := 1;
    VDrawStringPart(4, 4+(wconBufSize-f+1)*8, VStrAnsiToOem(emuConInputLine), stP,
      51, 7, 255);
    if stP > 1 then
      VDrawChar(4, 4+(wconBufSize-f+1)*8, #17, 14, 0);
    stP := Length(emuConInputLine)-stP+1;
    if GetTickCount() mod 1000 < 500 then
      VDrawChar(4+stP*6, 4+(wconBufSize-f+1)*8, #249, 15, 255);

    Dec(f, conScrollDelta); if f < 0 then f := 0;
  end
  else if config.conDumpLastLines and (emuWConDumpLinesLeft > 0) then
  begin
    // do scroll if necessary
    t := GetTickCount();
    if t >= emuWConDumpScroll then
    begin
      Dec(t, emuWConDumpScroll);
      ll := (Int64(t) div config.conDumpDelay)+1;
      if ll > emuWConDumpLinesLeft then ll := emuWConDumpLinesLeft;
      Dec(emuWConDumpLinesLeft, ll);
      if emuWConDumpLinesLeft = 0 then exit;
      f := config.conDumpDelay-(Int64(t) mod config.conDumpDelay);
      Inc(t, emuWConDumpScroll+LongWord(f));
      emuWConDumpScroll := t;
    end;
    f := wconBufSize-emuWConDumpLinesLeft+1; if f < 0 then f := 0;
    lout := emuWConDumpLinesLeft;
  end
  else begin f := maxInt; lout := 0; end;

  // show console text
  y := 4;
  while (f <= wconBufSize) and (lout > 0) do
  begin
    if emuWConHeight > 0 then VDrawString(4, y, XStrAnsiToOem(wconBuf[f]), 4, 255)
    else DrawTextOutline(4, y, XStrAnsiToOem(wconBuf[f]), 4, 0);
    //SDrawStringSS(4, y, XStrAnsiToOem(wconBuf[f]), 4, 0);
    Inc(f);
    Inc(y, 8);
    Dec(lout);
  end;
end;


procedure CmdPause (const cmd: string; const args: array of string);
var
  newP: Boolean;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('pause or resume Z80.');
    WriteLn('usage: ', cmd, ' [tan [imm]|ona]');
    exit;
  end;

  if cmd = 'pause' then
  begin
    newP := ConGetArgBoolDef(1, args, not emuPaused);
    if newP <> emuPaused then if newP then EmuPause(Length(args) > 1) else EmuUnpause();
    emuPaused := newP;
  end
  else
  begin
    newP := ConGetArgBoolDef(1, args, not emuTempPaused);
    if newP <> emuTempPaused then if newP then EmuTempPause() else EmuTempUnpause();
    emuTempPaused := newP;
  end;
end;

procedure CmdDefocusPause (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable defocus pausing.');
    Write('usage: defocus_pause [tan|ona]');
    exit;
  end;

  config.pauseOnDefocus := ConGetArgBoolDef(1, args, not config.pauseOnDefocus);
  WriteLn('defocus pausing: ', ConYesNoName(config.pauseOnDefocus));
end;

procedure CmdTapeAutostart (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable tape autostarting.');
    Write('usage: tape_autostart [tan|ona]');
    exit;
  end;

  config.tapeAutostart := ConGetArgBoolDef(1, args, not config.tapeAutostart);
  WriteLn('tape autostart: ', ConYesNoName(config.tapeAutostart));
end;

procedure CmdTapeAutopause (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable autopause after the tape stops.');
    WriteLn('usage: tape_autopause [tan|ona]');
    WriteLn(#0#1#1'this unique feature allows you to take snapshot of the game just after it loads.');
    exit;
  end;

  config.tapeAutopause := ConGetArgBoolDef(1, args, not config.tapeAutopause);
  WriteLn('tape autopause: ', ConYesNoName(config.tapeAutopause));
end;

procedure CmdTapeAutomax (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'enable/disable maxspeed mode when the tape is playing.');
    Write('usage: tape_automax [tan|ona]');
    exit;
  end;

  config.tapeAutomax := ConGetArgBoolDef(1, args, not config.tapeAutomax);
  WriteLn('tape automax: ', ConYesNoName(config.tapeAutomax));
end;

procedure CmdTapePrepause (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable pause before the tape stops.');
    WriteLn('usage: tape_prepause [tan|ona]');
    WriteLn('works only in tape_autopause mode.');
    exit;
  end;

  config.tapePrepause := ConGetArgBoolDef(1, args, not config.tapePrepause);
  WriteLn('tape prepause: ', ConYesNoName(config.tapePrepause));
end;

procedure CmdMALock (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable mouse autolocking.');
    WriteLn('usage: mouse_autolock [tan|ona]');
    WriteLn(#0#1#1'DSP can automatically lock the mouse when Speccy program uses it.');
    exit;
  end;

  config.autolockMouse := ConGetArgBoolDef(1, args, not config.autolockMouse);
  WriteLn('mouse autolocking: ', ConYesNoName(config.autolockMouse));
end;

procedure CmdMsSwap (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('swaps mouse buttons.');
    WriteLn('usage: kmouse_bswap [tan|ona]');
    exit;
  end;

  config.mouseSwap := ConGetArgBoolDef(1, args, not config.mouseSwap);
  WriteLn('mouse swapping: ', ConYesNoName(config.mouseSwap));
end;

procedure CmdAutoboot (const cmd: string; const args: array of string);
var
  s: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'enable/disable automatic boot adding to the boot-less disk images.');
    WriteLn('usage: autoboot scl|trd [tan|ona]');
    exit;
  end;

  s := ConGetArgStr(1, args);
  if s = '' then
  begin
    WriteLn('wtf?');
    exit;
  end;

  if s = 'scl' then
  begin
    config.autobootSCL := ConGetArgBoolDef(2, args, not config.autobootSCL);
    WriteLn('SCL autobooting: ', ConYesNoName(config.autobootSCL));
  end
  else if s = 'trd' then
  begin
    config.autobootTRD := ConGetArgBoolDef(2, args, not config.autobootTRD);
    WriteLn('TRD autobooting: ', ConYesNoName(config.autobootTRD));
  end
  else WriteLn('what the fuck is ', s, '?');

  sclAutoboot := config.autobootSCL;
  trdAutoboot := config.autobootTRD;
end;

procedure CmdBootFile (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set the boot file to use by autoboot feature.');
    WriteLn('usage: boot_file [filename]');
    WriteLn(#0#1#1'file must be HoBeta. when "filename" is not specified, DSP will use it''s default one-sector boot.');
    exit;
  end;

  config.defaultBoot := ConGetArgStr(1, args);
  if config.defaultBoot = '' then
  begin
    WriteLn('using default boot');
  end
  else
  begin
    WriteLn('boot: "', config.defaultBoot, '"');
  end;

  bootName := config.defaultBoot;
end;

procedure CmdLogging (const cmd: string; const args: array of string);
var
  s: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable logging. set log file.');
    WriteLn('usage: logging tan|ona [filename]');
    exit;
  end;

  if ConGetArgStr(1, args) = '' then
  begin
    WriteLn('wtf?');
    exit;
  end;

  if ConGetArgStr(2, args) <> '' then
  begin
    s := ConGetArgStr(2, args);
    if IsSpecialFile(s) then
    begin
      WriteLn('smart guy, eh?');
      exit;
    end;

    wconLogFile := s;
    WriteLn('using log file "', wconLogFile, '"');
  end;

  wconLogging := ConGetArgBoolDef(1, args, false);
  WriteLn('logging: ', ConYesNoName(wconLogging));
end;

procedure CmdRom (const cmd: string; const args: array of string);
var
  t, fn: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set corresponding rom file.');
    WriteLn('usage: rom type filename');
    WriteLn(' types:');
    WriteLn('  48k');
    WriteLn('  128k_0');
    WriteLn('  128k_1');
    WriteLn('  tr-dos');
    exit;
  end;

  t := ConGetArgStr(1, args); fn := ConGetArgStr(2, args);
  if (t = '') or (fn = '') then
  begin
    WriteLn('wtf?');
    exit;
  end;

  if Copy(t, 1, 2) = '48' then
  begin
    config.rom48k := fn;
    WriteLn('48k rom: "', config.rom48k, '"');
  end
  else if Copy(t, 1, 2) = 'tr' then
  begin
    config.romtrdos := fn;
    WriteLn('tr-dos rom: "', config.romtrdos, '"');
  end
  else if (t = '128k_0') or (t = '128_0') or (t = '128k0') then
  begin
    config.rom128k_0 := fn;
    WriteLn('128k rom 0: "', config.rom128k_0, '"');
  end
  else if (t = '128k_1') or (t = '128_1') or (t = '128k1') then
  begin
    config.rom128k_1 := fn;
    WriteLn('128k rom 1: "', config.rom128k_1, '"');
  end
  else WriteLn('what the shit is ', t, '?');
end;

procedure CmdConSlide (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable console sliding.');
    WriteLn('usage: con_slide [tan|ona]');
    exit;
  end;

  config.conSlide := ConGetArgBoolDef(1, args, not config.conSlide);
  WriteLn('console sliding: ', ConYesNoName(config.conSlide));
end;

procedure CmdConSlideDelay (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set console sliding delay.');
    WriteLn('usage: con_slide_delay milliseconds');
    exit;
  end;

  if not ConGetArgIntEx(1, args, f) then begin WriteLn('wtf?'); exit; end;
  if f < 5 then f := 5 else if f > 1500 then f := 1500;
  config.conSlideDelay := f;
  WriteLn('console sliding delay: ', config.conSlideDelay);
end;

procedure CmdConHeight (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set console height.');
    WriteLn('usage: con_height lines');
    exit;
  end;

  if not ConGetArgIntEx(1, args, f) then begin WriteLn('wtf?'); exit; end;
  if f < 2 then f := 2;
  config.conHeight := f;
  if config.conHeight > wconMaxHeight then config.conHeight := wconMaxHeight;
  WriteLn('console height: ', config.conHeight);
end;

procedure CmdConDumpLines (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'set number of visible lines when console is hidden.');
    WriteLn('usage: con_dump_lines lines');
    exit;
  end;

  if ConGetArgCount(args) < 1 then begin WriteLn('wtf?'); exit; end;

  f := ConGetArgIntDef(1, args, 0);
  if f < 1 then
  begin
    config.conDumpLastLines := false;
    WriteLn('dump console lines: ona');
  end
  else
  begin
    config.conDumpLastLines := true;
    if f > 18 then f := 18;
    config.conLinesToDump := f;
    WriteLn('dump console lines: ', config.conLinesToDump);
  end;
end;

procedure CmdConDumpDelay (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set console dump delay.');
    WriteLn('usage: con_dump_delay milliseconds');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    WriteLn('wtf?');
    exit;
  end;
  f := StrToIntDef(args[0], -1);
  if f < 50 then f := 50 else if f > 3500 then f := 3500;
  config.conDumpDelay := f;
  WriteLn('console dump delay: ', config.conDumpDelay);
end;

procedure CmdLogTime (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('should DSP write current time on each log line?');
    WriteLn('usage: log_time [tan|ona]');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    wconLogDateTime := not wconLogDateTime;
  end
  else wconLogDateTime := ConYesNoValue(args[0]);
  WriteLn('log time: ', ConYesNoName(wconLogDateTime));
end;

procedure CmdConsole (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('show/hide console.');
    WriteLn('usage: console [tan|ona]');
    exit;
  end;

  emuWConVisible := ConGetArgBoolDef(1, args, not emuWConVisible);
end;

procedure CmdSound (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable sound.');
    WriteLn('usage: sound [tan|ona]');
    exit;
  end;

  emuSoundEnabled := ConGetArgBoolDef(1, args, not emuSoundEnabled);
  EmuSoundDriverPause(not emuSoundEnabled);
  WriteLn('sound: ', ConYesNoName(emuSoundEnabled));
end;

procedure CmdSoundAY (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable AY-8910/12 emulation.');
    WriteLn('usage: sound_ay [tan|ona]');
    exit;
  end;

  emuAYEnabled := ConGetArgBoolDef(1, args, not emuAYEnabled);
  WriteLn('sound_ay: ', ConYesNoName(emuAYEnabled));
end;

procedure CmdSpeed (const cmd: string; const args: array of string);
var
  s: string;
  d: Double;
  e: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set emulation speed.');
    WriteLn('usage: speed mode [toggle|off]');
    WriteLn(#0#1#1'the second arg indicates that the speed mode must be either switched off or toggled.');
    WriteLn(' mode:');
    WriteLn('  normal -- speed of real Speccy');
    WriteLn('  max -- maximum speed');
    WriteLn('  <number> -- floating speed factor (2 is 2 times');
    WriteLn('    faster then a real Speccy, for exapmle)');
    exit;
  end;

  emuFramesToSkip := -1;
  s := ConGetArgStr(1, args);
  if s = 'normal' then
  begin
    emuMaxSpeed := false; emuSpeedFactor := 1.0;
    WriteLn('normal speed engaged.');
    EmuResetTimers();
  end
  else if s = 'max' then
  begin
    if ConGetArgStr(2, args) = 'off' then emuMaxSpeed := false
    else emuMaxSpeed := not emuMaxSpeed;

    if emuMaxSpeed then WriteLn('maximum speed mode engaged.')
    else WriteLn('maximum speed mode is swithed off.');
    EmuResetTimers();
  end
  else
  begin
    Val(s, d, e);
    if e <> 0 then begin WriteLn('incorrect speed factor'); exit; end;
    if d < 0.01 then d := 0.01 else if d > 8 then d := 8;
    emuSpeedFactor := d;
    Str(d:18:2, s); s := Trim(s);
    while s[Length(s)] = '0' do Delete(s, Length(s), 1);
    if s[Length(s)] = '.' then Delete(s, Length(s), 1);
    WriteLn('speed factor is ', Trim(s));
    EmuResetTimers();
  end;
end;

procedure CmdConAF (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable autofiring.');
    WriteLn('usage: autofire [tan|ona]');
    exit;
  end;

  config.autofiring := ConGetArgBoolDef(1, args, not config.autofiring);
  emuAutofireDelay := 1;
  WriteLn('autofire: ', ConYesNoName(config.autofiring));
end;

procedure CmdConAFRate (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set autofire rate in frames.');
    WriteLn('usage: autofire_rate frameno');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    WriteLn('wtf?');
    exit;
  end;
  f := StrToIntDef(args[0], -1);
  if f < 1 then f := 1 else if f > 1000 then f := 1000;
  config.autofireRate := f;
  emuAutofireDelay := 1;
  WriteLn('autofire rate: ', config.autofireRate);
end;

procedure CmdConAFHold (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'set number of frames to keep autofire pressed.');
    WriteLn('usage: autofire_hold frameno');
    exit;
  end;

  if (Length(args) < 1) or (args[0] = '') then
  begin
    WriteLn('wtf?');
    exit;
  end;
  f := StrToIntDef(args[0], -1);
  if f < 1 then f := 1 else if f > 1000 then f := 1000;
  config.autofireKeepPressed := f;
  emuAutofireDelay := 1;
  WriteLn('autofire holding: ', config.autofireKeepPressed);
end;

procedure CmdConClock (const cmd: string; const args: array of string);
var
  s: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'enable/disable displaying of the current time.');
    WriteLn('usage: clock tan|ona|fs');
    exit;
  end;

  s := ConGetArgStr(1, args);
  if s = 'fs' then config.showClock := 1
  else if ConGetArgBoolDef(1, args, config.showClock = 0) then config.showClock := 2
  else config.showClock := 0;
end;

procedure CmdJoyReset (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn(#0#1#1'reinit joysticks.');
    WriteLn('usage: joy-reset');
    exit;
  end;
  PCJoyReset();
end;

procedure CmdJoyDebugInfo (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable writing debug info.');
    WriteLn('usage: joy_debug_info [tan|ona]');
    exit;
  end;

  joyDebugInfo := ConGetArgBoolDef(1, args, not joyDebugInfo);
  WriteLn('joyDebugInfo: ', ConYesNoName(joyDebugInfo));
end;

procedure CmdFastTape (const cmd: string; const args: array of string);
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('enable/disable FastLoader(utm).');
    WriteLn('usage: fasttape [tan|ona]');
    exit;
  end;

  config.tapeFastLoader := ConGetArgBoolDef(1, args, not config.tapeFastLoader);
  WriteLn('FastLoader(utm): ', ConYesNoName(config.tapeFastLoader));
end;

procedure CmdTapeBit (const cmd: string; const args: array of string);
var
  f: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set tape bit state.');
    WriteLn('usage: tapebit 0|1|-1');
    WriteLn('  0: 0');
    WriteLn('  1: 1');
    WriteLn('  -1: tape emu');
    exit;
  end;

  if Length(args) < 1 then WriteLn('wtf?')
  else
  begin
    f := StrToIntDef(args[0], -1);
    config.tapeBit := f;
    Write('tape bit: ');
    case f of
      0: WriteLn('0');
      1: WriteLn('1');
      else WriteLn('tape emu');
    end;
  end;
end;


initialization
  PCJoyReset();
  LoadKeys();
  LoadConLogo();
  BuildOutFont();

  RegisterConsoleCommand('pause', @CmdPause);
  RegisterConsoleCommand('temp_pause', @CmdPause);
  RegisterConsoleCommand('defocus_pause', @CmdDefocusPause);

  RegisterConsoleCommand('speed', @CmdSpeed);

  RegisterConsoleCommand('tape_autostart', @CmdTapeAutostart);
  RegisterConsoleCommand('tape_automax', @CmdTapeAutomax);
  RegisterConsoleCommand('tape_autopause', @CmdTapeAutopause);
  RegisterConsoleCommand('tape_prepause', @CmdTapePrepause);

  RegisterConsoleCommand('mouse_autolock', @CmdMALock);
  RegisterConsoleCommand('kmouse_bswap', @CmdMsSwap);

  RegisterConsoleCommand('autoboot', @CmdAutoboot);
  RegisterConsoleCommand('boot_file', @CmdBootFile);

  RegisterConsoleCommand('logging', @CmdLogging);
  RegisterConsoleCommand('log_time', @CmdLogTime);

  RegisterConsoleCommand('rom', @CmdRom);

  RegisterConsoleCommand('con_slide', @CmdConSlide);
  RegisterConsoleCommand('con_slide_delay', @CmdConSlideDelay);
  RegisterConsoleCommand('con_height', @CmdConHeight);
  RegisterConsoleCommand('con_dump_lines', @CmdConDumpLines);
  RegisterConsoleCommand('con_dump_delay', @CmdConDumpDelay);

  RegisterConsoleCommand('console', @CmdConsole);

  RegisterConsoleCommand('sound', @CmdSound);
  RegisterConsoleCommand('sound_ay', @CmdSoundAY);

  RegisterConsoleCommand('autofire', @CmdConAF);
  RegisterConsoleCommand('autofire_rate', @CmdConAFRate);
  RegisterConsoleCommand('autofire_hold', @CmdConAFHold);

  RegisterConsoleCommand('clock', @CmdConClock);

  RegisterConsoleCommand('joy-reset', @CmdJoyReset);

  RegisterConsoleCommand('fasttape', @CmdFastTape);
  RegisterConsoleCommand('tapebit', @CmdTapeBit);

  RegisterConsoleCommand('joy_debug_info', @CmdJoyDebugInfo);
end.
