// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// ALISP3 API
//
unit dspLispApiU;

interface

uses
  SysUtils,
  z80U, dspEmuVarsU, dspEmuMiscU, dspVScrU, dspMachConstU, dspZXDrawU,
  dspLabelsU, zxFloatU, dzsLibU, xpackU,
  alisp3U, lispCon3U;



implementation


function IsFalse (n: Integer): Boolean;
begin
  result := ALIsFalse(n) or ALIsNil(n);
end;

function BoolVal (b: Boolean): Integer;
begin
  if b then result := ALTrue() else result := ALFalse();
end;


function ALCPause (n: Integer): Integer;
begin
  result := BoolVal(emuPaused);
  if not ALIsNil(n) then
  begin
    n := ALCar(n);
    if IsFalse(n) then ConExecute('pause ona') else ConExecute('pause tan');
  end;
end;

function ALCPeek (n: Integer): Integer;
// (peek addr [page])
var
  addr, page, maxPg: Integer;
begin
  if not ALNextIntArg('PEEK', n, addr, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextIntArg('PEEK', n, page, result) then exit;
  end
  else page := -255;
  addr := addr and $FFFF;

  if (addr >= 0) and (addr <= $3FFF) and (page < 0) then
  begin
    // 8: 128k rom, part 1
    // 9: 48k rom/128k standard 48k rom
    // 10: tr-dos rom
    if page = -1 then page := 9
    else if page = -2 then page := 8
    else if page = -3 then page := 10
    else page := emuPaging[0];
    result := ALNewIntAtom(emuMemory[page, addr]);
    exit;
  end;

  if (page < 0) or (addr < $C000) then
  begin
    if (addr < 0) or (addr > 65535) then
      result := ALError('invalid address', n)
    else
      result := ALNewIntAtom(emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)]);
  end
  else
  begin
    if config.pent512 then maxPg := 31 else maxPg := 7;
    if (addr > 65535) or (page < 0) or (page > maxPg) then
      result := ALError('invalid address', n)
    else
    begin
      if page > 7 then Inc(page, 3); // skip ROM pages
      result := ALNewIntAtom(emuMemory[page, (addr and $3FFF)]);
    end;
  end;
end;

function ALCPeekBuf (n: Integer): Integer;
// (peek-buf byte-buffer addr [page])
var
  bv, c, len, addr, page, pg, maxPg: Integer;
begin
  if not ALNextByteVectorArg('PEEK-BUF', n, bv, result) then exit;
  if not ALNextIntArg('PEEK-BUF', n, addr, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextIntArg('PEEK-BUF', n, page, result) then exit;
  end
  else page := -255;
  addr := addr and $FFFF;

  len := ALGetVectorLength(bv);
  if len < 1 then begin result := ALNil(); exit; end;
  if len > 65536 then len := 65536;
  addr := addr and $FFFF;

  if config.pent512 then maxPg := 31 else maxPg := 7;
  if page > 7 then Inc(page, 3); // skip ROM pages

  for c := 0 to len-1 do
  begin
    if (addr >= 0) and (addr <= $3FFF) and (page < 0) then
    begin
      // 8: 128k rom, part 1
      // 9: 48k rom/128k standard 48k rom
      // 10: tr-dos rom
      if page = -1 then pg := 9
      else if page = -2 then pg := 8
      else if page = -3 then pg := 10
      else pg := emuPaging[0];
      ALSetVectorItem(bv, c, emuMemory[pg, addr]);
    end
    else
    begin
      if (page < 0) or (addr < $C000) then
      begin
        ALSetVectorItem(bv, c, emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)]);
      end
      else
      begin
        if page <= maxPg then ALSetVectorItem(bv, c, emuMemory[page, (addr and $3FFF)])
        else ALSetVectorItem(bv, c, $FF);
      end;
    end;
    addr := (addr+1) and $FFFF;
  end;

  result := bv;
end;

function ALCPoke (n: Integer): Integer;
// (poke val addr [page])
var
  value, addr, page, maxPg: Integer;
begin
  if not ALNextIntArg('POKE', n, value, result) then exit;
  if not ALNextIntArg('POKE', n, addr, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextByteVectorArg('POKE', n, page, result) then exit;
  end
  else page := -255;
  addr := addr and $FFFF;

  if (value < -128) or (value > 255) then
  begin
    result := ALError('invalid value', n);
    exit;
  end;
  if value < 0 then value := 256+value;

  if (addr >= 0) and (addr <= $3FFF) and (page < 0) then
  begin
    // 8: 128k rom, part 1
    // 9: 48k rom/128k standard 48k rom
    // 10: tr-dos rom
    if page = -1 then page := 9
    else if page = -2 then page := 8
    else if page = -3 then page := 10
    else page := emuPaging[0];
    result := ALNewIntAtom(emuMemory[page, addr]);
    // no ROM pokez!
    exit;
  end;

  if (page < 0) or (addr < $C000) then
  begin
    if (addr < 0) or (addr > 65535) then
      result := ALError('invalid address', n)
    else
    begin
      result := ALNewIntAtom(emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)]);
      if addr > $3FFF then Z80PokeBInternal(emuZ80, addr, value);
    end;
  end
  else
  begin
    if config.pent512 then maxPg := 31 else maxPg := 7;
    if (addr > 65535) or (page < 0) or (page > maxPg) then
      result := ALError('invalid address', n)
    else
    begin
      if page > 7 then Inc(page, 3); // skip ROM pages
      n := emuPaging[(addr and $C000) shr 14];
      emuPaging[(addr and $C000) shr 14] := page;
      result := ALNewIntAtom(emuMemory[page, (addr and $3FFF)]);
      Z80PokeBInternal(emuZ80, addr, value);
      emuPaging[(addr and $C000) shr 14] := n;
    end;
  end;
end;

function ALCPokeBuf (n: Integer): Integer;
// (poke-buf byte-buf addr [page])
var
  bv, f, len, addr, page, maxPg: Integer;
begin
  if not ALNextByteVectorArg('PEEK-BUF', n, bv, result) then exit;
  if not ALNextIntArg('PEEK-BUF', n, addr, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextIntArg('PEEK-BUF', n, page, result) then exit;
  end
  else page := -255;
  addr := addr and $FFFF;

  if config.pent512 then maxPg := 31 else maxPg := 7;
  if page > 7 then Inc(page, 3); // skip ROM pages

  len := ALGetVectorLength(bv);
  for f := 0 to len-1 do
  begin
    if (addr >= 0) and (addr <= $3FFF) and (page < 0) then
    begin
      // 8: 128k rom, part 1
      // 9: 48k rom/128k standard 48k rom
      // 10: tr-dos rom
      (*if page = -1 then pg := 9
      else if page = -2 then pg := 8
      else if page = -3 then pg := 10
      else pg := emuPaging[0];*)
      // no ROM pokez!
    end
    else
    begin
      if (page < 0) or (addr < $C000) then
      begin
        if addr > $3FFF then Z80PokeBInternal(emuZ80, addr, ALGetVectorItem(bv, f));
      end
      else
      begin
        if page <= maxPg then
        begin
          n := emuPaging[(addr and $C000) shr 14];
          emuPaging[(addr and $C000) shr 14] := page;
          Z80PokeBInternal(emuZ80, addr, ALGetVectorItem(bv, f));
          emuPaging[(addr and $C000) shr 14] := n;
        end;
      end;
    end;
    addr := (addr+1) and $FFFF;
  end;

  result := ALTrue();
end;

function ALCPortIn (n: Integer): Integer;
// (port-in port)
var
  addr: Integer;
  oc: Boolean;
begin
  if not ALNextIntArg('PORT-IN', n, addr, result) then exit;
  if (addr < 0) or (addr > 65535) then result := ALError('invalid address', n)
  else
  begin
    emuEventDontCall := true;
    oc := config.contendedMemory; config.contendedMemory := false;
    result := ALNewIntAtom(Z80PortIn(emuZ80, addr));
    config.contendedMemory := oc; emuEventDontCall := false;
  end;
end;

function ALCPortOut (n: Integer): Integer;
// (port-out value port)
var
  addr, value: Integer;
  oc: Boolean;
begin
  if not ALNextIntArg('PORT-OUT', n, value, result) then exit;
  if not ALNextIntArg('PORT-OUT', n, addr, result) then exit;
  if (addr < 0) or (addr > 65535) then result := ALError('invalid address', n)
  else if (value < -128) or (value > 255) then result := ALError('invalid value', n)
  else
  begin
    if value < 0 then value := 256+value;
    emuEventDontCall := true;
    oc := config.contendedMemory; config.contendedMemory := false;
    result := ALNewIntAtom(Z80PortIn(emuZ80, addr));
    Z80PortOut(emuZ80, addr, value);
    config.contendedMemory := oc; emuEventDontCall := false;
  end;
end;

function ALCZ80Reg (n: Integer): Integer;
// (z80reg "reg" [new-val])
var
  value: Integer;
  s: string;
  doSet: Boolean;
begin
  if not ALNextStrArg('Z80REG', n, s, result) then exit;

  if not ALIsNil(n) then
  begin
    if not ALNextIntArg('Z80REG', n, value, result) then exit;
    doSet := true;
  end
  else
  begin
    doSet := false; value := 0;
  end;

  if doSet and ((value < -32768) or (value > 65535)) then
  begin
    result := ALError('invalid value', n);
    exit;
  end;
  if value < 0 then value := 65536+value;
  s := UpperCase(s);

  if s = 'IM' then
  begin
    if doSet and (value > 2) then begin result := ALError('invalid value', n); exit; end;
    result := ALNewIntAtom(emuZ80.im);
    if not doSet then exit;
    emuZ80.im := value;
    exit;
  end
  else if (s = 'IFF') or (s = 'IFF1') or (s = 'IFF2') then
  begin
    if doSet and (value > 1) then begin result := ALError('invalid value', n); exit; end;
    if (s = 'IFF') or (s = 'IFF1') then
    begin
      if emuZ80.iff1 then result := ALNewIntAtom(1) else result := ALNewIntAtom(0);
    end
    else if emuZ80.iff2 then result := ALNewIntAtom(1) else result := ALNewIntAtom(0);
    if not doSet then exit;
    if (s = 'IFF') or (s = 'IFF1') then emuZ80.iff1 := value <> 0;
    if (s = 'IFF') or (s = 'IFF2') then emuZ80.iff2 := value <> 0;
    exit;
  end;

  if Length(s) = 1 then value := value and $FF;
  if s = 'A' then
  begin
    result := ALNewIntAtom(emuZ80.regA);
    if doSet then emuZ80.regA := value;
  end
  else if s = 'F' then
  begin
    result := ALNewIntAtom(Z80GetF(emuZ80));
    if doSet then Z80SetF(emuZ80, value);
  end
  else if s = 'B' then
  begin
    result := ALNewIntAtom(emuZ80.regB);
    emuZ80.regB := value;
  end
  else if s = 'C' then
  begin
    result := ALNewIntAtom(emuZ80.regC);
    if doSet then emuZ80.regC := value;
  end
  else if s = 'D' then
  begin
    result := ALNewIntAtom((emuZ80.regDE shr 8) and $FF);
    if doSet then emuZ80.regDE := (emuZ80.regDE and $00FF) or (value shl 8);
  end
  else if s = 'E' then
  begin
    result := ALNewIntAtom(emuZ80.regDE and $FF);
    if doSet then emuZ80.regDE := (emuZ80.regDE and $FF00) or value;
  end
  else if s = 'H' then
  begin
    result := ALNewIntAtom((emuZ80.regHL shr 8) and $FF);
    if doSet then emuZ80.regHL := (emuZ80.regHL and $00FF) or (value shl 8);
  end
  else if s = 'L' then
  begin
    result := ALNewIntAtom(emuZ80.regHL and $FF);
    if doSet then emuZ80.regHL := (emuZ80.regHL and $FF00) or value;
  end
  else if s = 'AF' then
  begin
    result := ALNewIntAtom(Z80GetAF(emuZ80));
    if doSet then Z80SetAF(emuZ80, value);
  end
  else if s = 'BC' then
  begin
    result := ALNewIntAtom(Z80GetBC(emuZ80));
    if doSet then Z80SetBC(emuZ80, value);
  end
  else if s = 'DE' then
  begin
    result := ALNewIntAtom(emuZ80.regDE);
    if doSet then emuZ80.regDE := value;
  end
  else if s = 'HL' then
  begin
    result := ALNewIntAtom(emuZ80.regHL);
    if doSet then emuZ80.regHL := value;
  end
  else if s = 'IX' then
  begin
    result := ALNewIntAtom(emuZ80.regIX);
    if doSet then emuZ80.regIX := value;
  end
  else if s = 'IY' then
  begin
    result := ALNewIntAtom(emuZ80.regIY);
    if doSet then emuZ80.regIY := value;
  end
  else if s = 'SP' then
  begin
    result := ALNewIntAtom(emuZ80.regSP);
    if doSet then emuZ80.regSP := value;
  end
  else if s = 'PC' then
  begin
    result := ALNewIntAtom(emuZ80.regPC);
    if doSet then emuZ80.regPC := value;
  end
  else if s = 'I' then
  begin
    result := ALNewIntAtom(emuZ80.regI);
    if doSet then emuZ80.regI := value;
  end
  else if s = 'R' then
  begin
    result := ALNewIntAtom((emuZ80.regR and $80) or (emuZ80.regRTemp and $7F));
    if doSet then
    begin
      emuZ80.regR := value and $80; emuZ80.regRTemp := (value and $7F);
    end;
  end
  else if s = 'AF''' then
  begin
    result := ALNewIntAtom(emuZ80.regAFx);
    if doSet then emuZ80.regAFx := value;
  end
  else if s = 'BC''' then
  begin
    result := ALNewIntAtom(emuZ80.regBCx);
    if doSet then emuZ80.regBCx := value;
  end
  else if s = 'DE''' then
  begin
    result := ALNewIntAtom(emuZ80.regDEx);
    if doSet then emuZ80.regDEx := value;
  end
  else if s = 'HL''' then
  begin
    result := ALNewIntAtom(emuZ80.regHLx);
    if doSet then emuZ80.regHLx := value;
  end
  else
  begin
    result := ALError('invalid register name', n);
  end;
end;

function ALCTicks (n: Integer): Integer;
// (ticks [new-val])
var
  new: Integer;
begin
  if not ALIsNil(n) then
  begin
    if not ALNextIntArg('TICKS', n, new, result) then exit;
  end
  else new := -1;

  result := ALNewIntAtom(emuz80.tStates);
  if new > emuTicksIntr then new := emuTicksIntr;
  if new >= 0 then emuz80.tStates := new;
end;

function ALCIncTicks (n: Integer): Integer;
// (inc-ticks delta)
var
  new: Integer;
begin
  if not ALNextIntArg('INC-TICKS', n, new, result) then exit;
  result := ALNewIntAtom(emuz80.tStates);
  new := emuz80.tStates+new;
  if new < 0 then new := 0 else if new > emuTicksIntr then new := emuTicksIntr;
  emuz80.tStates := new;
end;

function ALCScrUpdate (n: Integer): Integer;
{var
  c: Integer;}
begin
  if not ALIsNil(n) and not ALIsFalse(n) then
  begin
    //ZXExUpdateScr(te-1, @emuMemory[emuScreenPage]);
    //for c := 0 to 239 do ZXUpdateLine(c, @(emuMemory[emuScreenPage]));
    EmuRenderScreen();
  end
  else {EmuRenderScreen(); //EmuDrawScreen(); }EmuDrawWCon();
  EmuDrawScreen();
  result := ALTrue();
end;

function ALCScrPeek (n: Integer): Integer;
// (scr-peek addr)
var
  addr: Integer;
begin
  if not ALNextIntArg('SCR-PEEK', n, addr, result) then exit;
  if (addr < 0) or (addr > 6143) then result := ALNewIntAtom(0)
  else
  begin
    result := ALNewIntAtom(emuMemory[emuScreenPage, addr]);
    //emuMemory[emuScreenPage, addr] :=
    //ZXExUpdateScr(emuz80.tStates-1, @emuMemory[emuScreenPage]);
  end;
end;

function ALCScrPoke (n: Integer): Integer;
// (scr-poke value addr)
var
  addr, value: Integer;
begin
  if not ALNextIntArg('SCR-POKE', n, value, result) then exit;
  if not ALNextIntArg('SCR-POKE', n, addr, result) then exit;

  if (addr < 0) or (addr > 6143) then result := ALNewIntAtom(0)
  else if (value < -128) or (value > 255) then
    result := ALError('invalid value', n)
  else
  begin
    if value < 0 then value := 256+value;
    result := ALNewIntAtom(emuMemory[emuScreenPage, addr]);
    emuMemory[emuScreenPage, addr] := value;
    ZXExUpdateScr(emuz80.tStates-1, @emuMemory[emuScreenPage]);
  end;
end;

function ALCScrPoint (n: Integer): Integer;
// (scr-point x y)
var
  x, y, addr, value: Integer;
begin
  if not ALNextIntArg('SCR-POINT', n, x, result) then exit;
  if not ALNextIntArg('SCR-POINT', n, y, result) then exit;
  if (x < 0) or (y < 0) or (x > 255) or (y > 191) then result := ALFalse()
  else
  begin
    addr := emuDrawZXLineOfs[y]+(x div 8);
    x :=7-(x and $07); // shift
    value := (emuMemory[emuScreenPage, addr] shr x) and $01;
    result := BoolVal(value <> 0);
  end;
end;

function ALCScrPlot (n: Integer): Integer;
// (scr-plot flag x y)
var
  x, y, addr, value, mask: Integer;
  doSet: Boolean;
begin
  doSet := not IsFalse(ALCar(n)); n := ALCdr(n);
  if not ALNextIntArg('SCR-PLOT', n, x, result) then exit;
  if not ALNextIntArg('SCR-PLOT', n, y, result) then exit;
  //WriteLn('plot at ', x, ',', y);
  if (x < 0) or (y < 0) or (x > 255) or (y > 191) then result := ALFalse()
  else
  begin
    addr := emuDrawZXLineOfs[y]+(x div 8);
    x := x and $07; // shift
    mask := $80 shr x;
    //WriteLn('addr=#', IntToHex(addr+16384, 4), '; mask=#', IntToHex(mask, 2));
    value := emuMemory[emuScreenPage, addr];
    result := BoolVal((value and mask) <> 0);
    if doSet then value := value or mask else value := value and (not mask);
    emuMemory[emuScreenPage, addr] := value;
    ZXExUpdateScr(emuz80.tStates-1, @emuMemory[emuScreenPage]);
  end;
end;

function ALCAttrGet (n: Integer): Integer;
// (attr-get x y)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('ATTR-GET', n, x, result) then exit;
  if not ALNextIntArg('ATTR-GET', n, y, result) then exit;
  if (x < 0) or (y < 0) or (x > 255) or (y > 191) then result := ALNewIntAtom(0)
  else
  begin
    addr := (y div 8)*32+(x div 8)+6144;
    result := ALNewIntAtom(emuMemory[emuScreenPage, addr]);
  end;
end;

function ALCAttrSet (n: Integer): Integer;
// (attr-get new x y)
var
  x, y, addr, value: Integer;
begin
  if not ALNextIntArg('ATTR-SET', n, value, result) then exit;
  if not ALNextIntArg('ATTR-SET', n, x, result) then exit;
  if not ALNextIntArg('ATTR-SET', n, y, result) then exit;
  if (x < 0) or (y < 0) or (x > 255) or (y > 191) then result := ALNewIntAtom(0)
  else if (value < -128) or (value > 255) then result := ALError('invalid value', n)
  else
  begin
    if value < 0 then value := 256+value;
    addr := (y div 8)*32+(x div 8)+6144;
    result := ALNewIntAtom(emuMemory[emuScreenPage, addr]);
    emuMemory[emuScreenPage, addr] := value;
    ZXExUpdateScr(emuz80.tStates-1, @emuMemory[emuScreenPage]);
  end;
end;

function ALCScrAddr (n: Integer): Integer;
// (scr-addr x y)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('SCR-ADDR', n, x, result) then exit;
  if not ALNextIntArg('SCR-ADDR', n, y, result) then exit;
  if (x < 0) or (y < 0) or (x > 255) or (y > 191) then addr := -1
  else addr := emuDrawZXLineOfs[y]+(x div 8);
  result := ALNewIntAtom(addr);
end;

function ALCAttrAddr (n: Integer): Integer;
// (attr-addr x y)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('ATTR-ADDR', n, x, result) then exit;
  if not ALNextIntArg('ATTR-ADDR', n, y, result) then exit;
  if (x < 0) or (y < 0) or (x > 255) or (y > 191) then addr := -1
  else addr := (y div 8)*32+(x div 8)+6144;
  result := ALNewIntAtom(addr);
end;

function ALCScrToAttr (n: Integer): Integer;
// (scr-to-attr addr)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('SCR-TO-ATTR', n, addr, result) then exit;
  if (addr < 0) or (addr > 6143) then addr := -1
  else
  begin
    x := (addr and $1F);
    y := ((addr shr 5) and $07); Inc(y, 8*((addr shr 11) and $03));
    addr := y*32+x+6144;
  end;
  result := ALNewIntAtom(addr);
end;

function ALCAttrToScr (n: Integer): Integer;
// (attr-to-scr addr)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('ATTR-TO-SCR', n, addr, result) then exit;
  if (addr < 6144) or (addr > 6911) then addr := -1
  else
  begin
    Dec(addr, 6144);
    x := (addr and $1F); y := (addr div 32)*8;
    addr := emuDrawZXLineOfs[y]+x;
  end;
  result := ALNewIntAtom(addr);
end;

function ALCScrLineUp (n: Integer): Integer;
// (scr-line-up addr)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('SCR-LINE-UP', n, addr, result) then exit;
  if (addr < 0) or (addr > 6143) then addr := -1
  else
  begin
    x := (addr and $1F);
    y := ((addr shr 5) and $07); Inc(y, 8*((addr shr 11) and $03));
    y := (y*8)+((addr shr 8) and $07);
    Dec(y);
    if (y < 0) or (y > 191) then addr := -1 else addr := emuDrawZXLineOfs[y]+x;
  end;
  result := ALNewIntAtom(addr);
end;

function ALCScrLineDown (n: Integer): Integer;
// (scr-line-down addr)
var
  x, y, addr: Integer;
begin
  if not ALNextIntArg('SCR-LINE-DOWN', n, addr, result) then exit;
  if (addr < 0) or (addr > 6143) then addr := -1
  else
  begin
    x := (addr and $1F);
    y := ((addr shr 5) and $07); Inc(y, 8*((addr shr 11) and $03));
    y := (y*8)+((addr shr 8) and $07);
    Inc(y);
    if (y < 0) or (y > 191) then addr := -1 else addr := emuDrawZXLineOfs[y]+x;
  end;
  result := ALNewIntAtom(addr);
end;

function ALCNum2ZXFloat (n: Integer): Integer;
// (num2zxfloat number byte-buffer)
var
  bv: Integer;
  fv: Single;
begin
  if not ALNextFloatArg('NUM2ZXFLOAT', n, fv, result) then exit;
  if not ALNextByteVectorArg('NUM2ZXFLOAT', n, bv, result) then exit;
  ALSetVectorLength(bv, 5);
  if not FloatToSpeccy(fv, ALGetByteVectorDataPtr(bv)^) then FloatToSpeccy(-1, ALGetByteVectorDataPtr(bv)^);
  result := bv;
end;

function ALCZXFloat2Num (n: Integer): Integer;
// (zxfloat2num byte-buffer)
var
  bv, len: Integer;
  fv: Single;
begin
  if not ALNextByteVectorArg('ZXFLOAT2NUM', n, bv, result) then exit;
  len := ALGetVectorLength(bv);
  if len < 5 then
  begin
    result := ALError('ZXFLOAT2NUM: byte-vector too small');
    exit;
  end;

  fv := SpeccyToFloat(ALGetByteVectorDataPtr(bv)^);
  result := ALNewFloatAtom(fv);
end;

function ALCLabelFindCode (n: Integer): Integer;
// (label-find-code addr)
var
  addr: Integer;
  s: string;
  ok: Boolean;
begin
  if not ALNextIntArg('LABEL-FIND-CODE', n, addr, result) then exit;
  addr := addr and $FFFF;
  s := ZFindCodeLabel(addr, ok); if not ok then s := '';
  result := ALNewStrAtom(s);
end;

function ALCLabelFindData (n: Integer): Integer;
// (label-find-data addr)
var
  addr: Integer;
  s: string;
  ok: Boolean;
begin
  if not ALNextIntArg('LABEL-FIND-DATA', n, addr, result) then exit;
  addr := addr and $FFFF;
  s := ZFindDataLabel(addr, ok); if not ok then s := '';
  result := ALNewStrAtom(s);
end;

function ALCLabelFindName (n: Integer): Integer;
// (label-find-name name)
var
  s: string;
  addr: Word;
  size: Integer;
  sys: Boolean;
begin
  if not ALNextStrArg('LABEL-FIND-NAME', n, s, result) then exit;
  if ZFindLabelByName(s, addr, size, sys) then
  begin
    result := ALCons(ALNewIntAtom(addr), ALNil());
    ALPushCell(result); n := result;
    ALSetCdr(n, ALCons(ALNewIntAtom(size), ALNil())); n := ALCdr(n);
    if sys then ALSetCdr(n, ALCons(ALTrue(), ALNil()))
    else ALSetCdr(n, ALCons(ALFalse(), ALNil()));
    n := ALCdr(n);
    ALSetCdr(n, ALCons(ALNewStrAtom(s), ALNil()));
    ALPopCells(1);
  end
  else result := ALNil();
end;

function ALCLabelDelete (n: Integer): Integer;
// (label-delete name)
var
  s: string;
  sys: Boolean;
begin
  if not ALNextStrArg('LABEL-DELETE', n, s, result) then exit;
  n := ALCar(n); sys := not (ALIsNil(n) or ALIsFalse(n));
  ZDeleteLabel(s, sys);
  result := ALTrue();
end;

function ALCLabelAdd (n: Integer): Integer;
// (label-add name addr size sys-flag)
var
  addr, size: Integer;
  s: string;
  sys: Boolean;
begin
  if not ALNextStrArg('LABEL-ADD', n, s, result) then exit;
  if not ALNextIntArg('LABEL-ADD', n, addr, result) then exit;
  if not ALNextIntArg('LABEL-ADD', n, size, result) then exit;
  n := ALCar(n); sys := not (ALIsNil(n) or ALIsFalse(n));
  ZAddLabel(s, addr and $FFFF, size, sys);
  result := ALTrue();
end;

function ALCLabelsClear (n: Integer): Integer;
// (labels-clear sys-flag)
var
  sys: Boolean;
begin
  n := ALCar(n); sys := not (ALIsNil(n) or ALIsFalse(n));
  ZClearLabels(sys);
  result := ALTrue();
end;

function ALCLabelsLoad (n: Integer): Integer;
// (labels-load fname sys-flag)
var
  s: string;
  sys: Boolean;
begin
  if not ALNextStrArg('LABELS-LOAD', n, s, result) then exit;
  n := ALCar(n); sys := not (ALIsNil(n) or ALIsFalse(n));
  ZLoadLabels(s, sys);
  result := ALTrue();
end;

function ALCLabelsSave (n: Integer): Integer;
// (labels-save fname sys-flag)
var
  s: string;
  sys: Boolean;
begin
  if not ALNextStrArg('LABELS-SAVE', n, s, result) then exit;
  n := ALCar(n); sys := not (ALIsNil(n) or ALIsFalse(n));
  ZSaveLabels(s, sys);
  result := ALTrue();
end;

function ALCInfoMessage (n: Integer): Integer;
// (info-message msg [ms-to-keep])
var
  pause: Integer;
  s: string;
begin
  if not ALNextStrArg('INFO-MESSAGE', n, s, result) then exit;
  if ALIsNil(n) then pause := 2000
  else if not ALNextIntArg('INFO-MESSAGE', n, pause, result) then exit;

  if pause < 500 then pause := 500;
  if pause > 60000 then pause := 60000;
  EmuInfoMessage(s, LongWord(pause));

  result := ALTrue();
end;

function PackCodec (bv, maxSize: Integer; cname: string): Integer;
var
  len: Integer;
  destBuf: packed array of Byte;
  srcBuf: Pointer;
  destSz: Integer;
begin
  if (maxSize < 1) or (maxSize > 1024*1024) then begin result := ALFalse(); exit; end;
  len := ALGetVectorLength(bv); srcBuf := ALGetByteVectorDataPtr(bv);
  SetLength(destBuf, maxSize);
  if cname = 'compress-rle' then
    destSz := Z80RLEPackBuffer(srcBuf^, len, destBuf[0], maxSize)
  else if cname = 'compress-x' then
    destSz := XPackBuffer(srcBuf^, len, destBuf[0], maxSize)
  else if cname = 'decompress-rle' then
    destSz := Z80RLEUnpackBuffer(srcBuf^, len, destBuf[0], maxSize)
  else if cname = 'decompress-x' then
    destSz := XUnpackBuffer(srcBuf^, len, destBuf[0], maxSize)
  else destSz := -1;

  if destSz = -1 then result := ALFalse()
  else
  begin
    ALSetVectorLength(bv, destSz);
    Move(destBuf[0], ALGetByteVectorDataPtr(bv)^, destSz);
    result := bv;
  end;
end;

function ALCCompressRLE (n: Integer): Integer;
// (compress-rle byte-buffer maxsize)
var
  bv, maxSz: Integer;
begin
  if not ALNextByteVectorArg('COMPRESS-RLE', n, bv, result) then exit;
  if not ALNextIntArg('COMPRESS-RLE', n, maxSz, result) then exit;
  PackCodec(bv, maxSz, 'compress-rle');
end;

function ALCCompressX (n: Integer): Integer;
// (compress-x byte-buffer maxsize)
var
  bv, maxSz: Integer;
begin
  if not ALNextByteVectorArg('COMPRESS-X', n, bv, result) then exit;
  if not ALNextIntArg('COMPRESS-X', n, maxSz, result) then exit;
  PackCodec(bv, maxSz, 'compress-x');
end;

function ALCDecompressRLE (n: Integer): Integer;
// (decompress-rle byte-buffer maxsize)
var
  bv, maxSz: Integer;
begin
  if not ALNextByteVectorArg('DECOMPRESS-RLE', n, bv, result) then exit;
  if not ALNextIntArg('DECOMPRESS-RLE', n, maxSz, result) then exit;
  PackCodec(bv, maxSz, 'decompress-rle');
end;

function ALCDecompressX (n: Integer): Integer;
// (decompress-x byte-buffer maxsize)
var
  bv, maxSz: Integer;
begin
  if not ALNextByteVectorArg('DECOMPRESS-X', n, bv, result) then exit;
  if not ALNextIntArg('DECOMPRESS-X', n, maxSz, result) then exit;
  PackCodec(bv, maxSz, 'decompress-x');
end;

function ALCEscapeScr (n: Integer): Integer;
// (escape-str str)
var
  nn: Integer;
  s, t: string;
begin
  nn := ALCar(n);
  if not ALNextStrArg('ESCAPE-STR', n, s, result) then exit;
  t := ALEscapeStr(s);
  if t <> s then
  begin
    result := ALNewStrAtom(Copy(t, 2, Length(t)-2))
  end
  else result := nn;
end;


begin
  ALAddPrimitive('pause', ALCPause, false, 0, 1);
  ALAddPrimitive('peek', ALCPeek, false, 1, 2);
  ALAddPrimitive('poke', ALCPoke, false, 2, 3);
  ALAddPrimitive('peek-buf', ALCPeekBuf, false, 2, 3);
  ALAddPrimitive('poke-buf', ALCPokeBuf, false, 2, 3);
  ALAddPrimitive('port-in', ALCPortIn, false, 1, 1);
  ALAddPrimitive('port-out', ALCPortOut, false, 2, 2);
  ALAddPrimitive('ticks', ALCTicks, false, 0, 1);
  ALAddPrimitive('inc-ticks', ALCIncTicks, false, 1, 1);
  ALAddPrimitive('z80reg', ALCZ80Reg, false, 1, 2);
  ALAddPrimitive('scr-update', ALCScrUpdate, false, 0, 1);
  ALAddPrimitive('scr-peek', ALCScrPeek, false, 1, 1);
  ALAddPrimitive('scr-poke', ALCScrPoke, false, 2, 2);
  ALAddPrimitive('attr-get', ALCAttrGet, false, 2, 2);
  ALAddPrimitive('attr-set', ALCAttrSet, false, 3, 3);
  ALAddPrimitive('scr-addr', ALCScrAddr, false, 2, 2);
  ALAddPrimitive('attr-addr', ALCAttrAddr, false, 2, 2);
  ALAddPrimitive('scr-to-attr', ALCScrToAttr, false, 1, 1);
  ALAddPrimitive('attr-to-scr', ALCAttrToScr, false, 1, 1);
  ALAddPrimitive('scr-line-up', ALCScrLineUp, false, 1, 1);
  ALAddPrimitive('scr-line-down', ALCScrLineDown, false, 1, 1);
  ALAddPrimitive('scr-point', ALCScrPoint, false, 2, 2);
  ALAddPrimitive('scr-plot', ALCScrPlot, false, 3, 3);
  ALAddPrimitive('zxfloat2num', ALCZXFloat2Num, false, 1, 1);
  ALAddPrimitive('num2zxfloat', ALCNum2ZXFloat, false, 1, 1);
  ALAddPrimitive('label-find-code', ALCLabelFindCode, false, 1, 1);
  ALAddPrimitive('label-find-data', ALCLabelFindData, false, 1, 1);
  ALAddPrimitive('label-find-name', ALCLabelFindName, false, 1, 1);
  ALAddPrimitive('label-delete', ALCLabelDelete, false, 4, 4);
  ALAddPrimitive('label-add', ALCLabelAdd, false, 2, 2);
  ALAddPrimitive('labels-clear', ALCLabelsClear, false, 1, 1);
  ALAddPrimitive('labels-load', ALCLabelsLoad, false, 2, 2);
  ALAddPrimitive('labels-save', ALCLabelsSave, false, 2, 2);
  ALAddPrimitive('info-message', ALCInfoMessage, false, 1, 2);
  ALAddPrimitive('compress-rle', ALCCompressRLE, false, 2, 2);
  ALAddPrimitive('compress-x', ALCCompressX, false, 2, 2);
  ALAddPrimitive('decompress-rle', ALCDecompressRLE, false, 2, 2);
  ALAddPrimitive('decompress-x', ALCDecompressX, false, 2, 2);

  ALAddPrimitive('escape-str', ALCEscapeScr, false, 1, 1);
end.
