// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// DirectDraw renderer
//
unit dspDDrawU;

interface

uses
  Windows, SysUtils, ActiveX,
  DirectDraw7, dspVScrU, dspEmuVarsU, dspEmuMouseU;


type
  TDDrawPutLine = procedure (src, dst: Pointer); register;


var
  // ddraw stuff
  pDDraw: IDirectDraw7 = nil;
  pPrimary: IDirectDrawSurface7 = nil;
  pBackBuffer: IDirectDrawSurface7 = nil;
  pClipper: IDirectDrawClipper = nil;
  ddrawLocked: Boolean = false;
  pBPP: Integer = 0;
  ddrawPutLine: TDDrawPutLine = nil;
  ddrawScaleHW: Boolean = true;


procedure DDrawDeinit ();
function DDrawInit (): Integer;
procedure DDrawUpdate ();
procedure DDrawReleaseScr ();
function DDrawLockScr (): Pointer;
procedure DDrawActualizeVScreen2x ();
procedure DDrawActualizeVScreen ();

function IsDDrawHere (): Boolean;


implementation


function DetermineShift (v: LongWord): Integer;
begin
  result := 0;
  if v = 0 then exit;
  while (v and $01) = 0 do begin Inc(result); v := v shr 1; end;
end;

function DetermineSize (v: LongWord): Integer;
begin
  result := 0;
  if v = 0 then exit;
  while (v and $01) = 0 do v := v shr 1;
  while (v and $01) <> 0 do begin Inc(result); v := v shr 1; end;
end;

{$INCLUDE putline.inc}

procedure DDrawDeinit ();
begin
  if pDDraw <> nil then
  begin
    pDDraw.RestoreDisplayMode();
    pDDraw.SetCooperativeLevel(dspWinHandle, DDSCL_NORMAL);
  end;
  pClipper := nil; pBackBuffer:= nil; pPrimary := nil; pDDraw := nil;
  ddrawPutLine := nil;
end;

function DDrawInit (): Integer;
var
  desc: TDDSurfaceDesc2;
  width, height: Integer;
  drvCaps, hwCaps: TDDCaps;
  sr, sg, sb: Integer;
  zr, zg, zb: Integer;
begin
  if not Assigned(DirectDrawCreate) then begin result := -1; exit; end;
  width := vScrWidth*2; height := vScrHeight*2;
  DDrawDeinit();

  // create DirectDraw object
  if Failed(DirectDrawCreateEx(nil, pDDraw, IID_IDirectDraw7, nil)) then
  begin
    result := -2; exit;
  end;

  // set cooperative level
  if Failed(pDDraw.SetCooperativeLevel(dspWinHandle, DDSCL_NORMAL)) then
  begin
    DDrawDeinit(); result := -3; exit;
  end;

  // create the primary surface
  ZeroMemory(@desc, SizeOf(desc));
  desc.dwSize := SizeOf(desc);
  desc.dwFlags := DDSD_CAPS;
  desc.ddsCaps.dwCaps := DDSCAPS_PRIMARYSURFACE or DDSCAPS_VIDEOMEMORY;
  if Failed(pDDraw.CreateSurface(desc, pPrimary, nil)) then
  begin
    DDrawDeinit(); result := -4; exit;
  end;

  // make a clipper for it
  if Failed(pDDraw.CreateClipper(0, pClipper, nil)) then
  begin
    DDrawDeinit(); result := -5; exit;
  end;
  if Failed(pClipper.SetHWnd(0, dspWinHandle)) then
  begin
    DDrawDeinit(); result := -6; exit;
  end;

  // assign the clipper to the primary surface
  if Failed(pPrimary.SetClipper(pClipper)) then
  begin
    DDrawDeinit(); result := -7; exit;
  end;

  // create the back buffer
  ZeroMemory(@desc, SizeOf(desc));
  desc.dwSize := SizeOf(desc);
  desc.dwFlags := DDSD_CAPS or DDSD_HEIGHT or DDSD_WIDTH;
  desc.ddsCaps.dwCaps := DDSCAPS_OFFSCREENPLAIN;
  desc.dwWidth := width; desc.dwHeight := height+4;
  if Failed(pDDraw.CreateSurface(desc, pBackBuffer, nil)) then
  begin
    DDrawDeinit(); result := -8; exit;
  end;

  if Failed(pBackBuffer.Lock(nil, desc,
       DDLOCK_SURFACEMEMORYPTR or DDLOCK_WRITEONLY or DDLOCK_WAIT, 0)) then
  begin
    DDrawDeinit(); result := -9; exit;
  end;
  pBackBuffer.Unlock(nil);

  pBPP := desc.ddpfPixelFormat.dwRGBBitCount;
  {WriteLn('r mask: $', IntToHex(desc.ddpfPixelFormat.dwRBitMask, 8));
  WriteLn('g mask: $', IntToHex(desc.ddpfPixelFormat.dwGBitMask, 8));
  WriteLn('b mask: $', IntToHex(desc.ddpfPixelFormat.dwBBitMask, 8));
  WriteLn('a mask: $', IntToHex(desc.ddpfPixelFormat.dwRGBAlphaBitMask, 8));}
  sr := DetermineShift(desc.ddpfPixelFormat.dwRBitMask);
  zr := DetermineSize(desc.ddpfPixelFormat.dwRBitMask);
  sg := DetermineShift(desc.ddpfPixelFormat.dwGBitMask);
  zg := DetermineSize(desc.ddpfPixelFormat.dwGBitMask);
  sb := DetermineShift(desc.ddpfPixelFormat.dwBBitMask);
  zb := DetermineSize(desc.ddpfPixelFormat.dwBBitMask);
  case pBPP of
    16:
      begin
        if zr = 4 then
        begin
          if (zg = 4) and (zb = 4) then ddrawPutLine := @PutLineR4G4B4;
        end
        else if zr = 5 then
        begin
          if (zg = 5) and (zb = 5) then ddrawPutLine := @PutLineR5G5B5;
          if (zg = 6) and (zb = 5) then ddrawPutLine := @PutLineR5G6B5;
        end;
      end;
    24:
      begin
        if (sr = 16) and (sg = 8) and (sb = 0) then ddrawPutLine := @PutLineR8G8B8_24;
      end;
    32:
      begin
        if (sr = 16) and (sg = 8) and (sb = 0) then ddrawPutLine := @PutLineR8G8B8;
      end;
  end;
  if not Assigned(ddrawPutLine) then begin DDrawDeinit(); result := -10; exit; end;

  ZeroMemory(@drvCaps, SizeOf(drvCaps)); drvCaps.dwSize := SizeOf(drvCaps);
  ZeroMemory(@hwCaps, SizeOf(hwCaps)); hwCaps.dwSize := SizeOf(hwCaps);
  if not Failed(pDDraw.GetCaps(@drvCaps, @hwCaps)) then
  begin
    //WriteLn('memory: total=', drvCaps.dwVidMemTotal, '; free=', drvCaps.dwVidMemFree);
  end;

  result := 0;
end;

procedure DDrawUpdate ();
var
  srcRect, destRect: TRect;
  pt: TPoint;
  fx: TDDBLTFX;
  //hr: HRESULT;
begin
  // draw the back buffer to the front buffer
  GetClientRect(dspWinHandle, destRect);
  destRect.Left := 0; destRect.Top := 0;
  if emuScale = 1 then
  begin
    destRect.Right := vScrWidth; destRect.Bottom := vScrHeight;
  end
  else
  begin
    destRect.Right := vScrWidth*2; destRect.Bottom := vScrHeight*2;
  end;
  pt.x := 0; pt.y := 0; ClientToScreen(dspWinHandle, pt);
  OffsetRect(destRect, pt.x, pt.y);

  ZeroMemory(@fx, SizeOf(fx));
  fx.dwSize := SizeOf(fx);
  //fx.dwDDFX := DDBLTFX_MIRRORUPDOWN;

  srcRect.Left := 0; srcRect.Top := 0;
  if ddrawScaleHW or (emuScale = 1) then
  begin
    srcRect.Right := vScrWidth; srcRect.Bottom := vScrHeight+1;
    if emuScale = 1 then Dec(srcRect.Bottom);
    pPrimary.Blt(@destRect, pBackBuffer, @srcRect,
      {DDBLT_DDFX or }DDBLT_ASYNC or DDBLT_WAIT, @fx);
  end
  else
  begin
    srcRect.Right := vScrWidth*2; srcRect.Bottom := vScrHeight*2;
    pPrimary.Blt(@destRect, pBackBuffer, @srcRect,
      {DDBLT_DDFX or }DDBLT_ASYNC or DDBLT_WAIT, @fx);
    //pPrimary.BltFast(0, 0, pBackBuffer, @srcRect, DDBLTFAST_DONOTWAIT);
  end;
  (*{hr := }pPrimary.Blt(@destRect, pBackBuffer, @srcRect,
    DDBLT_DDFX or DDBLT_ASYNC or DDBLT_WAIT, @fx);*)
  //if hr = DDERR_SURFACELOST then exit;
  //if hr <> DDERR_WASSTILLDRAWING then exit;
  //!pPrimary.BltFast(destRect.Left, destRect.Top, pBackBuffer, @srcRect, DDBLTFAST_DONOTWAIT);
end;

procedure DDrawReleaseScr ();
begin
  pBackBuffer.Unlock(nil);
  ddrawLocked := false;
end;

function DDrawLockScr (): Pointer;
var
  hr: HRESULT;
  desc: TDDSurfaceDesc2;
begin
  result := nil;
  ZeroMemory(@desc, SizeOf(desc));
  desc.dwSize := SizeOf(desc);

  // lock the back buffer surface to get its surface pointer
  hr := pBackBuffer.Lock(nil, desc,
    DDLOCK_SURFACEMEMORYPTR or DDLOCK_WRITEONLY or DDLOCK_WAIT or DDLOCK_DISCARDCONTENTS, 0);
  if Failed(hr) then exit;
  ddrawLocked := true;
  result := desc.lpSurface;
end;

procedure DDrawActualizeVScreen2x ();
var
  y: Integer;
  a, scr: PChar;
begin
  scr := DDrawLockScr();
  if scr = nil then exit;
  if ddrawScaleHW then
  begin
    for y := vScrHeight-1 downto 0 do
    begin
      a := @(dspVScreenRam[y, 0]);
      ddrawPutLine(a, scr);
      Inc(a, vScrWidth); Inc(scr, 640*(pBPP div 8));
    end;
    FillChar(scr^, 640*(pBPP div 8), 0);
  end
  else
  begin
    VScale2X();
    for y := vScrHeight-1 downto 0 do
    begin
      a := @(vScreenRam2x[y*2, 0]);
      ddrawPutLine(a, scr); Inc(a, vScrWidth); Inc(scr, 320*(pBPP div 8));
      ddrawPutLine(a, scr); Inc(a, vScrWidth); Inc(scr, 320*(pBPP div 8));
      ddrawPutLine(a, scr); Inc(a, vScrWidth); Inc(scr, 320*(pBPP div 8));
      ddrawPutLine(a, scr); Inc(a, vScrWidth); Inc(scr, 320*(pBPP div 8));
    end;
  end;
  DDrawReleaseScr();
end;

procedure DDrawActualizeVScreen ();
var
  y: Integer;
  a, scr: PChar;
begin
  scr := DDrawLockScr();
  if scr = nil then exit;
  for y := vScrHeight-1 downto 0 do
  begin
    a := @(dspVScreenRam[y, 0]);
    ddrawPutLine(a, scr);
    Inc(a, vScrWidth); Inc(scr, 640*(pBPP div 8));
  end;
  FillChar(scr^, 640*(pBPP div 8), 0);
  DDrawReleaseScr();
end;


var
  tDDraw: IDirectDraw7 = nil;

function IsDDrawHere (): Boolean;
begin
  result := false; tDDraw := nil;
  if not Assigned(DirectDrawCreate) then exit;
  if Failed(DirectDrawCreateEx(nil, tDDraw, IID_IDirectDraw7, nil)) then exit;
  tDDraw := nil;
  result := true;
end;


end.
