// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
unit dspWLispApiU;

interface

uses
  SysUtils,
  dspVScrU, dspWinU,
  alisp3U, lispCon3U;



implementation


function IsFalse (n: Integer): Boolean;
begin
  result := ALIsFalse(n) or ALIsNil(n);
end;

function BoolVal (b: Boolean): Integer;
begin
  if b then result := ALTrue() else result := ALFalse();
end;


function ALGetSym (n: Integer): Integer;
//(get-sym "str")
var
  s: string;
begin
  if not ALNextStrArg ('GET-SYM', n, s, result) then exit;
  result := ALCdr(ALFindGlobalSymbol(s));
end;

function ALKillSym (n: Integer): Integer;
//(kill-sym "str")
var
  s: string;
begin
  if not ALNextStrArg ('KILL-SYM', n, s, result) then exit;
  //ALDeleteGlobalSymbol(s);
  ALDeleteSymbolFromCurrentContext(s);
  result := ALNil();
end;

function ALSetSym (n: Integer): Integer;
//(set-sym "str" value)
var
  s: string;
begin
  if not ALNextStrArg ('SET-SYM', n, s, result) then exit;
  //ALAddGlobalSymbol(s, ALCar(n));
  ALSetGlobalSymbol(s, ALCar(n), true);
  result := ALCar(n);
end;

function ALPostBroadcast (n: Integer): Integer;
//(post-broadcast "name" bvalue ivalue tvalue)
var
  s: string;
  vb: Boolean;
  vi: Integer;
  vt: string;
begin
  if not ALNextStrArg ('POST-BROADCAST', n, s, result) then exit;
  if ALIsNil(ALCar(n)) or ALIsFalse(ALCar(n)) then vb := false else vb := true;
  n := ALCdr(n);
  if ALIsNil(ALCar(n)) then vi := 0
  else if not ALNextIntArg ('POST-BROADCAST', n, vi, result) then exit;
  if ALIsNil(ALCar(n)) then vt := ''
  else if not ALNextStrArg ('POST-BROADCAST', n, vt, result) then exit;

  if s <> '' then
  begin
    ZXWindowsPostMsg(s, nil, vb, vi, vt);
  end;
  result := ALNil();
end;

function ALPostMsg (n: Integer): Integer;
//(post-msg "ctrl" "name" bvalue ivalue tvalue)
var
  s, nm: string;
  vb: Boolean;
  vi: Integer;
  vt: string;
begin
  if ALIsNil(ALCar(n)) then begin result := ALNil(); exit; end;
  if not ALNextStrArg ('POST-MSG', n, s, result) then exit;
  if not ALNextStrArg ('POST-MSG', n, nm, result) then exit;
  if ALIsNil(ALCar(n)) or ALIsFalse(ALCar(n)) then vb := false else vb := true;
  n := ALCdr(n);
  if ALIsNil(ALCar(n)) then vi := 0
  else if not ALNextIntArg ('POST-MSG', n, vi, result) then exit;
  if ALIsNil(ALCar(n)) then vt := ''
  else if not ALNextStrArg ('POST-MSG', n, vt, result) then exit;

  if s <> '' then
  begin
    try
      ZXWindowsPostMsg(nm, TZXControl(StrToInt('$'+Copy(s, 1, Length(s)-1))), vb, vi, vt);
    except
    end;
  end;
  result := ALNil();
end;

function ALGetProp (n: Integer): Integer;
//(get-prop "control" "str")
var
  cs, ps: string;
begin
  if not ALNextStrArg ('GET-PROP', n, cs, result) then exit;
  if not ALNextStrArg ('GET-PROP', n, ps, result) then exit;
  result := ALCdr(ALFindGlobalSymbol(cs+ps));
end;

function ALSetProp (n: Integer): Integer;
//(set-prop "control" "str" value)
var
  v: Integer;
  cs, ps: string;
begin
  if not ALNextStrArg ('SET-PROP', n, cs, result) then exit;
  if not ALNextStrArg ('SET-PROP', n, ps, result) then exit;
  v := ALCar(n);
  //result := ALAddGlobalSymbol(cs+ps, v);
  ALSetGlobalSymbol(cs+ps, v, true);
  result := v;
end;

function ALNewControl (n: Integer): Integer;
//(new-control "type" "parent" proplist)
// proplist: "name" value ... "name" value
var
  cts, ctp, s: string;
  parent, ct: TZXControl;
begin
  if not ALNextStrArg ('NEW-CONTROL', n, cts, result) then exit;
  if ALIsNil(ALCar(n)) then begin ctp := ''; n := ALCdr(n); end
  else if not ALNextStrArg ('NEW-CONTROL', n, ctp, result) then exit;
  try
    if ctp = '' then parent := nil
    else
    begin
      parent := Pointer(StrToInt('$'+Copy(ctp, 1, Length(ctp)-1)));
      if not (TObject(parent) is TZXControl) then
      begin
        result := ALError('invalid control parent: "'+ctp+'"', n); exit;
      end;
    end;

    if cts = 'control' then ct := TZXControl.Create(parent)
    else if cts = 'button' then ct := TZXControl.Create(parent)
    else if cts = 'input' then ct := TZXTextInput.Create(parent)
    else if cts = 'label' then
    begin
      ct := TZXControl.Create(parent);
      ct.SetIntProperty('colors.frame', 255);
      ct.SetIntProperty('colors.disbled.frame', 255);
      ct.SetIntProperty('colors.back', 1);
      ct.SetIntProperty('colors.disbled.back', 1);
      ct.SetBoolProperty('clickable', false);
      ct.SetIntProperty('align.caption', zxaVCenter);
    end
    else if cts = 'window' then ct := TZXWindow.Create()
    else ct := nil;

    if ct = nil then
    begin
      result := ALError('invalid control type: "'+cts+'"', n); exit;
    end;
    if (parent = nil) and not (ct is TZXWindow) then zxWinList.Add(ct);

    while not ALIsNil(n) do
    begin
      if not ALNextStrArg ('NEW-CONTROL', n, s, result) then exit;
      ct.SetPropertyAtom(s, ALCar(n));
      n := ALCdr(n);
    end;
    result := ALNewStrAtom(IntToHex(Integer(ct), 8)+'.');
  except
    result := ALError('new-control fatal error', n);
  end;
end;


begin
  ALAddPrimitive('get-sym', ALGetSym, false, 1, 1);
  ALAddPrimitive('set-sym', ALSetSym, false, 2, 2);
  ALAddPrimitive('kill-sym', ALKillSym, false, 1, 1);
  ALAddPrimitive('post-broadcast', ALPostBroadcast, false, 1, -1);
  ALAddPrimitive('post-msg', ALPostMsg, false, 2, -1);
  ALAddPrimitive('get-prop', ALGetProp, false, 2, 2);
  ALAddPrimitive('set-prop', ALSetProp, false, 3, 3);
  ALAddPrimitive('new-control', ALNewControl, false, 2, -1);
end.
