// packing/depacking routines, v0.0.1
// coded by Ketmar // Vampire Avalon
// public domain (except of packer & depacked itself, which (c) Jibz)
// packer engine: WDOSX-Pack v1.07, (c) 1999-2001 by Joergen Ibsen / Jibz
// for data and executable compression software:
//   http://apack.cjb.net/   mailto:jibz@hotmail.com
// Delphi translation: Ketmar // Vampire Avalon
//
// block file format:
//   signature: 'WXB0'
//   word: # of blocks (not including the last partially filled one)
//   word: bytes in the last block
//     this field can be 0 if the last block is empty
// then full blocks:
//   4 bytes: unpacked data adler32
//   word: packed data size
//     if this field is 0: unpacked 65536 bytes follows,
//     else packed bytes follows
// then last block (if not empty):
//   4 bytes: adler32
//   byte: packed flag (<>0: block is packed)
//     if packed:
//       word: packed data size
//     data
//
{.$DEFINE WDXPACK_CRYPT_BLOCKZ}
{$DEFINE WDXPACK_INCLUDE_BLOCK_FUNCZ}

{$DESCRIPTION 'WDXPACK v0.0.1 by Ketmar // Vampire Avalon'}

{$MESSAGE ''}
{$MESSAGE 'compiling WDXPACK v0.0.1'}


{$IFDEF WDXPACK_INCLUDE_BLOCK_FUNCZ}
  {$IFDEF WDXPACK_CRYPT_BLOCKZ}
    {$MESSAGE 'WDXPACK: crypted block i/o included'}
  {$ELSE}
    {$MESSAGE 'WDXPACK: block i/o included'}
  {$ENDIF}
{$ENDIF}
{$MESSAGE ''}

{.$DEFINE DEBUG}

{$ASSERTIONS ON}
{$ALIGN 8}
{$BOOLEVAL OFF}
{$EXTENDEDSYNTAX ON}
{$IMPORTEDDATA OFF} // was ON
{$LONGSTRINGS ON}
{$IOCHECKS ON}
{$WRITEABLECONST OFF}
{$MINENUMSIZE 1}
{$OPENSTRINGS ON}
{$OVERFLOWCHECKS OFF}
{$RANGECHECKS OFF}
{$REALCOMPATIBILITY OFF}
{$SAFEDIVIDE OFF}
{$TYPEDADDRESS OFF}
{$TYPEINFO OFF}
{$VARSTRINGCHECKS OFF}
{$E-,F-,K-,N+,S-} //???

{$IFDEF DEBUG}
  {$DEBUGINFO ON}
  {$LOCALSYMBOLS ON}
  {$DEFINITIONINFO ON}
  {$HINTS ON}
  {$WARNINGS ON}
  {$OPTIMIZATION OFF}
  {$STACKFRAMES ON}
{$ELSE}
  {$DEBUGINFO OFF}
  {$LOCALSYMBOLS OFF}
  {$DEFINITIONINFO OFF}
  {$HINTS ON}
  {$WARNINGS OFF}
  {$OPTIMIZATION ON}
  {$STACKFRAMES OFF}
{$ENDIF}
unit wdxpackU;

interface

uses
  Windows, {$IFDEF WDXPACK_INCLUDE_BLOCK_FUNCZ}SysUtils, Classes{$ENDIF};


function WDXPackBuffer (const source; length: Integer;
  out destination; destSize: Integer): Integer;
  // uncompressible data may be expanded by up to 1 bit per byte.
  // this means that the destination should be length+((length+7)/8)+2 bytes.
  // returns size of the packed data or -1 on error
function WDXUnpackBuffer (const source; srcLength: Integer;
  out destination; length: Integer): Integer;
  // unpack length bytes from source to destination
  // returns # of unpacked bytes
  // destination buffer must be at least length bytes
  // returns -1 on error (out of source buffer)


{$IFDEF WDXPACK_INCLUDE_BLOCK_FUNCZ}
const
  WDXBLOCKFILESIGN = 'WXB0';


type
  TWDXBlockFile = class
  protected
    fStream: TStream;
    fCount: Integer;
    fOfs: TList; // offset to each block

    fLastBlockData: packed array [0..65536] of Byte;
      // last buffer
    fLastBlockUsed: Integer;
      // bytes used in the last buffer. can be 0 if
      // the last buffer is empty
    fChanged: Boolean;

    fCachedBlockData: packed array [0..65536] of Byte;
    fCachedBlock: Integer;
    fCurBlock: Integer;
    fCurOfs: Integer;

    function GetSize (): Integer;
    function GetPosition (): Integer;
    procedure SetPosition (value: Integer);

  public
    constructor Create (pStream: TStream; testAll: Boolean = false;
      skipSign: Boolean = false);
    constructor CreateNew (pStream: TStream; skipSign: Boolean = false);
    destructor Destroy (); override;

    function ReadBlock (index: Integer; out buf): Integer;
      // returns unpacked block size (1-65536)
      // raises an exception if there's no such block

    procedure ReadBuffer (out buf; count: Integer);
    procedure WriteBuffer (const buf; count: Integer);

    procedure Flush ();

    property Count: Integer read fCount;
    property Size: Integer read GetSize;
    property Position: Integer read GetPosition write SetPosition;
  end;

  TWDXStream = class (TStream)
  protected
    fBF: TWDXBlockFile;
    fNewPos: Integer;
    fPosChanged: Boolean;

  public
    destructor Destroy (); override;

    function Read (var buffer; count: LongInt): LongInt; override;
    function Write (const buffer; count: LongInt): LongInt; override;
    function Seek (const offset: Int64; origin: TSeekOrigin): Int64; override;
  end;

  TWDXUnpackStream = class (TWDXStream)
  public
    constructor Create (pStream: TStream; testAll: Boolean = false;
      skipSign: Boolean = false);
  end;

  TWDXPackStream = class (TWDXStream)
  public
    constructor Create (pStream: TStream; skipSign: Boolean = false);
  end;


function WDXIsBlockStream (st: TStream; seekBack: Boolean = true): Boolean;
// DOES BACK SEEKS!

{$ENDIF}


function WDXAdler32 (adler: LongWord; const buf; len: Integer): LongWord;
  // update a running Adler-32 checksum with the bytes buf[0..len-1] and
  // return the updated checksum.
  // the required initial value for the checksum is 1.
  // an Adler-32 checksum is almost as reliable as a CRC32 but can be
  // computed much faster.

function WDXAdler32Init (): LongWord;
  // return the initial value for the checksum calculation.


implementation


const
  LEN_2_LIMIT = 1920;


type
  PInt = ^Integer;
  PIntArr = ^TIntArr;
  TIntArr = packed array [0..(maxInt div 4)-128] of Integer;

  TMatch = record
    pos: Integer;
    len: Integer;
  end;


function WDXPackBuffer (const source; length: Integer;
  out destination; destSize: Integer): Integer;
  // global variables
  var
    nextBackEntry: Integer;
    backPtr, sourceOfs: PAnsiChar;
    inBuffer, outBuffer, tagByte: PAnsiChar;
    bitCount: Integer;
    outEnd: Integer; // end if output buffer

    // back-table
    backTable: PIntArr;

    // lookup-table
    lookup: packed array [0..255, 0..255] of Integer;

    lastMatchPos: Integer;
    lastWasMatch: Boolean;
    error: Boolean;


  procedure AdvanceTagByte (bit: Boolean);
  var
    b: Byte;
  begin
    if error then exit;
    // check bitcount and then decrement
    if bitCount = 0 then
    begin
      bitCount := 8;
      tagByte := outBuffer; Inc(outBuffer); //tagByte^ := #0;
    end;
    if Integer(tagByte) >= outEnd then begin error := true; exit; end;
    // shift in bit
    b := Byte(tagByte^) shl 1; if bit then b := b or $01;
    tagByte^ := Char(b);
    Dec(bitCount);
  end;

  // output Gamma2-code for val in range [2..?] ...
  procedure OutputGamma (val: Integer);
  var
    invertlen, invert: Integer;
  begin
    invertlen := 0; invert := 0;
    // rotate bits into invert (except last)
    repeat
      invert := (invert shl 1) or (val and $0001);
      Inc(invertlen);
      val := (val shr 1) and $7FFFFFFF;
    until not (val > 1);
    //while ((val >>= 1) > 1);

    // output Gamma2-encoded bits
    Dec(invertlen);
    while invertlen > 0 do
    begin
      AdvanceTagByte((invert and $01) <> 0);
      AdvanceTagByte(true);
      invert := invert shr 1;
      Dec(invertlen);
    end;
    AdvanceTagByte((invert and $01) <> 0);
    AdvanceTagByte(false);
  end;

  procedure OutputLiteral (lit: Char);
  begin
    //WriteLn('lit');
    lastWasMatch := false;
    AdvanceTagByte(false); // FALSE indicates a literal
    if error then exit;
    if Integer(outBuffer) >= outEnd then begin error := true; exit; end;
    outBuffer^ := lit; Inc(outBuffer); // output the literal
  end;

  procedure OutputCodePair (pos, len: Integer; buffer: PAnsiChar);
  begin
    //WriteLn(pos, ':', len);
    // if we just had a match, don't use lastMatchPos
    if lastWasMatch then
    begin
      // if a short match is too far away, encode it as two literals instead
      if (pos > LEN_2_LIMIT) and (len = 2) then
      begin
        OutputLiteral(buffer^);
        OutputLiteral((buffer+1)^);
      end
      else
      begin
        AdvanceTagByte(true); // TRUE indicates a match
        // a match more than LEN_2_LIMIT bytes back will be longer than 2
        if pos > LEN_2_LIMIT then Dec(len);
        OutputGamma(len); // output length
        lastMatchPos := pos; Dec(pos);
        ASSERT(pos >= 0);
        OutputGamma(((pos shr 6) and $3FFFFFFF)+2); // output high part of position
        // output low 6 bits of position
        AdvanceTagByte((pos and $20) <> 0);
        AdvanceTagByte((pos and $10) <> 0);
        AdvanceTagByte((pos and $08) <> 0);
        AdvanceTagByte((pos and $04) <> 0);
        AdvanceTagByte((pos and $02) <> 0);
        AdvanceTagByte((pos and $01) <> 0);
      end
    end
    else
    begin
      lastWasMatch := true;
      // if a short match is too far away, encode it as two literals instead
      if (pos > LEN_2_LIMIT) and (len = 2) and (pos <> lastMatchPos) then
      begin
        OutputLiteral(buffer^);
        OutputLiteral((buffer+1)^);
      end
      else
      begin
        // TRUE indicates a match
        AdvanceTagByte(true);
        // a match more than LEN_2_LIMIT bytes back will be longer than 2
        if (pos > LEN_2_LIMIT) and (pos <> lastMatchPos) then Dec(len);
        OutputGamma(len); // output length
        // output position
        if pos = lastMatchPos then
        begin
          // a match with position 0 means use last position
          AdvanceTagByte(false);
          AdvanceTagByte(false);
        end
        else
        begin
          lastMatchPos := pos; Dec(pos);
          ASSERT(pos >= 0);
          OutputGamma(((pos shr 6) and $3FFFFFFF)+3); // output high part of position
          // output low 6 bits of position
          AdvanceTagByte((pos and $20) <> 0);
          AdvanceTagByte((pos and $10) <> 0);
          AdvanceTagByte((pos and $08) <> 0);
          AdvanceTagByte((pos and $04) <> 0);
          AdvanceTagByte((pos and $02) <> 0);
          AdvanceTagByte((pos and $01) <> 0);
        end;
      end;
    end;
  end;

  procedure OutputEOD ();
  begin
    // just switch last tagByte into position
    if error then exit;
    if bitCount = 8 then exit;
    tagByte^ := Char(Byte(tagByte^) shl bitCount);
  end;

  procedure FindMatch (var thematch: TMatch; buffer: PAnsiChar; lookback, lookforward: Integer);
  var
    backPos, matchLen, bestMatchLen, bestMatchPos: Integer;
    ptr: PAnsiChar;
    i0, i1: Integer;
    idx0, idx1: Integer;
  begin
    // temporary variables to avoid indirect addressing into the match
    bestMatchLen := 0; bestMatchPos := 0;

    // update lookup- and backtable up to current position
    while backPtr < buffer do
    begin
      idx0 := Byte(backPtr^);
      idx1 := Byte(PAnsiChar(backPtr+1)^);
      //PInt(Integer(backTable)+nextBackEntry*4)^ := lookup[idx0, idx1];
      backTable^[nextBackEntry] := lookup[idx0, idx1];
      lookup[idx0, idx1] := nextBackEntry;
      Inc(nextBackEntry); Inc(backPtr);
    end;

    // get position by looking up next two bytes
    backPos := lookup[Byte(buffer^), Byte(PAnsiChar(buffer+1)^)];

    if (backPos <> 0) and (lookforward > 1) then
    begin
      ptr := backPos+sourceOfs;

      // go backwards until before buffer
      while (ptr >= buffer) and (backPos <> 0) do
      begin
        //backPos := PInt(Integer(backTable)+backPos*4)^;
        backPos := backTable^[backPos];
        ptr := backPos+sourceOfs;
      end;

      // search through table entries
      while (backPos <> 0) and (buffer-ptr <= lookback) do
      begin
        matchLen := 2;
        // if this position has a chance to be better
        if PAnsiChar(ptr+bestMatchLen)^ = PAnsiChar(buffer+bestMatchLen)^ then
        begin
          // scan it
          while (PAnsiChar(ptr+matchLen)^ = PAnsiChar(buffer+matchLen)^) and
                (matchLen < lookforward) do Inc(matchLen);
          // check it
          if buffer-ptr = lastMatchPos then i0 := 1 else i0 := 0;
          if bestMatchPos = lastMatchPos then i1 := 1 else i1 := 0;
          if matchLen+i0 > bestMatchLen+i1 then
          begin
            bestMatchLen := matchLen;
            if bestMatchLen = lookforward then backPos := 0;
            bestMatchPos := buffer-ptr;
          end;
        end;
        // move backwards to next position
        backPos := PInt(Integer(backTable)+backPos*4)^;
        ptr := backPos+sourceOfs;
      end; // while ((backPos) && (buffer - ptr <= lookback))
    end; // if ((backPos) && (lookforward > 1))

    // forget match if too far away
    if (bestMatchPos > LEN_2_LIMIT) and
       (bestMatchLen = 2) and
       (bestMatchPos <> lastMatchPos) then
    begin
      bestMatchLen := 0; bestMatchPos := 0;
    end;

    // update the match with best match
    thematch.len := bestMatchLen;
    thematch.pos := bestMatchPos;
  end;

var
  match, nextmatch, literalmatch, testmatch: TMatch;
  pos, lastpos, literalCount: Integer;
  i, j: Integer;
  i0, i1: Integer;
  workMem: Pointer;
begin
  if length < 1 then begin result := 0; exit; end;
  if destSize < 1 then begin result := -1; exit; end;
  GetMem(workMem, (length+4)*4);
  error := false;

  try
    inBuffer := @source; outBuffer := @destination;
    outEnd := Integer(outBuffer)+destSize;
    sourceOfs := inBuffer-1;
    backTable := Pointer(workMem);

    // init lookup- and backtable
    for i := 0 to 255 do for j := 0 to 255 do lookup[i][j] := 0;

    backPtr := inBuffer; backTable^[0] := 0; nextBackEntry := 1;

    lastpos := -1; lastMatchPos := -1;
    lastWasMatch := false; literalCount := 0;

    // the first byte is sent verbatim
    outBuffer^ := inBuffer^;
    Inc(outBuffer); Inc(inBuffer);

    // init tag-byte
    bitCount := 8; tagByte := outBuffer; Inc(outBuffer);

    // pack data
    pos := 1;
    while pos < length do
    begin
      // find best match at current position (if not already found)
      if pos = lastpos then
      begin
        match.len := nextmatch.len; match.pos := nextmatch.pos;
      end
      else FindMatch(match, inBuffer, pos, length-pos);

      // if we found a match, find the best match at the next position
      if match.len <> 0 then
      begin
        FindMatch(nextmatch, inBuffer+1, pos+1, length-(pos+1));
        lastpos := pos+1;
      end
      else nextmatch.len := 0;

      // decide if we should output a match or a literal
      if match.pos = lastMatchPos then i0 := 1 else i0 := 0;
      if nextmatch.pos = lastMatchPos then i1 := 1 else i1 := 0;
      if (match.len <> 0) and (match.len+i0 >= nextmatch.len+i1) then
      begin
        // output any pending literals
        if literalCount <> 0 then
        begin
          if literalCount = 1 then
          begin
            OutputLiteral(PAnsiChar(inBuffer-1)^);
          end
          else
          begin
            // check if there is a closer match with the required length
            FindMatch(testmatch, inBuffer-literalCount, literalmatch.pos, literalCount);
            if testmatch.len >= literalCount then
            begin
              OutputCodePair(testmatch.pos, literalCount, inBuffer-literalCount);
            end
            else
            begin
              OutputCodePair(literalmatch.pos, literalCount, inBuffer-literalCount);
            end;
          end;
          literalCount := 0;
        end;
        // output match
        OutputCodePair(match.pos, match.len, inBuffer);
        Inc(inBuffer, match.len);
        Inc(pos, match.len);
      end
      else
      begin
        // if ((match.len) && ( ... )
        // check if we are allready collecting literals
        if literalCount <> 0 then
        begin
          // if so, continue..
          Inc(literalCount);
          // have we collected as many as possible?
          if literalCount = literalmatch.len then
          begin
            OutputCodePair(literalmatch.pos, literalCount, inBuffer-literalCount+1);
            literalCount := 0;
          end;
        end
        else
        begin
          // if (literalCount)
          // if we had a match which was not good enough, then save it..
          if match.len <> 0 then
          begin
            literalmatch.len := match.len;
            literalmatch.pos := match.pos;
            Inc(literalCount);
          end
          else
          begin
            // if not, we have to output the literal now
            OutputLiteral(inBuffer^);
          end;
        end; // if (literalCount)
        Inc(inBuffer); Inc(pos);
      end; // if ((match.len) && ( ... )
    end;

    // output any remaining literal bytes
    if literalCount <> 0 then
    begin
      if literalCount = 1 then OutputLiteral(PAnsiChar(inBuffer-1)^)
      else OutputCodePair(literalmatch.pos, literalCount, inBuffer-literalCount);
      //literalCount := 0;
    end;

    // do EOD stuff
    OutputEOD();

    if error then result := -1
    else result := Integer(outBuffer)-Integer(@destination);
  finally
    FreeMem(workMem);
  end;
end;


function WDXUnpackBuffer (const source; srcLength: Integer;
  out destination; length: Integer): Integer;
var
  fbyte: Byte;
  src, dest, pp: PAnsiChar;
  lastMatchPos: Integer;
  lastWasMatch: Boolean;
  len, pos: Integer;
  error: Boolean;

  function GetBit (): Boolean;
  begin
    result := (fbyte >= $80); fbyte := (fbyte and $7F) shl 1;
    if error then exit;
    if fbyte = 0 then
    begin
      if srcLength < 1 then begin error := true; exit; end;
      fbyte := Byte(src^); Inc(src); Dec(srcLength);
      result := (fbyte >= $80); fbyte := ((fbyte and $7F) shl 1) or $01;
    end;
  end;

  function GetGamma (): Integer;
  begin
    result := 1;
    if error then exit;
    repeat
      Inc(result, result); //result := result shl 1;
      if GetBit() then Inc(result); //result := result or $01;
    until not GetBit();
  end;

  function GetLoPos (pos: Integer): Integer;
  // get 6 low bits of position
  var
    f: Integer;
  begin
    for f := 0 to 5 do
    begin
      Inc(pos, pos); //pos := pos shl 1;
      if GetBit() then Inc(pos); //pos := pos or $01;
    end;
    result := pos;
  end;

begin
  result := 0; if length < 1 then exit;
  if srcLength < 1 then begin result := -1; exit; end;
  src := @source; dest := @destination;
  // the first byte was sent verbatim

  dest^ := src^; Inc(dest); Inc(src); Dec(length); Dec(srcLength);
  if length < 1 then begin result := 1; exit; end;
  lastMatchPos := 0; lastWasMatch := false; fbyte := 0;

  error := false;
  repeat
    if not GetBit() then
    begin
      // literal
      //WriteLn('lit');
      if srcLength < 1 then begin error := true; break; end;
      dest^ := src^; Inc(dest); Inc(src); Dec(length); Dec(srcLength);
      lastWasMatch := false;
    end
    else
    begin
      // match
      len := GetGamma(); // get length
      if error then break;
      // same position as last match?
      if lastWasMatch then
      begin
        pos := GetGamma()-2;
        if error then break;
        pos := GetLoPos(pos)+1;
        if error then break;
        lastMatchPos := pos;
        if pos > LEN_2_LIMIT then Inc(len);
      end
      else
      begin
        lastWasMatch := true;
        pos := GetGamma()-2;
        if error then break;
        if pos = 0 then pos := lastMatchPos
        else
        begin
          pos := GetLoPos(pos-1)+1;
          if error then break;
          lastMatchPos := pos;
          if pos > LEN_2_LIMIT then Inc(len);
        end;
      end;
      //WriteLn(pos, ':', len);
      if error then break;
      // copy match
      pp := dest; Dec(pp, pos);
      while (len > 0) and (length > 0) do
      begin
        dest^ := pp^; Inc(dest); Inc(pp); Dec(length); Dec(len);
      end;
    end;
  until (length < 1) or error;

  if error then result := -1 else result := Integer(dest)-Integer(@destination);
end;

const
  BASE = 65521; // largest prime smaller than 65536
  NMAX = 5552;
   // NMAX = 5552; original code with unsigned 32 bit integer
   // NMAX = 3854; code with signed 32 bit integer
   // NMAX is the largest n such that 255*n*(n+1)/2+(n+1)*(BASE-1)<=2^32-1
   // the penalty is the time loss in the extra MOD-calls.

function WDXAdler32 (adler: LongWord; const buf; len: Integer): LongWord;
var
  s1, s2, k: LongWord;
  p: PChar;
begin
  result := adler;
  if len < 0 then exit;
  p := @buf;
  s1 := (adler and $FFFF);
  s2 := (adler shr 16) and $FFFF;
  while len > 0 do
  begin
    if len < NMAX then k := len else k := NMAX;
    Dec(len, k);
    while k > 0 do
    begin
      Inc(s1, Byte(p^)); Inc(s2, s1);
      Inc(p); Dec(k);
    end;
    s1 := s1 mod BASE;
    s2 := s2 mod BASE;
  end;
  result := (s2 shl 16) or s1;
end;

function WDXAdler32Init (): LongWord;
begin
  result := 1;
end;


{$IFDEF WDXPACK_CRYPT_BLOCKZ}
function RandIntSeeded (var seed: Integer): Integer;
var
  k1, ix: Integer;
begin
  ix := seed;

  k1 := ix div 127773;
  ix := 16807*(ix-k1*127773)-k1*2836;
  if ix < 0 then Inc(ix, 2147483647);
  seed := ix;

  result := ix;
end;

function RandByteSeeded (var seed: Integer): Byte;
begin
  result := (RandIntSeeded(seed) and 255);
end;

procedure XorCopyBuf (const src; len: Integer; out dest; seed: Integer);
var
  s, d: PChar;
begin
  s := @src; d := @dest;
  while len > 0 do
  begin
    d^ := Char(Byte(s^) xor RandByteSeeded(seed));
    Inc(s); Inc(d); Dec(len);
  end;
end;

procedure XorBuf (const src; len: Integer; seed: Integer);
var
  s, d: PChar;
begin
  s := @src; d := @src;
  while len > 0 do
  begin
    d^ := Char(Byte(s^) xor RandByteSeeded(seed));
    Inc(s); Inc(d); Dec(len);
  end;
end;
{$ENDIF}


{$IFDEF WDXPACK_INCLUDE_BLOCK_FUNCZ}
function WDXIsBlockStream (st: TStream; seekBack: Boolean = true): Boolean;
var
  sign: packed array [0..3] of Char;
begin
  result := (st.Size-st.Position) >= 8;
  if result then
  begin
    st.ReadBuffer(sign[0], 4);
    if seekBack then st.Seek(-4, soFromCurrent);
    result := (sign = WDXBLOCKFILESIGN);
  end;
end;

procedure NTFSClearCompression (fh: Cardinal);
var
  bytesReturned: Cardinal;
  buffer: Short;
begin
  buffer := 0;
  DeviceIoControl(fh, $0009C040, @buffer, SizeOf(Short), nil, 0, bytesReturned, nil);
end;

procedure NTFSFileClearCompression (const fileName: string);
var
  fh: Integer;
begin
  fh := FileOpen(fileName, fmOpenReadWrite or fmShareExclusive);
  if fh > -1 then
  begin
    NTFSClearCompression(fh);
    FileClose(fh);
  end;
end;

{$IFDEF WDXPACK_CRYPT_BLOCKZ}
function RandIntSeeded (var seed: Integer): Integer;
var
  k1, ix: Integer;
begin
  ix := seed;

  k1 := ix div 127773;
  ix := 16807*(ix-k1*127773)-k1*2836;
  if ix < 0 then Inc(ix, 2147483647);
  seed := ix;

  result := ix;
end;

function RandByteSeeded (var seed: Integer): Byte;
begin
  result := (RandIntSeeded(seed) and 255);
end;

procedure XorCopyBuf (const src; len: Integer; out dest; seed: Integer);
var
  s, d: PChar;
begin
  s := @src; d := @dest;
  while len > 0 do
  begin
    d^ := Char(Byte(s^) xor RandByteSeeded(seed));
    Inc(s); Inc(d); Dec(len);
  end;
end;

procedure XorBuf (const src; len: Integer; seed: Integer);
var
  s, d: PChar;
begin
  s := @src; d := @src;
  while len > 0 do
  begin
    d^ := Char(Byte(s^) xor RandByteSeeded(seed));
    Inc(s); Inc(d); Dec(len);
  end;
end;
{$ENDIF}


{ TWDXBlockFile }
constructor TWDXBlockFile.CreateNew (pStream: TStream; skipSign: Boolean = false);
// create new block file
var
  w: Word;
  sign: packed array [0..3] of Char;
begin
  if pStream is THandleStream then
    NTFSClearCompression(THandleStream(pStream).Handle);
  fCachedBlock := -1; fCurBlock := 0; fCurOfs := 0; fCount := 1;
  fStream := pStream;
  // write header
  if not skipSign then
  begin
    sign := WDXBLOCKFILESIGN; fStream.WriteBuffer(sign[0], 4);
  end;
  w := 0; fStream.WriteBuffer(w, 2); // # of blocks
  {w := 0; }fStream.WriteBuffer(w, 2); // bytes in the last block
  fLastBlockUsed := 0;
  fChanged := false;

  fOfs := TList.Create();
  fOfs.Add(Pointer(fStream.Position));
end;

constructor TWDXBlockFile.Create (pStream: TStream; testAll: Boolean = false;
  skipSign: Boolean = false);
var
  w: Word;
  f, c, usz: Integer;
  ad32, cad32: LongWord;
  buf: Pointer;
  b: Byte;
  sign: packed array [0..3] of Char;
begin
  fCachedBlock := -1; fCurBlock := 0; fCurOfs := 0; fChanged := false;
  fStream := pStream;
  // read header
  if not skipSign then
  begin
    fStream.ReadBuffer(sign[0], 4);
    if sign <> WDXBLOCKFILESIGN then raise EReadError.Create('BlockFile: invalid signature');
  end;
  fStream.ReadBuffer(w, 2); fCount := Integer(w)+1;
  fStream.ReadBuffer(w, 2); fLastBlockUsed := w;

  GetMem(buf, 65536);
  try
    fOfs := TList.Create();
    // scan blocks
    try
      for f := 2 to fCount do  // fCount-1 full blocks
      begin
        fOfs.Add(Pointer(fStream.Position));
        fStream.ReadBuffer(ad32, SizeOf(ad32)); // skip md5
        fStream.ReadBuffer(w, 2); // size of the packed data
        c := w; if c = 0 then c := 65536; // size 0: unpacked block
        if testAll then
        begin
          // unpack & test the block
          if c = 65536 then
          begin
            fStream.ReadBuffer(fLastBlockData[0], 65536);
            {$IFDEF WDXPACK_CRYPT_BLOCKZ}
            XorBuf(fLastBlockData[0], 65536, 666);
            {$ENDIF}
          end
          else
          begin
            fStream.ReadBuffer(buf^, c);
            {$IFDEF WDXPACK_CRYPT_BLOCKZ}
            XorBuf(buf^, c, 666);
            {$ENDIF}
            usz := WDXUnpackBuffer(buf^, c, fLastBlockData[0], 65536);
            if usz <> 65536 then
              raise EReadError.Create('BlockFile: block #'+IntToHex(f-2, 4)+' is corrupted');
          end;
          cad32 := WDXAdler32Init();
          cad32 := WDXAdler32(cad32, fLastBlockData[0], 65536);

          if ad32 <> cad32 then
            raise EReadError.Create('BlockFile: block #'+IntToHex(f-2, 4)+' is corrupted (adler32)');

          if fCachedBlock = -1 then
          begin
            // cache first block
            fCachedBlock := f-2;
            Move(fLastBlockData[0], fCachedBlockData[0], 65536);
          end;
        end
        else fStream.Seek(c, soFromCurrent); // skip block data
      end;

      // add last partially full block
      fOfs.Add(Pointer(fStream.Position));
      if fLastBlockUsed > 0 then
      begin
        // there is some data, so cache last block
        fStream.ReadBuffer(ad32, SizeOf(ad32));
        fStream.ReadBuffer(b, 1); // "packed" flag
        if b = 0 then
        begin
           // not packed
          fStream.ReadBuffer(fLastBlockData[0], fLastBlockUsed);
          {$IFDEF WDXPACK_CRYPT_BLOCKZ}
          XorBuf(fLastBlockData[0], fLastBlockUsed, 666);
          {$ENDIF}
        end
        else
        begin
          // packed
          fStream.ReadBuffer(w, 2); c := w; // size of the packed data
          ASSERT(w <> 0); // this should never fail!
          fStream.ReadBuffer(buf^, c);
          {$IFDEF WDXPACK_CRYPT_BLOCKZ}
          XorBuf(buf^, c, 666);
          {$ENDIF}
          usz := WDXUnpackBuffer(buf^, c, fLastBlockData[0], fLastBlockUsed);
          if usz <> fLastBlockUsed then
            raise EReadError.Create('BlockFile: the last block is corrupted');
        end;

        // check adler32
        cad32 := WDXAdler32Init();
        cad32 := WDXAdler32(cad32, fLastBlockData[0], fLastBlockUsed);
        if ad32 <> cad32 then raise EReadError.Create('BlockFile: the last block is corrupted (adler32)');
      end;
    except
      fOfs.Free();
      raise;
    end;
  finally
    ReallocMem(buf, 0);
  end;
end;

destructor TWDXBlockFile.Destroy ();
begin
  if fChanged then Flush();
  fOfs.Free();
end;

function TWDXBlockFile.GetPosition (): Integer;
begin
  result := fCurBlock*65536+fCurOfs;
end;

procedure TWDXBlockFile.SetPosition (value: Integer);
var
  sz: Integer;
begin
  sz := (fCount-1)*65536+fLastBlockUsed;
  if value < 0 then value := 0
  else if value > sz then value := sz;
  fCurBlock := value div 65536; fCurOfs := value mod 65536;
end;

function TWDXBlockFile.GetSize (): Integer;
begin
  result := (fCount-1)*65536+fLastBlockUsed;
end;

function TWDXBlockFile.ReadBlock (index: Integer; out buf): Integer;
// returns unpacked block size (1-65536)
// raises an exception if there's no such block
var
  rbuf: Pointer;
  w: Word;
  ad32, cad32: LongWord;
  c: Integer;
begin
  if (index < 0) or (index >= fCount) then
    raise EReadError.Create('BlockFile: block #'+IntToHex(index, 4)+' not found');
  if index = fCount-1 then
  begin
    result := fLastBlockUsed;
    if result <> 0 then Move(fLastBlockData[0], buf, result);
    exit;
  end;

  result := 65536;
  fStream.Position := Integer(fOfs[index]);
  GetMem(rbuf, 65536);
  try
    fStream.ReadBuffer(ad32, SizeOf(ad32));
    fStream.ReadBuffer(w, 2); c := w;
    if c = 0 then
    begin
      fStream.ReadBuffer(buf, 65536);
      {$IFDEF WDXPACK_CRYPT_BLOCKZ}
      XorBuf(buf, 65536, 666);
      {$ENDIF}
    end
    else
    begin
      fStream.ReadBuffer(rbuf^, c);
      {$IFDEF WDXPACK_CRYPT_BLOCKZ}
      XorBuf(rbuf^, c, 666);
      {$ENDIF}
      c := WDXUnpackBuffer(rbuf^, c, buf, 65536);
      if c <> 65536 then
        raise EReadError.Create('BlockFile: block '+IntToHex(index, 4)+' is corrupted');
    end;

    cad32 := WDXAdler32Init();
    cad32 := WDXAdler32(cad32, buf, 65536);

    if ad32 <> cad32 then
      raise EReadError.Create('BlockFile: block #'+IntToHex(index, 4)+' is corrupted (adler32)');
  finally
    ReallocMem(rbuf, 0);
  end;
end;

procedure TWDXBlockFile.Flush ();
var
  buf: Pointer;
  c: Integer;
  w: Word;
  ad32: LongWord;
begin
  if not fChanged then exit;

  ASSERT(fLastBlockUsed > 0);
  ASSERT(fLastBlockUsed <= 65535);

  fStream.Position := Integer(fOfs[0])-2;
  w := fLastBlockUsed; fStream.WriteBuffer(w, 2);

  fStream.Position := Integer(fOfs[fCount-1]);
  if fLastBlockUsed > 0 then
  begin
    ad32 := WDXAdler32Init();
    ad32 := WDXAdler32(ad32, fLastBlockData[0], fLastBlockUsed);

    GetMem(buf, 65536);
    try
      fStream.WriteBuffer(ad32, SizeOf(ad32));

      c := WDXPackBuffer(fLastBlockData[0], fLastBlockUsed, buf^, 65535);
      if (c = -1) or (c >= fLastBlockUsed-2) then
      begin
        // the block is uncompressible
        w := 0; fStream.WriteBuffer(w, 1); // "uncompressed" flag
        {$IFDEF WDXPACK_CRYPT_BLOCKZ}
        XorBuf(fLastBlockData[0], fLastBlockUsed, 666);
        {$ENDIF}
        fStream.WriteBuffer(fLastBlockData[0], fLastBlockUsed);
        {$IFDEF WDXPACK_CRYPT_BLOCKZ}
        XorBuf(fLastBlockData[0], fLastBlockUsed, 666);
        {$ENDIF}
      end
      else
      begin
        w := 1; fStream.WriteBuffer(w, 1); // "compressed" flag
        w := c; fStream.WriteBuffer(w, 2);
        {$IFDEF WDXPACK_CRYPT_BLOCKZ}
        XorBuf(buf^, c, 666);
        {$ENDIF}
        fStream.WriteBuffer(buf^, c);
      end;
    finally
      ReallocMem(buf, 0);
    end;
  end;

  fStream.Size := fStream.Position;
  fChanged := false;
end;

procedure TWDXBlockFile.WriteBuffer (const buf; count: Integer);
var
  p: PChar;
  wrt, c: Integer;
  ad32: LongWord;
  wbuf: Pointer;
  w: Word;
begin
  p := @buf;
  while count > 0 do
  begin
    wrt := count;
    if fLastBlockUsed+wrt > 65536 then wrt := 65536-fLastBlockUsed;
    Move(p^, fLastBlockData[fLastBlockUsed], wrt);
    Inc(fLastBlockUsed, wrt); Inc(p, wrt); Dec(count, wrt);
    fChanged := true;
    ASSERT(fLastBlockUsed <= 65536);
    if fLastBlockUsed = 65536 then
    begin
      // write full block
      if fCount = 65535 then raise EWriteError.Create('BlockFile: too much blocks');
      fStream.Position := Integer(fOfs[fCount-1]);

      ad32 := WDXAdler32Init();
      ad32 := WDXAdler32(ad32, fLastBlockData[0], 65536);

      GetMem(wbuf, 65536);
      try
        c := WDXPackBuffer(fLastBlockData[0], 65536, wbuf^, 65535);
        if c = -1 then
        begin
          // the block is uncompressible
          w := 0; Move(fLastBlockData[0], wbuf^, 65536); c := 65536;
        end
        else w := c;

        fStream.WriteBuffer(ad32, SizeOf(ad32));
        fStream.WriteBuffer(w, 2);
        {$IFDEF WDXPACK_CRYPT_BLOCKZ}
        XorBuf(wbuf^, c, 666);
        {$ENDIF}
        fStream.WriteBuffer(wbuf^, c);
      finally
        ReallocMem(wbuf, 0);
      end;

      Inc(fCount);
      fOfs.Add(Pointer(fStream.Position));
      fStream.Position := Integer(fOfs[0])-4;
      w := fCount-1; fStream.WriteBuffer(w, 2); // block count
      w := 0; fStream.WriteBuffer(w, 2); // last block data size
      fLastBlockUsed := 0;
      fChanged := false;
    end;
  end;
end;

procedure TWDXBlockFile.ReadBuffer (out buf; count: Integer);
var
  dp, sp: PChar;
  blkSz: Integer;
  rd: Integer;
begin
  dp := @buf;
  while count > 0 do
  begin
    if GetPosition() >= GetSize() then
      raise EWriteError.Create('BlockFile: read error (out of data)');

    if fCurBlock = fCount-1 then
    begin
      sp := @(fLastBlockData[0]); blkSz := fLastBlockUsed;
    end
    else
    begin
      if fCurBlock <> fCachedBlock then
      begin
        try
          fCachedBlock := fCurBlock;
          ReadBlock(fCachedBlock, fCachedBlockData[0]);
        except
          fCachedBlock := -1;
          raise;
        end;
      end;
      sp := @(fCachedBlockData[0]);
      blkSz := 65536;
    end;

    rd := count;
    if fCurOfs+rd > blkSz then rd := blkSz-fCurOfs;
    Move(sp[fCurOfs], dp^, rd);
    Inc(fCurOfs, rd); Inc(dp, rd); Dec(count, rd);
    ASSERT(fCurOfs <= 65536);
    if fCurOfs = 65536 then begin fCurOfs := 0; Inc(fCurBlock); end;
  end;
end;


{ TWDXStream }
destructor TWDXStream.Destroy ();
begin
  fBF.Free();
  inherited Destroy();
end;

function TWDXStream.Seek (const offset: Int64; origin: TSeekOrigin): Int64;
begin
  case origin of
    soCurrent:
      begin
        if offset = 0 then
        begin
          if fPosChanged then result := fNewPos else result := fBF.Position;
        end
        else
        begin
          fPosChanged := true;
          result := fBF.Position+offset;
        end;
      end;
    soEnd:
      begin
        if offset = 0 then
        begin
          fPosChanged := true; result := fBF.Size;
        end
        else
        begin
          fPosChanged := true;
          result := fBF.Size+offset;
        end;
      end;
    else // soBeginning
      begin
        fPosChanged := true;
        fBF.Position := offset;
        result := offset;
      end;
  end;
  fBF.Position := result;
end;

function TWDXStream.Read (var buffer; count: LongInt): LongInt;
begin
  fBF.Position := fNewPos;
  fBF.ReadBuffer(buffer, count); result := count;
  fNewPos := fBF.Position;
end;

function TWDXStream.Write (const buffer; count: LongInt): LongInt;
begin
  fBF.WriteBuffer(buffer, count); result := count;
end;


{ TWDXUnpackStream }
constructor TWDXUnpackStream.Create (pStream: TStream; testAll: Boolean = false;
  skipSign: Boolean = false);
begin
  inherited Create();
  fBF := TWDXBlockFile.Create(pStream, testAll, skipSign);
  fNewPos := 0;
end;


{ TWDXPackStream }
constructor TWDXPackStream.Create (pStream: TStream; skipSign: Boolean = false);
begin
  inherited Create();
  fBF := TWDXBlockFile.CreateNew(pStream, skipSign);
  fNewPos := 0;
end;
{$ENDIF}


end.
